/* {{{ preamble
 * regmod.c
 *
 * Command-line Regedit-like functionality
 *
 * Authors:	Sean Loaring
 *
 * Copyright (c) 2005-2006 Sean Loaring
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 }}} */

// {{{ Includes
#include <iconv.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <errno.h>

#include <signal.h>
#include <ctype.h>
#include <wchar.h>
#include <locale.h>

#include "../lib/regtypes.h"
#include "../lib/libhive.h"
#include "../lib/defines.h"
#include "../misc/uerr.h"
#include "../hivetools/nstdreg.h"
#include "../misc/tools.h"
#include "../misc/globals.h"

// }}}
// {{{ Defines

#define RM_OUTPUT_8  0
#define RM_OUTPUT_16 1

#define RM_MODE_EXPORT 0
#define RM_MODE_IMPORT 1
#define RM_MODE_DELETE 2
#define RM_MODE_MK     3
#define RM_MODE_MV     4
// }}}

int outmode = RM_OUTPUT_16;

FILE *outfd;

char utf16_pref[] = { 0xff, 0xfe };

int write_output( char *buf, int blen ) // {{{ Write contents of buf (less zero-termination) to outfd in utf8 or 16
{
	VBUF_STATIC( utf16buf, 10 );
	static int wrote_utf16_prefix = 0;

	switch( outmode ) {
		case RM_OUTPUT_8:
			if( 1 != fwrite( buf, blen - 1, 1, outfd ) )
				DIE("Write error: %s", strerror( errno ) );
			break;
		case RM_OUTPUT_16:

			if( !wrote_utf16_prefix ) {
				if( 1 != fwrite( utf16_pref, 2, 1, outfd ) ) {
					//TODO: error
					fprintf(stderr, "Can't write UTF16 prefix.\n");
					return -1;
				}
				wrote_utf16_prefix = 1;
			}
			if( -1 == vb_utf8to16( utf16buf, buf, blen ) )
				WARN("Problem converting to UTF-16: %s",  uerr_str( uerr_get() ) );

			if( 1 != fwrite( utf16buf->b, utf16buf->dlen - 2, 1, outfd ) )
				DIE("Write error: %s", strerror( errno ) );
			break;
		default:
			DIE("Unknown output mode %d", outmode );
	}

	return 0;
} // }}}
int export_reg( char *path ) // {{{
{
	return ns_reg_export( path, stupid_cr, &write_output );
} // }}}
int import_reg( char *path ) // {{{
{
	if( 0 != ns_reg_merge( path ) ) {
		fprintf(stderr, "Error merging file %s into registry: %s\n",
				path, uerr_str(uerr_get()) );
		return -1;
	}

	return 0;
} // }}}
void usage(void) // {{{
{
	fprintf(stderr,
		"Usage: regmod options mount_options\n"
		"Modify an NT registry file from the command line\n"
		"\n"
		" -h                        This message\n"                                                            
		" -v                        Be a little more verbose (for debuging)\n"
		" -8                        Output in UTF-8 (default UTF-16)\n"
		" -i file                   Import .reg file\n"
		" -d                        Debug outout\n"
		" -o file                   output to file (default stdout)\n"
		//" -N                        No allocation mode. Only (old style) same length overwrites possible\n"
		//" -u                        Ugly output (for script processing).\n"
		" -e key                    Export a key in .reg format\n"
		" -d key/val                Delete a key or value\n"
		" -m key                    Make a new key\n"
		" -s val_path=type=value    Create a new value\n"
		"\n"
		"     e.g.   regmod -s /HKEY_LOCAL_MACHINE/SOFTWARE/foo/bleh=SZ=\"This is a string.\"\n\n"
		"See readme file on how to extract/read/write the NT's SAM file\n"
		"if it's on an NTFS partition!\n\n"
	);

	fprintf(stderr, "%s", hl_mount_help() );
}
// }}}
void set_outfile( char *fname ) // {{{
{
	if( NULL == ( outfd = fopen( fname, "w" ) ) ) 
		DIE("Couldn't open output file (%s) for write: %s\n", fname, strerror( errno ) );
} // }}}
int set_mode( int mode ) // {{{
{
	static int setmode=-1;

	if( -1 != setmode ) {
		usage();
		fprintf(stderr, "More than one operation specified.");
		exit(1);
	}

	return mode;

} // }}}
int main(int argc, char **argv) // {{{
{
	int mode = -1;
	extern int optind;
	extern char *optarg;
	char *outfile = NULL;
	char c;
	char *regpath = NULL;
	char *del_path = NULL;
	char *import_str = NULL;
	char *mk_path = NULL;
	char *mk_str = NULL;
	int r = 0; //return value

	char *options = "ve:i:d:o:8hm:s:";

	outfd = stdout;

	if( 1 == argc ) {
		usage();
		exit(1);
	}

	unicode_init();
	libhive_init();
	ns_init();

	while((c=getopt(argc,argv,options)) > 0) {
		switch(c) {
			case 'v': gverbose = 1; break;
			case 'e': mode = set_mode( RM_MODE_EXPORT ); regpath    = optarg; break;
			case 'i': mode = set_mode( RM_MODE_IMPORT ); import_str = optarg; break;
			case 'd': mode = set_mode( RM_MODE_DELETE ); del_path   = optarg; break;
			case 'm': mode = set_mode( RM_MODE_MK ); mk_path = optarg; break;
			case 's': mode = set_mode( RM_MODE_MV ); mk_str = optarg; break;
			case 'o': set_outfile( optarg );break;
			case '8': outmode = RM_OUTPUT_8; break;
			case 'h': usage(); exit(0); break;
			default: usage(); exit(1); break;
		}
	}

	if( 0 != hl_mount_list( argc - optind, argv + optind ) ) {
		fprintf(stderr, "There was a problem loading the given hives: %s\n", uerr_str( uerr_get() ));
		return 1;
	}

	switch( mode ) {
		case RM_MODE_EXPORT:
			if( 0 != export_reg( regpath ) ) {
				fprintf(stderr, "doexportNEW() failed: %s\n", uerr_str( uerr_get() ) );
				return 1;
			}
			break;
		case RM_MODE_IMPORT:
			import_reg( import_str );
			break;

		case RM_MODE_DELETE:
			if( 0 !=  regmod_delete( del_path ) ) { 
				r = 1;
			}
			break;

		case RM_MODE_MK:
			DB(DB_VERB, "Creating key <%s>\n", mk_path);
			if( 0 != ns_mkdir( mk_path, 0 ) ) {
				DB(DB_ERR, "%s\n", uerr_str( uerr_get() ) );
				r = 1;
			}
			break;


		case RM_MODE_MV:
			r = make_value( mk_str );
			break;


		case -1:
			usage();
			exit(1);
			break;

		default:
			DIE("Unknown mode %d", mode );
	}

	hl_close_all();

	return r;

}
// }}}

int regmod_delete( char *del_path ) {
	int type = ns_gettype( del_path );

	DB(DB_VERB, "Deleting <%s>\n", del_path );
	switch( type ) {
		case RTYPE_KEY:
			if( 0 != ns_rmdir( del_path ) ) {
				DB(DB_ERR, "%s\n", uerr_str( uerr_get() ) );
				return -1;
			}
			break;
		case RTYPE_VALUE:
			if( 0 != ns_unlink( del_path ) ) {
				DB(DB_ERR, "%s\n", uerr_str( uerr_get() ) );
				return -1;
			}
			break;
		default:
			DIE("Unknown type: %d", type );
	}
} // }}}
int make_value( char *str ) // {{{
{
	VBUF_STATIC( vbpath, 10 );
	VBUF_STATIC( vbtype, 10 );
	VBUF_STATIC( vbval, 10 );
	VBUF_STATIC( vb, 10 );
	struct varbuf *bufs[3];
	int n=0;
	int type=-1;
	bufs[0] = vbpath;
	bufs[1] = vbtype;
	bufs[2] = vbval;

	vbuf_strset( bufs[0], "" );
	vbuf_strset( bufs[1], "" );
	vbuf_strset( bufs[2], "" );

	DB(DB_VERB, "Creating value <%s>\n", str );
	int is_esc, c;

	char *p = str;
	while( p && ( p = tok_esc_char2( p, &is_esc, &c ) ) ) {
		if( !is_esc && ('=' == c || '\0' == c ) )
			n++;
		else
			vbuf_charcat( bufs[n], c );

		if( '\0' == c ) break;
	}

	if( n < 3 ) {
		fprintf(stderr, "Malformed arugment.  I need \"val_path=type=value\", see help.\n");
		return -1;
	}

	if( 0 > ( type = str2valtype( vbtype->b ) ) ) {
		fprintf(stderr, "Could not get valid type from string %s\n", vbtype->b );
		return -1;

	}

	if( 0 != regp_udata( vb, type, vbval ) ) {
                fprintf(stderr, "parse_data(%s) failed: %s\n", vbval->b, uerr_str( uerr_get() ) );
                return;
        }

	//WARN("Not creting value from string %s -> %s, %s, %s\n", str, path->b, type->b, val->b );

	if( -1 == ns_set_valueT( vbpath->b, type, vb->b, vb->dlen ) ) {
                fprintf(stderr, "Couldnt' set value: %s\n", uerr_str( uerr_get() ) );
                return;
        }

	return -1;
		
	/*
	if( 0 != ns_mkdir( md_path ) ) {
		DB(DB_ERR, "%s\n", uerr_str( uerr_get() ) );
	}
	*/
}

// vim600: set foldlevel=0 foldmethod=marker:
