/* {{{ preamble
 * sam.c
 *
 * Command line access to Windows SAM
 *
 * Author: Sean Loaring
 *
 * Copyright (c) 2006 Sean Loaring
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 }}} */
// {{{ includes
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "../lib/libhive.h" //win32-style registry interface
#include "../hivetools/nstdreg.h" //my attempt at a posix-like interface to libhive
#include "../hivetools/sam.h" // sam-specific functions (set user password, etc)
#include "../misc/tools.h" //
#include "../misc/uerr.h" // like errno and strerror...kindof
// }}}

#define TMODE_USAGE 		0
#define TMODE_ULIST 		1
#define TMODE_ULIST_LONG 	2
#define TMODE_PASSWD		3
#define TMODE_PASSWD_NAMED	4
#define TMODE_ULIST_LONG_NAMED  5

char *pass_str = NULL;
char *user = NULL;

static int mode = TMODE_USAGE;

// {{{ forward defs
int usage(); 
int parse_args( int *argc, char ***argv ); 
int load_hives( int argc, char **argv );
int list_users();
int list_named_user_long( char *name ); 
int list_users_long();
int set_passwd( char *str ); 
int set_passwd_named( char *str );
// }}}
int main(int argc, char **argv ) // {{{
{
	int rid;
	//init
        unicode_init();
        libhive_init();
        ns_init();

	//operate
	if( 0 != parse_args( &argc, &argv ) ) { //sets mode
		fprintf(stderr, "Error parsing arguments: %s\n", uerr_str( uerr_get() ) );
		goto MAINERR;
	}

	if( 0 != load_hives( argc, argv ) ) { 
		fprintf(stderr, "Error loading hives: %s\n", uerr_str( uerr_get() ) );
		goto MAINERR;
	}

	switch( mode ) {
		case TMODE_ULIST:
			list_users();
			break;
		case TMODE_ULIST_LONG:
			list_users_long();
			break;
		case TMODE_PASSWD:
			if( 0 != set_passwd( pass_str ) ) return 1;
			break;
		case TMODE_PASSWD_NAMED:
			if( 0 != set_passwd_named( pass_str ) ) return 1;
			break;
		case TMODE_ULIST_LONG_NAMED:
			if( 0 != list_named_user_long( user ) ) return 1;
			break;
		case TMODE_USAGE:
			usage();
			break;
		default:
			DIE("Unknown mode.");
	}
	//shutdown

	return 0;

MAINERR:
	return 1;
} // }}}
int parse_args( int *argc, char ***argv ) // {{{ parse leading arguments
{
        extern int optind;
        extern char* optarg;
        char c;

        char *options = "hlLD:p:P:u:";

        while((c=getopt(*argc,*argv,options)) > 0) {
                switch(c) {
                        case 'h': mode = TMODE_USAGE;      break;
                        case 'p': mode = TMODE_PASSWD;pass_str=optarg; break;
                        case 'P': mode = TMODE_PASSWD_NAMED;pass_str=optarg; break;
                        case 'l': mode = TMODE_ULIST;      break;
                        case 'L': mode = TMODE_ULIST_LONG; break;
                        case 'u': mode = TMODE_ULIST_LONG_NAMED;user=optarg; break;
			case 'D':
			{
				int dbval;
				if( 1 != sscanf(optarg, "%d", &dbval ) ) {
					fprintf(stderr, "Couldn't set debug level to \"%s\", is it an integer?\n", optarg );
					exit(1);
				}
				DEBUG_LEVEL = dbval;
				break;
			}
                        default: 
				  fprintf(stderr, "Unrecognized option %c.  Run with -h for help.\n", c);
				  exit(1);
                }
        }

	*argc -= optind;
	*argv += optind;

	return 0;
} // }}}
int usage() // {{{
{
	fprintf(stderr,
		"Usage: sam [options] mount_list\n"
		"Read and write data from SAM (System Accounts Manager) hive file.\n"
		"\n"
		"options:\n"
                " -h           This message\n"
		" -l           List users\n"
                " -L           List users (long)\n"
		" -u user      List user (long)\n"
		" -p RID=pass  Set password for user with given RID\n"
		" -P USER=pass Set password for user\n"
                " -D n         Enable debug output level n\n\n"
        );

	fprintf(stderr, "%s", hl_mount_help() );

	/*
	fprintf(stderr, 
                "See readme file on how to extract/read/write the NT's SAM file\n"
                "if it's on an NTFS partition!\n");
		*/

	return 0;
} // }}}
int load_hives( int argc, char **argv ) // {{{ 
{
//        if( !argv[optind] ) {
//                fprintf(stderr, "No hive files specified.\n");
//                return 1;
//        }

	//HERE
        if( 0 != hl_mount_list( argc, argv  ) ) {
                //TODO: error
                fprintf(stderr, "There was a problem loading the given hives: %s\n", uerr_str( uerr_get() ) );
                return 1;
	}

	return 0;
} // }}}
int list_users() // {{{
{
	VBUF_STATIC( vb, 10 );
	VBUF_STATIC( nbuf, 10 );
        int *ridlist;
        int i;

        if( 0 != sam_user_list( vb ) ) {
		fprintf(stderr, "Error reading user RID list: %s\n", uerr_str( uerr_get() ) );
		return -1;
	}

        ridlist = (int*)vb->b;

        for( i=0; i<vb->dlen; i++ ) {
                if( 0 == sam_get_user_name( nbuf, ridlist[i] ) ) {
                        printf("%08x %s\n", ridlist[i], nbuf->b );
                }
                else {
                        fprintf(stderr, "Error reading data for RID %x: %s\n", ridlist[i], uerr_str( uerr_get() ) );
                }
        }

	return 0;
} // }}}
int list_named_user_long( char *name ) // {{{
{
	int rid;
	if( 0 > ( rid = sam_uname2rid( user ) ) ) {
		fprintf(stderr, "Error getting RID for user \"%s\":  %s\n", user, uerr_str( uerr_get() ) );
		return -1;
	}
	return list_user_long( rid );
} // }}}
int list_user_long( int rid ) // {{{
{
	VBUF_STATIC( vb, 10 );
	VBUF_STATIC( nbuf, 10 );
        int i;

	vbuf_strset( nbuf, "" );
	if( 0 != sam_print_user( nbuf, rid ) ) {
		fprintf(stderr, "Error printing user data for RID %x: %s\n", rid, uerr_str( uerr_get() ) );
		return -1;
	}

	printf( "%s", nbuf->b );

	return 0;
} // }}}
int list_users_long() // {{{
{
	VBUF_STATIC( vb, 10 );
	VBUF_STATIC( nbuf, 10 );
        int *ridlist;
        int i;

        if( 0 != sam_user_list( vb ) ) {
		fprintf(stderr, "Error reading user RID list: %s\n", uerr_str( uerr_get() ) );
		return -1;
	}

        ridlist = (int*)vb->b;

	printf("Excessivly verbose user list:\n");
        for( i=0; i<vb->dlen; i++ ) {
		list_user_long( ridlist[i] );
		printf("------------------------------------------------\n%s", nbuf->b );
        }

	return 0;
} // }}}
int set_passwd( char *str ) // {{{
{
	VBUF_STATIC (vb, 10);
	int rid;

	vbuf_resize( vb, strlen( str ) + 1 );

	sscanf( str, "%x=%s", &rid,  vb->b );
	if( 0 != sam_set_nt_password( rid, vb->b ) ) {
		fprintf(stderr, "Error setting NT password: %s\n", uerr_str( uerr_get() ) );
		return -1;
	}

	if( 0 != sam_set_lm_password( rid, vb->b ) ) {
		fprintf(stderr, "Error setting Lanman password: %s\n", uerr_str( uerr_get() ) );
		return -1;
	}
	return 0;
} // }}}
int set_passwd_named( char *str ) // {{{
{
	VBUF_STATIC (uname, 10);
	VBUF_STATIC (pass, 10);
	char *off;
	int rid;

	vbuf_resize( uname, strlen( str ) + 1 );
	vbuf_resize( pass, strlen( str ) + 1 );

	if( NULL == ( off = strchr( str, '=' ) ) ) {
		fprintf(stderr, "Expecting user=password, got \"%s\".\n", str );
		return -1;
	}
	vbuf_strnset( uname, str, off - str );
	vbuf_strset( pass, off+1 );
	/*
	printf("USER: .%s.\nPASS: .%s.\n", uname->b, pass->b );
	return 0;
	*/
	if( 0 > ( rid = sam_uname2rid( uname->b ) ) ) {
		fprintf(stderr, "Error getting RID for user \"%s\":  %s\n", uname->b, uerr_str( uerr_get() ) );
		return -1;
	}
	if( 0 != sam_set_nt_password( rid, pass->b ) ) {
		fprintf(stderr, "Error setting NT password: %s\n", uerr_str( uerr_get() ) );
		return -1;
	}

	if( 0 != sam_set_lm_password( rid, pass->b ) ) {
		fprintf(stderr, "Error setting Lanman password: %s\n", uerr_str( uerr_get() ) );
		return -1;
	}
	return 0;
} // }}}
// vim600: set foldlevel=0 foldmethod=marker:
