/* {{{ preamble
 * nstdreg.c
 *
 * NonStandardRegistry
 *
 * POSIX-like routines that provide access to the registry via rlReg* functions
 *
 * Authors:	Sean Loaring
 *
 * Copyright (c) 2005-2006 Sean Loaring
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * commit
 *
 }}} */

#include "../lib/libhive.h"
#include "nstdreg.h"
#include "../misc/tools.h"
#include "../misc/uerr.h"
#include "errno.h"
#include <string.h>
#include <stdio.h>
#include <libgen.h>

int ns_errno;

char *ns_dirname( char *p, char delim );
char *ns_basename( char *p, char delim );
char *pathjoin( char *buf, int blen, char *p1, char *p2 ); 
int ns_cd( char *path );        //attempt to change cwd, will try to work with relative paths, etc.

char *process_pathA_junk( char *path, char *buf, int blen );

void *vb_pathjoin( vbuf *vb, char *p1, char *p2 ); 

int path_end_is_delim( char *p ) // {{{ Determine if the last token in a path is a delimiter (escaped delimiters don't count)
{
	int len = strlen( p );
	int r = p[len-1] == '/' &&  !( len >= 2 && p[len-2] == '\\' );
	//WARN("path_end_is_delim( %s ): returns %d with len %d", p, r, len );
	return r;
} // }}}

char *hivenames[] = // {{{
	{ "HKEY_LOCAL_MACHINE",
	"HKEY_CURRENT_USER",
	"HKEY_USERS",
	"HKEY_CURRENT_CONFIG"
	 }; // }}}

NREGDIR *cwd = NULL;
rl_hkey *hkeys;
#define N_HIVES 4

#define PATH_RELATIVE 1
#define PATH_ABSOLUTE 2

//#define RTYPE_VALUE	0
//#define RTYPE_KEY	1

void ns_init() // {{{
{
	libhive_init();
	//TODO: free this at shutdown time
	hkeys = F_MALLOC( N_HIVES * sizeof( rl_hkey ) );

	hkeys[0] = rlHKEY_LOCAL_MACHINE;
	hkeys[1] = rlHKEY_CURRENT_USER;
	hkeys[2] = rlHKEY_USERS;
	hkeys[3] = rlHKEY_CURRENT_CONFIG;
} // }}}

int strappend( char *buf, int blen, int *off, char *str ) //{{{ append to a string && increment off || return !0
{
	//TODO: errno
	int od=0;
	if( blen - *off <= (od += snprintf( buf + *off, blen - *off, str ) ) ) {
		DIE("strappend() failed.  Check this for memory corruption.");
		return 0; 
	}
	*off+=od;

	return 1;
} // }}}

int ns_findhive( char *hivename ) // {{{
{
	int i;
	for( i=0; i<N_HIVES ; i++ ) {
		if( 0 == strcmp( hivename, hivenames[i] ) ) {
			return i;
		}
	}

	ERR_UNIX( ENOENT, "ns_findhive(): invalid or unknown hive name: %s", hivename ); 
	return -1;
} // }}}
int ns_gettype( char *pathin ) // {{{ 
{
	//rl_hkey h, h2;
	rl_hkey h2;
	int result = RTYPE_NOT_FOUND;
	VBUF_STATIC( vdn, 100 );
	VBUF_STATIC( vbn, 100 );
	VBUF_STATIC( regbn, 100 );
	VBUF_STATIC( vpathtmp, 100 );
	VBUF_STATIC( vpath, 100 );
	char *path;
	NREGDIR *d;

	VBUF_STATIC( cp, 100 );

	vb_pathjoin( cp, ns_getcwd(), pathin );

	//WARN("GETTYPE JOIN");
	path = cp->b;

	if( NULL != ( d = ns_opendir( path ) ) )  { //this should work if we have a RTYPE_KEY
		ns_closedir( d );

		return RTYPE_KEY;
	}

	//fprintf(stderr, "ns_gettype(%s) ns_opendir() returned NULL: %s\n", cp->b, uerr_str( uerr_get() ) );

	ns_vdirname( vdn, path );
	ns_vbasename( vbn, path );

	//path is either a value or it doesn't exist

	if( 0 != ns_path_ns2regvalname( regbn, vbn->b ) ) return -1;
	if( NULL == ( d = ns_opendir( vdn->b ) ) ) return -1;

	if( rlERROR_SUCCESS == rlRegQueryValueEx( ns_get_hkey( d ), regbn->b, NULL, NULL, NULL ) ) 
		result = RTYPE_VALUE;

	return result;

} // }}}

int ns_test( char *path, char *t ) // {{{ like test(1), but less functional, pass things like '-d' and '-e'
{
	VBUF_STATIC( vdn, 100 );
	VBUF_STATIC( vbn, 100 );
	//NREGDIR *dir;

	if( t[0] != '-' || t[1] == '\0' || '\0' != t[2] ) { // {{{ error
		ERR_UNIX(EINVAL, "ns_test(): I don't understand the test you gave me.");
		return -1;
	} // }}}

	int type = ns_gettype( path );

	//ns_closedir( dir );

	switch( t[1] ) {
		case 'e':
			if( RTYPE_NOT_FOUND != type ) return 1;
			return 0;
		case 'd':
			if( RTYPE_KEY == type || RTYPE_FAKEKEY == type ) return 1;
			return 0;
		case 'f':
			if( RTYPE_VALUE == type ) return 1;
			return 0;
		default:
			ERR_UNIX( EINVAL, "ns_test(): unknown test \"%c\".", t[1] );
			return 0; //TODO: is this right?
	}

	return 0;
} // }}}

int ns_rmdir( char *pathin ) // {{{ Delete a key
{
	char *path;
	VBUF_STATIC( vdn, 100 );
	VBUF_STATIC( vbn, 100 );
	VBUF_STATIC( vbregname, 100 );
	NREGDIR *dir;
	rl_hkey hk;
	int r=0;

	VBUF_STATIC( cp, 100 );
	vb_pathjoin( cp, ns_getcwd(), pathin );
	//WARN("rmdir JOIN");

	path = cp->b;

	ns_vdirname( vdn, path );
	ns_vbasename( vbn, path );
	if( NULL == ( dir = ns_opendir( vdn->b ) ) ) return -1;
	
	if( 0 != ns_path_ns2regkeypath( vbregname, vbn->b ) ) return -1;

	if( rlERROR_SUCCESS != rlRegDeleteKey( ns_get_hkey( dir ), vbregname->b ) )
		r = -1;
	
	ns_closedir( dir );
	return r;
} // }}}
int ns_unlink( char *pathin ) // {{{ Delete a value
{
	char *path;
	int r=0;
	VBUF_STATIC( vdn, 100 );
	VBUF_STATIC( vbn, 100 );
	VBUF_STATIC( vb_valname, 100 );
	NREGDIR *dir;
	rl_hkey hk;

	VBUF_STATIC( cp, 100 );
	vb_pathjoin( cp, ns_getcwd(), pathin );
	path = cp->b;

	ns_vbasename( vbn, path );
	ns_vdirname( vdn, path );

	if( 0 != ns_path_ns2regvalname( vb_valname, vbn->b ) ) return -1;
	if( NULL == ( dir = ns_opendir( vdn->b ) ) ) return -1;

	if( rlERROR_SUCCESS != rlRegDeleteValue( ns_get_hkey(dir), vb_valname->b )) {
		r = -1;
	}
	ns_closedir( dir );

	return r;
} // }}}
int ns_chdir( char *path ) // {{{ set current working directory
{
	NREGDIR *tcwd;
	//WARN("chdir()");
	if( NULL == ( tcwd = ns_opendir( path ) ) ) {
		return -1;
		//DIE("Couldn't open directory %s: %s.", path, uerr_str( uerr_get() ));
	}
	cwd = tcwd;
	return 0;
} // }}}
char *ns_getcwd() // {{{ returns the current working directory
{ // TODO: This returns stuff from the NSDIRENT struct, is this good?
	if( NULL != cwd )
		return ns_getpath( cwd );
	return "/";
} // }}}
void ns_dumpfulldir( int n, vbuf *vb ) // {{{
{
	struct ns_dirent *da;
	int i;
	//WARN("dedump");
	da = (struct ns_dirent*)vb->b;
	for( i=0; i < n; i++ ) {
		WARN("entry %d: %s", i, da[i].filename );
	}

} // }}}
int ns_recurse( char *base, char *rel, int (*f)(char *base, char *rel, long type, void *opt ), int flags, void *opt ) // {{{ calls f on all subordinate of hk.
{
	NREGDIR* dir = NULL;
	VBUF_STATIC( curpath, 10 );
	struct ns_dirent *d;
	int r=0;
	//vbuf *compath;
	vbuf *vbdirents;
	int nents=-1, i=0;
	struct ns_dirent *ents;
	char term[] = "";
	long type;

	if( rel == NULL ) rel = term;
	vbuf *vbrel;

	vb_pathjoin( curpath, base, rel );

	//WARN("Join: %s, %s -> %s.", base, rel, curpath->b );

	vbrel = vbuf_alloc(10);

	if( 0 > (type = ns_gettype( curpath->b )) )
		goto error;

	//compath = vbuf_alloc(10);
	vbdirents = vbuf_alloc( 10 );

	//WARN("recurse -> ns_opendir");
	if( NULL == ( dir = ns_opendir( curpath->b ) ) ) {
		//WARN("HereX: %s", curpath->b);
		goto error;
	}

	if( ! (NSR_FLAG_PARENT_LAST & flags) ) {
		if( 0 != (*f)( base, rel, type, opt ) ) 
			goto error;
	}

	if( 0 > ( nents =  ns_readfulldir( vbdirents, dir ) ) ) goto error;

	ents = (struct ns_dirent*)vbdirents->b;

	for( i=0; i< nents; i++ ) {

		d = &ents[i];

		vb_pathjoin( vbrel, rel, d->filename );

		//WARN("Join2: %s, %s -> %s.", rel, d->filename, vbrel->b );

		//vb_pathjoin( compath, path, d->filename );

		if( d->type == RTYPE_KEY || d->type == RTYPE_FAKEKEY ) {
			int r;

			if( 0 != ( r = ns_recurse( base, vbrel->b, f, flags, opt ) ) )  {
				if( flags & NSR_FLAG_DITCH_ON_FAIL ) {
					goto error;
				}
				else
					WARN("recurse into \"%s\" from \"%s\" failed: %s", vbrel->b, base, uerr_str(uerr_get()) );
			}
			/*
			if( 0 != ( r = ns_recurse( base, compath->b, f, flags, opt ) ) )  {
				if( flags & NSR_FLAG_DITCH_ON_FAIL ) {
					goto error;
				}
				else
					WARN("recurse into \"%s\" failed: %s", compath->b, uerr_str(uerr_get()) );
			}
			*/

		} else {
			if( 0 != (*f)(base, vbrel->b, d->type, opt) ) {
				//TODO: set error!
				WARN("ns_recurse() *f call: %s + %s", base, vbrel->b);
				goto error;
			}
			/*
			if( 0 != (*f)(compath->b, d->type, opt) ) {
				//TODO: set error!
				WARN("Uh-oh: %s", compath->b);
				goto error;
			}
			*/
		}
	}
	
	if( NSR_FLAG_PARENT_LAST & flags ) {
		vb_pathjoin( curpath, base, rel );

		//WARN("Join3: %s, %s -> %s.", rel, d->filename, vbrel->b );
		if( 0 != (*f)(base, rel, ns_gettype( curpath->b ), opt) ) {
			//TODO: set error!
			WARN("Uh-oh: %s, %s", curpath->b, uerr_str( uerr_get() ));
			goto error;
			//return -1;
		}
	}

	goto done;
error:
	r = -1;
done:
	if( dir ) ns_closedir( dir );
	//vbuf_free( compath );
	vbuf_free( vbdirents );
	vbuf_free( vbrel );
	return r;
} // }}}
int ns_path_element_nreg2ns( vbuf *dest, int n, char *src ) // {{{
{
	int i;

	vbuf_strset( dest, "" );

	// the special null case
	if( src[0] == '\0' ) {
		vbuf_strset( dest, "\\0" );
		return 0;
	}

	// The 'dots only' case
	int doprefix=1;
	for( i=0; i<n; i++ ) 
		if( src[i] != '.' ) doprefix = 0;
	if( doprefix ) vbuf_strncat( dest, "..", 3 );
	
	//everything else
	for( i=0; i<n; i++ ) {
		switch(src[i]) {
			case '\\':
				vbuf_strncat( dest, "\\\\", 3 );
				break;
			case '/':
				vbuf_strncat( dest, "\\", 3 );
				vbuf_strncat( dest, &src[i], 1 );
				break;
			case '\0':
				break;

			default:
				vbuf_strncat( dest, &src[i], 1 );
		}
	}

	return 0;
} // }}}

// ns_opendir(), ns_readdir(), etc {{{
int ns_readfulldir( vbuf *vb, NREGDIR *dir ) // {{{ Reads ns_dirents into an array
{
	int i=0;
	struct ns_dirent *d;
	struct ns_dirent *da;
	vbuf_clear( vb );
	PRO_START;
	//WARN("entering...");
	while( d = ns_readdir( dir ) ) {
		//WARN("Got %s", d->filename );
		PRO_START;
		vbuf_append( vb, d, sizeof( struct ns_dirent ) );
		PRO_END("ns_readfulldir.append");
		//ns_dumpfulldir( i, vb );
		i++;
	}

	PRO_END("ns_readfulldir");

	return i;
} // }}}
NREGDIR *NREGDIR_ALLOC( char *path, rl_hkey hk ) // {{{
{
	NREGDIR *result = F_MALLOC( sizeof( struct ns_handle ) );
	result->hk = hk;
	result->vb_path = vbuf_alloc( strlen( path ) + 1 );

	vbuf_strset( result->vb_path, path );

	ns_rewinddir( result );

	return result;
} // }}}
void NREGDIR_FREE( NREGDIR *dir ) // {{{
{
	vbuf_free( dir->vb_path );
	F_FREE( dir );
} // }}}
NREGDIR *ns_opendir( char *pathin ) // {{{ open an ns handle
{
	//char tok[PATH_MAX];
	rl_hkey hk = NULL;
	char rlpath[PATH_MAX];
	char *p;
	char *path;
	int ttype;
	int hiveno;
	int off;
	VBUF_STATIC( vb_path, 100 );
	VBUF_STATIC( vb_rpath, 100 );
	VBUF_STATIC( vb_tok, 100 );

	VBUF_STATIC( cp, 100 );

	//WARN("ns_opendir() called on %s in %s", pathin, ns_getcwd() );

	vb_pathjoin( cp, ns_getcwd(), pathin );

	if( 0 != vb_process_path( vb_path, cp->b ) ) return NULL;

	if( !path_end_is_delim( vb_path->b ) ) vbuf_charcat( vb_path, '/' );

	path = vb_path->b;
	p = path;

	//  ensure absolute & eat leading '/' from *p
	if( TOK_DELIM != vb_path_token( vb_tok, &p ) ) { 
		ERR_UNIX(EINVAL, "ns_opendir(): path is not absolute.", path);
		return NULL;
	} // 


	ttype = vb_path_token( vb_tok, &p ); //gettoken( tok, PATH_MAX, &p, '/' ); //TODO: do I need to set anything in result before returning root?
	if( TOK_EMPTY == ttype ) { //allocate some fakeness if the path is just '/'
		return NREGDIR_ALLOC(vb_path->b, NULL);
	}

	//The first element after '/' should be the name of a hive
	//  try to load given hive

	if( TOK_ELEMENT != ttype ) { 
		ERR_UNIX( EINVAL, "ns_opendir(): In path \"%s\" I was expecting something like HKEY_LOCAL_MACHINE, got something else.", path);
		return NULL;
	} 

	hiveno = ns_findhive( vb_tok->b );

	if( hiveno < 0 ) return NULL;

	if( 0 != ns_path_ns2regkeypath( vb_rpath, p ) ) return NULL;

	if( rlERROR_SUCCESS != rlRegOpenKeyEx( hkeys[hiveno], vb_rpath->b, 0, KEY_ALL_ACCESS, &hk ) ) { 

		//WARN("Fail: %s -> %s\n", pathin, vb_rpath->b);
		return NULL;
	}  

	//rl_hkey_dump( hk );
	return NREGDIR_ALLOC( vb_path->b, hk );
} // }}}
void ns_closedir( NREGDIR *h ) // {{{ close a ns handle
{
	if( h->hk ) rlRegCloseKey( h->hk );

	NREGDIR_FREE( h );
} // }}}
struct ns_dirent *ns_readdir( NREGDIR *h ) // {{{
{
	VBUF_STATIC( vbfname, 100 );
	static struct ns_dirent de;
	PRO_START;
	if( h->hk ) {
		//WARN("Entering ns_readdir() w/ %p, %p %d", h, h->hk, h->hk->type);
	}
	if( 0 == strcmp( h->vb_path->b, "/" ) ) { //1 == strlen(h->path) && h->path[0] == '/' ) {
		if( h->rdidx >= N_HIVES ) return NULL;
		strncpy( de.filename, hivenames[ h->rdidx ], PATH_MAX );
		de.type = RTYPE_FAKEKEY;
		h->rdidx++;
	}
	else {
		ulong tlen = PATH_MAX;
		if( 0 > h->rd_donekeys ) {
			PRO_START;
			long cr = rlRegEnumKeyEx( h->hk, h->rdidx, de.filename, &tlen, NULL, NULL, NULL );
			de.type = RTYPE_KEY;

			switch( cr ) {
				case rlERROR_SUCCESS:
					ASSERT( tlen < PATH_MAX, "Buffer issues." );
					h->rdidx++;
					break;
				case rlERROR_NO_MORE_ITEMS:
					h->rd_donekeys = h->rdidx;
					break;
				default:
					DIE("rlRegEnumKeyEx() returned something odd.");
			}

			PRO_END( "ns_readdir.keys");
		}
		if( 0 <= h->rd_donekeys ) {
			int cr = rlRegEnumValue( h->hk, h->rdidx - h->rd_donekeys, de.filename, &tlen, NULL, NULL, NULL );
			de.type = RTYPE_VALUE;

			PRO_START;

			switch( cr ) {
				case rlERROR_SUCCESS:
					ASSERT( tlen < PATH_MAX, "Buffer issues." );
					h->rdidx++;
					break;
				case rlERROR_NO_MORE_ITEMS:
					return NULL;
					break;
				default:
					DIE("rlRegEnumKeyEx() returned something odd.");
			}

			PRO_END( "ns_readdir.vals");
		}

		ns_path_element_reg2ns( vbfname, de.filename );

		switch( de.type ) {
			case RTYPE_VALUE:
				break;
			case RTYPE_KEY:
				vbuf_charcat( vbfname, '/' );
				break;
			default:
				DIE("I am confused.");
		}

		ASSERT( vbfname->dlen < PATH_MAX, "vbfname->dlen > MAX_PATH, sorry.");
		memcpy( de.filename, vbfname->b, vbfname->dlen );
	}

	PRO_END("ns_readdir.full");

	return &de;
} // }}}
void ns_rewinddir( NREGDIR *h ) // {{{
{
	h->rdidx = 0;
	h->rd_donekeys = -1;
} // }}}
// }}}

int ns_set_sz( char *path, char *str ) // {{{ in: path,str=utf8 string; out: str->utf16->registry
{
	VBUF_STATIC( vb, 100 );
	int nl = ( strlen( str ) + 1 )* 2;
	if( 0 != vb_utf8to16( vb, str, strlen( str ) + 1 ) ) return -1;

	return ns_set_valueT( path, REG_SZ, vb->b, vb->dlen );
} // }}}
int ns_set_valueT( char *pathin, int type, char *value, int len ) // {{{ in: raw value data; out: ->registry
{
	VBUF_STATIC( vpath, 100 );
	VBUF_STATIC( vdn, 100 );
	VBUF_STATIC( vbn, 100 );
	VBUF_STATIC( vb, 100 );
	VBUF_STATIC( regbn, 100 );
	NREGDIR *dir = NULL;
	rl_hkey hk;
	char *ds = value;
	int *dlen = &len;
	char *path;
	int r=0;

	VBUF_STATIC( pp, 100 );

	//WARN("Eek.");
	vb_pathjoin( pp, ns_getcwd(), pathin );

	path = pp->b;

	ns_vdirname( vdn, path );
	ns_vbasename( vbn, path );
	if( NULL == ( dir = ns_opendir( vdn->b ) ) ) goto err;

	if( NULL == ( hk = ns_get_hkey( dir ) ) ) {
		ERR_UNIX( EINVAL, "ns_set_value(): You can not set a value within this key because it is not real.");
		goto err;
	}

	if( 0 != ns_path_ns2regvalname( regbn, vbn->b ) ) goto err;
	if( rlERROR_SUCCESS != rlRegSetValueEx( hk, regbn->b, 0, type, ds, *dlen) ) goto err;

	goto done;
err:
	r = -1;
done:
	if( NULL != dir ) ns_closedir( dir );
	return r;
} // }}}
void ns_pairpath( vbuf *dn, vbuf *bn, char *path ) // {{{ cleans path, sets dn to dirname and bn to basename
{
	VBUF_STATIC( vpath, 100 );

	vb_process_path( vpath, path );
	ns_vdirname( dn, path );
	ns_vbasename( bn, path );
} // }}}
int ns_read_valueT( char *path, unsigned long *type, struct varbuf *vb ) // {{{
{
	char *dn = NULL;
	char *bn = NULL;
	NREGDIR *dir = NULL;
	NREGH* rh;
	rl_hkey hk;
	unsigned long needed;
	VBUF_STATIC( vbn, 100 );
	VBUF_STATIC( vdn, 100 );
	VBUF_STATIC( regbn, 100 );

	ns_pairpath( vdn, vbn, path );

	//WARN("What be this: %s, %s, %s", path, vbn->b, vdn->b );
	if( NULL == ( dir = ns_opendir( vdn->b ) ) ) goto broken;

	if( NULL == ( hk = ns_get_hkey( dir ) ) ) goto broken;

	needed=0;
	if( 0 != ns_path_ns2regvalname( regbn, vbn->b ) ) return -1;

	//WARN("How much do we need?" );
        if( rlERROR_SUCCESS != rlRegQueryValueEx( hk, regbn->b, type, NULL, &needed ) ) goto broken;
	if( !vb ) return rlERROR_SUCCESS;
	//WARN("Needed: %d", needed );
        vbuf_resize( vb, needed );

	//WARN("Calling rlRegQueryValueEx w/: %d", needed );
        if( rlERROR_SUCCESS != rlRegQueryValueEx( hk, regbn->b, type, vb->b, &needed ) ) goto broken; 
	vb->dlen = needed;

	ns_closedir( dir );

	return 0;
broken:
	{
		WARN("Yuck, broken: %d", needed );
	struct UERR *myuerr = uerr_get();
	if( myuerr->errclass == UERR_WIN && myuerr->errnumber == rlERROR_NOT_FOUND )
		ERR_UNIX( ENOENT, "%s: %s", path, uerr_str( myuerr ));
	if( NULL != dir ) ns_closedir( dir );
	return -1;
	}

} // }}}
int ns_mkdir_p( char *path ) // {{{  Create (if they don't exist) all of the registry keys in a path (like mkdir -p)
{
	int ttype;
	VBUF_STATIC( tok, 1 );
	VBUF_STATIC( bpath, 100 );
	char *p = path + 1;

	vbuf_strset( bpath, "" );

	while( TOK_EMPTY != (ttype = vb_path_token( tok, &p )) ){
		switch( ttype ) {
			case TOK_ERROR:
				return -1;
			case TOK_ELEMENT:

				vbuf_charcat( bpath, '/' );
				vbuf_strcat( bpath, tok->b );
				vbuf_charcat( bpath, '/' );

				if( !ns_test( bpath->b, "-d") ) {
					if( 0 != ns_mkdir( bpath->b, 0 ) )
						return -1;
				}

				continue;
			case TOK_DELIM:
				continue;
			case TOK_PARENT:
			case TOK_CURRENT:
				DIE("Got TOK_PARENT or TOK_CURRENT, shouldn't be possible. %d", ttype);
			default:
				DIE("Unknown token.");
		}
	}

	return 0;
} // }}}
int ns_mkdir( char *pathin, int flags ) { // {{{ Create a new registry key 
	char *path;
	long werr;
	NREGDIR *dir;
	rl_hkey hk;
	rl_hkey ret;
	VBUF_STATIC( dn, 100 );
	VBUF_STATIC( bn, 100 );
	VBUF_STATIC( regn, 100 );

	VBUF_STATIC( cp, 100 );

	vb_pathjoin( cp, ns_getcwd(), pathin );
	path = cp->b;

	if( *path != '/' ) { 
		ERR_UNIX( EINVAL, "ns_mkdir(): I only work with absolute paths, got %s.", path );
		return -1;
	} 

	if( MKDIR_P & flags ) {
		return ns_mkdir_p( path );
	}

	ns_vdirname( dn, path );
	ns_vbasename( bn, path );
	if( 0 != ns_path_ns2regvalname( regn, bn->b ) ) return -1;

	if( NULL == ( dir = ns_opendir( dn->b ) ) ) return -1;
	if( NULL == ( hk = ns_get_hkey( dir ) ) ) return -1;
	if( rlERROR_SUCCESS != ( werr = rlRegCreateKeyEx( hk, regn->b, NULL, 0, 0, NULL, &ret, NULL ) ) ) { 
		ns_closedir( dir );
		return -1;
	} 

	rlRegCloseKey( ret );

	ns_closedir( dir );

	return 0;
} // }}}

// Path Manipulation {{{
int path_is_absolute( char *p ) // {{{
{
	if( p[0] == '/' ) return 1;
	return 0;
} // }}}
void *vb_pathjoin( vbuf *vb, char *p1, char *p2 ) // {{{ attach p2 relative to p1
{
	VBUF_STATIC( cp, 100 );

	char *result;

	if( path_is_absolute( p2 ) ) {
		vbuf_strset( cp, p2 );
	}else {

		if( strlen( p1 ) > 0 ) {
			vbuf_strset( cp, p1 );
			if( ! path_end_is_delim( p1 ) ) {
				vbuf_strcat( cp, "/" );
			}

			vbuf_strcat( cp, p2 );

		}
		else {
			vbuf_strset( cp, p2 );
		}

	}

	//WARN("Jeez: %s, %s -> %s", p1, p2, cp->b);

	vb_process_path( vb, cp->b );



} // }}}
int vb_process_path( vbuf *vb, char *path ) // {{{ "/foo/.././dogs" -> "/dogs"
{
	//TODO: this function is dumpy
	char *p = path;
	int ttype;
	VBUF_STATIC(tok,100);
	int off = 0;
	int loff = -1;
	int last_cur=0;

	//WARN("vb_process_path(): called on <%s>\n", path );

	//WARN("aQ zero2 d: %d %p %d", 0, hkeys[0], hkeys[0]->type );
	vbuf_strset( vb, "" );

	while( TOK_EMPTY != ( ttype = vb_path_token( tok, &p ) ) ) {
		//WARN("TOK: %s", tok->b );
		last_cur=0;
		switch( ttype ) {
			case TOK_ERROR:
				ERR_UNIX( EINVAL,
					"vb_process_path(): Error parsing path %s around %s\n",
					path, p );
				return -1;

			case TOK_DELIM:
				//pp->type = PATH_ABSOLUTE;
				//if( path_is_absolute( path ) ) {
				//WARN("DELIM: %s", p );
				if( '\0' == vb->b[0] ) loff = -1;
				if( '/' != vs_last( vb ) )
					vbuf_charcat( vb, '/' );
				while( '/' == *p ) p++;
				//}
				break;

			case TOK_PARENT:
			{
				// ""
				// "foo"  	-> ""
				// "/foo" 	-> "/"
				// "/foo/"	-> "/"
				// "/../"	-> "/../../"
				// ".."		-> "/../../"

				//loff should point to the start of the last appended element
				//printf("PP: %s\n", vb->b );
				if( loff > 0 )
					vbuf_strtrunc( vb, loff );
				else {
					if( !path_is_absolute( vb->b ) ) 
						vbuf_strcat( vb, ".." );
				}

				//printf("PoP: %s\n", vb->b );

				break;
			}

			case TOK_CURRENT:
				//ignore all '.' but one at the start
				if( 1 == vb->dlen ) {
					vbuf_strset( vb, "." );
				}
				else {
					last_cur=1;
				}
			break;

			case TOK_ELEMENT:

				// We want to separate elements with delimiters, but we don't want a leading delimiter (it would have been provided by TOK_DELIM)
				if( vb->dlen > 1 && ! path_end_is_delim( vb->b ) ) vbuf_strcat( vb, "/" );
				/*
				else 
					WARN("Element returned but no strcat: %s %d %d", tok->b, path_end_is_delim( vb->b ), loff  );
					*/

				loff = vb->dlen - 1;

				vbuf_strcat( vb, tok->b );

				break;

			default:
				DIE("process_path(): Got unknown token type.");
		}

		//WARN("postTOK: %s %s:%d (%d), vb: %s", p, tok->b, ttype, TOK_DELIM, vb->b );
	}

	//WARN("Q2 zero2 d: %d %p %d", 0, hkeys[0], hkeys[0]->type );
	//I am removing this without knowing if this is correct
	//if( last_cur ) vbuf_charcat( vb, '.' );

	//WARN("Pre-slice: %s\n", vb->b);
	//TODO: this is gross and is it even correct?
	//while( vb->dlen > 2 && '/' == vb->b[vb->dlen-2] ) vbuf_strtrunc( vb, vb->dlen-2 );

	//TODO: This could be just a syntax error, it shouldn't kill the program
	ASSERT( vb->b[vb->dlen-1] != '\\', "Final character is an escape (which is illegal).");

	//WARN("vb_process_path(): returns %s\n", vb->b);
	return 0;
} // }}}
int ns_path_element_reg2ns( vbuf *dest, char *src ) // {{{
{
	return ns_path_element_nreg2ns( dest, strlen(src), src );
} // }}}
int ns_path_regkey2ns( vbuf *dest, char *str ) // {{{
{
	VBUF_STATIC( tmp, 100 );

	vbuf_strset( dest, "/" );

	char *this = str;
	if( '\\' == *this ) {
		ERR_UNIX( EINVAL, "Registry key path (%s) starts with \\, which is invalid.");
		return -1;
	}
	char *next;

	do {
		int n;
		next = index( this, '\\' );
		n = (!next)?strlen( this ):next - this;
		if( 0 != ns_path_element_nreg2ns( tmp, n, this ) )
			return -1;
		vbuf_strcat( dest, tmp->b );

		if( next ) {
			vbuf_charcat( dest, '/' );
			this = next+1;
		}
		else
			this = NULL;
	} while(this);

	return 0;
} // }}}
int ns_path_ns2regkeypath( vbuf *dest, char *src ) // {{{ in: ns dirname; out: reg key path
{
	return ns_path_ns2regt( dest, src, RTYPE_KEY );
} // }}}
int ns_path_ns2regvalname( vbuf *dest, char *src ) // {{{ in: ns basename; out: reg value name
{
	int r = ns_path_ns2regt( dest, src, RTYPE_VALUE );

	//WARN("Path conv: %s -> %s.\n", src, dest->b);
	return r;
} // }}}
int ns_path_ns2regt( vbuf *dest, char *src, int t ) // {{{
{
	//TODO: holy gross function, batman
	char *p, *lastp;
	int all_period=1;
	int c;
	int is_esc;

	vbuf_strset( dest, "" );

	if( '\0' == src[0] ) return 0;

	if( t == RTYPE_KEY )
		while( *src == '/' ) src++;


	p = src;
	while( *p ) {
		lastp = p;
		p = tok_esc_char( p, &is_esc, &c );
		if(is_esc && '0' == c ) {
			switch( t ) {
				case RTYPE_KEY:
					ERR_UNIX( EINVAL, "Path contains excaped zero, illegal in key names.");
					return -1;

				case RTYPE_VALUE:
					c = '\0';
					break;
				default:
					return -1;
					DIE("Unknown key type %d", t );
			}
		}
		switch( c ) {
			case '\0':
				if(p=src)all_period=0;
				while(*p)p++;
				break;
			case '/':
				all_period=0;
				if( is_esc ) {
					vbuf_charcat(dest, '/');
				}
				else {
					while( *p == '/' ) p++;
					if( '\0' != *p )
						vbuf_strncat(dest, "\\", 1);
				}
				break;
			case '.':
				vbuf_charcat(dest, c );
				break;
				
			case '\\':
				if( is_esc ) {
					switch( t ) {
						case RTYPE_KEY:
							ERR_UNIX( EINVAL, "Path contains escape, this is illegal in key names.");
							WARN("Wonk! %s, %s", p, src);
							return -1;
						default:
							DIE("Unknown type.");
							break;

						case RTYPE_VALUE:
							all_period=0;
							break;

					}
				}
			default:
				all_period=0;
				vbuf_charcat(dest, c );
				break;
		}
	}

	if( 0 == strlen( src ) ) all_period = 0;

	//the 'dot only' case
	if( all_period ) {
		if( strlen( src ) < 3 ) {
			WARN("WTF: >%s<", src );
			ERR_UNIX( EINVAL, "ns_path_ns2reg(): path contains '.' or '..'");
			return -1;
		}
		vbuf_skip(dest, 2);
	}

	return 0;
} // }}}
int rfinddelim( char *p, char delim ) // {{{ Seek last unescaped delimiter unless using escapes as delimters :P
{
	int i;
	int stl = strlen( p );
	for( i=stl; i>=0; i-- ) {
		if( p[i] == '\\' && delim != '\\' ) { //we skip escaped stuff unless we are using a retarded delimiter
			//TODO: what exactly should we do in the retarded
			//delimiter case?  What if someone is is passing stuff
			//like "\\blah\\blah"?
			p++;
			if( p[i] == '\0' ) return -1;
			p++;
		}
		else if( p[i] == delim ) {
			return i;
		}
	}

	return -1;
} // }}}
#if 0
int ns_vdirnamer( vbuf *vb, char *path ) // {{{ Another dirname() function
{
	int e;
	int t;
	int r=0;
	char *p = tok_esc_char( path, &e, &t );
	if( !e && t == '/' ) { // root directory is prepended '/'
		vbuf_strset( vb, "/" );
		while( *p == '/' ) p++;
		r+=1;
		return r + ns_vdirnamer( vb, p );
	}

	if( p = esc_index( path, '/' ) ) { // if we are not the last element then we append
		char *p2 = p;
		while( *p2 == '/' )p2++;
		if(!*p2) return 0;

		if( vb->dlen > 1 && vb->b[vb->dlen-2] != '/' )
			vbuf_charcat( vb, '/' );
		vbuf_strncat( vb, path, p - path );
		return ns_vdirnamer( vb, p2 );
	}

	return 0;
	//if( !*p ) return;

} // }}}
void ns_vdirname( vbuf *vb, char *path ) // {{{ Another dirname() function
{
	vbuf_strset( vb, "" );
	ns_vdirnamer( vb, path );
	if( 1 == vb->dlen ) vbuf_strset( vb, "." );
	//WARN("ns_vdirname %s -> %s", path, vb->b );
} // }}}
#endif
#if 0
int ns_vbasenamer( vbuf *vb, char *p, char **start ) // {{{ Another basename() function
{ //TODO: this is probably junk now
	int is_esc, c;
	ASSERT( p>= vb->b && p- vb->b < vb->dlen, "Stray pointer.");
	if( 0 == strlen( p ) ) return 0;

	p = tok_esc_char( p, &is_esc, &c );
	if( !is_esc && '/' == c ) {
		int r = ns_vbasenamer( vb, p, start );
		if( 0 == r ) {
			if( p == vb->b ) { // '^////' -> '^/'
				vbuf_strtrunc( vb, 1);
				return 1;
			}
			else { // '///$' -> ''
				vbuf_strtrunc( vb, p - 1 - vb->b );
				return 0;
			}
		}
		else { // 'foo/...' -> '...' 
			ASSERT( p >= vb->b && p - vb->b < vb->dlen, "Stray pointer.");
			//printf("Setting start: %p/%s ->", *start, *start );
			if( p > *start ) *start = p;
			//printf("%p/%s\n", *start, *start );
			return ns_vbasenamer( vb, p, start ); 
		}
	}
	else {
		return ns_vbasenamer( vb, p, start ) + 1;
	}
} // }}}
#endif
void ns_vdirname( vbuf *vb, char *path ) // {{{ similar to dirname(3)
{
	int type;
	char *p = path;
	VBUF_STATIC( tok, 10 );
	VBUF_STATIC( lastelem, 10 );

	vbuf_strset( lastelem, "" );
	vbuf_strset( vb, "" );
	int lastdelim = 0;
	int rcount = 0;

	//WARN("In this function.");

	while( TOK_EMPTY != (type = vb_path_token( tok, &p ) ) ) {
		//WARN("APPEND: %s", lasttok->b );
		rcount++;
		switch( type ) {
			case TOK_ELEMENT:
			case TOK_PARENT:
			case TOK_CURRENT:
				vbuf_strcat( vb, lastelem->b );
				if( 1 == lastdelim )
					vbuf_strset( lastelem, "/" );
				else
					vbuf_strset( lastelem, "" );
				vbuf_strcat( lastelem, tok->b );
				lastdelim=0;
				break;
			case TOK_DELIM:
				//WARN("TOAPPEND: %s", tok->b );
				if( 1 == rcount )
					vbuf_strset( vb, "/");
				else
					lastdelim = 1;
				break;
			case TOK_ERROR:
				//TODO: perhaps this function should return an int for failure?
				DIE("I can't deal with errors.");

			default:
				DIE("Unknown token type.");
		}
	}

	if( 0 == strcmp( vb->b, "" ) ) {
		vbuf_strset( vb, "." );
	}

} // }}}
void ns_vbasename( vbuf *vb, char *path ) // {{{ similar to basename(3)
{
	VBUF_STATIC( tok, 100 );
	int type;
	char *result = NULL;
	char *p = path;
	int c, is_esc;
	int is_set = 0;

	vbuf_strset( vb, "" );

	while( TOK_EMPTY != (type = vb_path_token( tok, &p ) ) ) {
		switch( type ) {
			case TOK_DELIM:
				if(!is_set) {
					is_set=1;
					vbuf_strset(vb, "/");
				}
				break;

			case TOK_ELEMENT:
			case TOK_PARENT:
			case TOK_CURRENT:
				is_set = 1;
				vbuf_overwrite( vb, tok );
				break;
			case TOK_ERROR:
				//TODO: perhaps this function should return an int for failure?
				DIE("I can't deal with errors.");

			default:
				DIE("Unknown token type.");
		}
	}

} // }}}
// }}}
NREGH* ns_open( char *pathin ) // {{{ open an ns handle
{
	NREGDIR *d;
	char *path;
	rl_hkey h;
	VBUF_STATIC( vdn, 100 );
	VBUF_STATIC( vbn, 100 );
	VBUF_STATIC( regbn, 100 );
	NREGH *result;

	//WARN("ns_open() starting with %s\n", path );

	VBUF_STATIC( cp, 100 );
	vb_pathjoin( cp, ns_getcwd(), pathin );
	path = cp->b;

	ns_vdirname( vdn, path );
	ns_vbasename( vbn, path );

	//WARN("HEREx");
	if( NULL == ( d = ns_opendir( vdn->b ) ) ) return NULL;
	if( NULL == ( h = ns_get_hkey( d ) ) ) return NULL;

	result = F_MALLOC( sizeof( NREGH ) );
	result->blen = 0;

	if( 0 != ns_path_ns2regvalname( regbn, vbn->b ) ) return NULL;

	long r = rlRegQueryValueEx( h, regbn->b, &result->type, NULL, &result->blen );
	if( rlERROR_MORE_DATA != r && rlERROR_SUCCESS != r ) {
		return NULL;
	}

	result->buf = F_MALLOC( result->blen );

	if( rlERROR_SUCCESS != rlRegQueryValueEx( h, regbn->b, &result->type, result->buf, &result->blen ) ) {
		DIE("This is broken");
		free( result->buf );
		free( result );
		return NULL;
	}

	return result;
} // }}}
char *ns_getpath( NREGDIR *h ) // {{{
{
	return h->vb_path->b;
} // }}}
void ns_close( NREGH *h ) // {{{ close a n
{
	free( h->buf );
	free( h );
} // }}}

rl_hkey ns_get_hkey( NREGDIR *h )  // {{{ return the hkey associated with cwd (if one exists)
{
	if( !h->hk ) {
		ERR_UNIX( EINVAL, "Attempt to retrieve hkey from imaginary NREGDIR.");
		DIE("I am going to crash here because this should probably not happen (TODO: this happens if no hive is loaded).");
	}
	return h->hk;
} // }}}
// Registry Export {{{
struct exp { // {{{
	ns_export_callback callback;
	char *cr;

}; // }}}
void export_keypath( char *s, char *cr, ns_export_callback callback ) // {{{
{
        VBUF_STATIC( rpath, 10 );
        VBUF_STATIC( utf8buf, 10 );

        ns_path_ns2regkeypath( rpath, s );

        if( 0 == rpath->dlen - 1 ) return;

        vbuf_printf( utf8buf, "%s[%s]%s", cr, rpath->b, cr );
        //write_output( utf8buf, outfd );

	callback( utf8buf->b, utf8buf->dlen );
} // }}}
int export_value( char *path, ns_export_callback callback ) // {{{
{
        VBUF_STATIC( valbuf, 10 );

        if( 0 != ns_regp_val( valbuf, path ) ) {
                WARN( "Couldn't export value \"%s\": %s", path, uerr_str( uerr_get() ) );
                return -1;
        }

        //write_output( valbuf, outfd );

	callback( valbuf->b, valbuf->dlen );
        return 0;
} // }}}
int export_item(char *base, char *rel, long type, void *ine ) // {{{
{
	VBUF_STATIC( path, 10 );

	//WARN("EXPORT JOIN");

	if( 0 != vb_pathjoin( path, base, rel ) ) return -1;

	//WARN("EXPORT JOIN");

	struct exp *e = ine;
        switch( type ) {
                case RTYPE_KEY:
			export_keypath( path->b, e->cr, e->callback );
                        break;
                case RTYPE_VALUE:
                        export_value( path->b, e->callback );
                        break;
                default:
                        DIE("Unknown type: %d", type );
        }
        return 0;
} // }}}
int ns_reg_export( char *path, char *cr, ns_export_callback callback ) // {{{
{
	struct exp e;
	VBUF_STATIC( utf16buf, 10 );
        VBUF_STATIC( utf8buf, 10 );

        vbuf_printf( utf8buf, "Windows Registry Editor Version 5.00%s", cr);

        callback( utf8buf->b, utf8buf->dlen );

	e.callback = callback;
	e.cr = cr;
        return ns_recurse( path, NULL, &export_item, 0, &e );
} // }}}
// }}}
int ns_reg_merge( char *path ) // {{{ Attempt to merge reg into the registry
{
	int r=0;
        struct regfile *rf;
	VBUF_STATIC( name, 100 );
	VBUF_STATIC( val, 100 );
	VBUF_STATIC( cwd, 100 );
	VBUF_STATIC( keypath, 100 );
	VBUF_STATIC( prpath, 100 );
	VBUF_STATIC( cname, 100 );
	VBUF_STATIC( dname, 100 );
        int type;
        int ttok;
	static int ditchcnt=0;

	name->dlen = 0;
	val->dlen = 0;
	cwd->dlen = 0;
	keypath->dlen = 0;

        if( NULL == ( rf = regfile_open( path ) ) )
		return -1;

        while( -1 != ( ttok = regfile_next( name, val, &type, rf ) ) && RTK_EOF != ttok ) {
                switch( ttok ) {
                        case RTK_VAL:
				if( 0 == cwd->dlen ) {
					ERR_UNIX( EINVAL, "reg_merge(): found value before finding a key name.");
					return -1;
				}
				vbuf_overwrite( keypath, cwd );
				vbuf_charcat( keypath, '/' );
				if( 0 == strcmp( name->b, "@" ) ) {
					vbuf_strcat( keypath, "\\0" );
				}
				else {
					parse_escaped_string( dname, name->b, name->dlen );
					ns_path_element_reg2ns( cname, dname->b );
					vbuf_strncat( keypath, cname->b, cname->dlen );
				}
				DB(DB_VERB, "Setting value(%x) %s\n", val->dlen, keypath->b );
				if( 0 != ns_set_valueT( keypath->b, type, val->b, val->dlen ) ) {
					WARN("Failed to add value <%s>: %s\n", keypath->b, uerr_str( uerr_get() ) );
				}
				ditchcnt++;
                                break;
                        case RTK_KEYNAME:
				ns_path_regkey2ns( prpath, name->b );
				vbuf_strnprepend( prpath, "/", 1 );

				DB(DB_VERB, "Creating key: %s\n", prpath->b );

				if( 0 != ns_mkdir( prpath->b, MKDIR_P ) ) {
					WARN("ns_mkdir() failed on %s: %s\n",
							prpath->b,
							uerr_str( uerr_get() ) );
				}
				vbuf_overwrite( cwd, prpath );
                                break;
                        case RTK_EOF:
                                break;
                        default:
				ERR_UNIX( EINVAL, "Unknown token type %d.", ttok);
				r = -1;
				goto end;
                }
        }

	switch( ttok ) {
		case RTK_EOF:
			break;
		case -1:
			r=-1;
			goto end;
		default:
			r=-1;
			ERR_UNIX( EINVAL, "reg_merge(): ttok == %d -> broken", ttok );
			goto end;
	}

end:
        regfile_close( rf );
	return r;
} // }}}
int ns_regp_val( vbuf *out, char *path ) // {{{
{
	long loff=0;
	unsigned long type;
        VBUF_STATIC( bn, 10 );
        VBUF_STATIC( rbn, 10 );
        VBUF_STATIC( val, 10);
	VBUF_STATIC( valout, 10 );

	vbuf_strset( out, "" );

	ns_vbasename( bn, path );
        ns_path_ns2regvalname( rbn, bn->b );


        if( 0 != ns_read_valueT( path, &type, val ) ) 
                return -1;

        if( 0 == strcmp( rbn->b, "" ) )
		loff += vbuf_catprintf( out, "@=" );
        else 
		loff += vbuf_catprintf( out, "\"%s\"=", rbn->b );

	vbprint_val8( valout, val->b, type, val->dlen, loff );

	vbuf_strcat( out, valout->b );

	return 0;

} // }}}
int ns_read_dword( char *path, int *out ) // {{{
{
	VBUF_STATIC( buf, 10 );
	unsigned long type = 0;

	//WARN("Going in: %s", path);
	hexdump( (char*)&type, 0, sizeof( type ), 1 );
	if( 0 != ns_read_valueT( path, &type, buf ) ) return -1;
	if( type != REG_DWORD ) {
		ERR_UNIX(EINVAL, "Expected a dword, got %d\n", type );
		hexdump( (char*)&type, 0, sizeof( type ), 1 );
		return -1;
	}
	if( buf->dlen != 4 ) {
		ERR_UNIX(EINVAL, "dword length seems incorrect, 4 != %d\n", buf->dlen );
		return -1;
	}

	return vb_dword2int( buf, out );
} // }}}


//TODO: ns_set_dword
// vim600: set foldlevel=0 foldmethod=marker:
