/* {{{ preamble
 * libhive.c
 *
 * Registry Hive access library API
 *
 * Authors:     Sean Loaring
 *              Petter Nordahl-Hagen
 *
 * Copyright (c) 2005-2006 Sean Loaring
 * Copyright (c) 1997-2004 Petter Nordahl-Hagen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * 
 * TODO: this file talks to the outside world in UTF8, while the win32api talks
 * in UTF16.  ntreg.c converts utf8 to 16 to deal with the raw registry files.
 * What should be going on, since I want this to be as win32 as possible, is
 * this should be UTF16 as should ntreg.c.  The caller (nstdreg.c,
 * whateverelse.c) should talk to this in UTF16 and convert as it sees fit.
 *
 * TODO: all of those "reserved" parameters aren't included in prototypes, they
 * should be
 *
 }}} */

/* {{{ behaviour of windows buffer functions
 * Functions that take
 *   - a buffer length variable
 *   - a buffer
 *
 * Will express the following behavior:
 *
 * buffer           length variable
 * NULL             set to required buffer length
 * Large enough     set to size of data coppied to buffer
 * Too small        ERROR_MORE_DATA, set to required buffer length
 *
 }}}*/

// {{{ #include *
#include "libhive.h"
#include "winerr.h"
#include "ntreg.h"
#include "../misc/uerr.h"
#include "../misc/tools.h"

#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#ifdef WITH_WIN32_PASSTHROUGH
#define UNICODE 
#define _UNICODE
#include <windows.h>

//This is shity
#define MAX_KEY_LENGTH_WCHR 255
#define MAX_VALUE_NAME_WCHR 16383

#endif
//}}}

#define MAX_HIVES 10

// return 0 to make the compiler stop yelling
#define STUB { fprintf(stderr, "STUB!\n"); raise(SIGSEGV); return 0; }

#define NK_NORMAL 0
#define NK_META   1
#define NK_NATIVE 2

#define META_NORMAL 0
#define META_ROOT   1
#define META_NATIVE 2

/* rl_hkey rlHKEY_* {{{
 * Under win32 one can pass these generic rl_hkeys to registry functions.  They
 * are allocated with type = NK_META and are handled differently than NK_NORMAL
 * */
rl_hkey rlHKEY_CLASSES_ROOT;
rl_hkey rlHKEY_CURRENT_USER;
rl_hkey rlHKEY_LOCAL_MACHINE;
rl_hkey rlHKEY_USERS;
rl_hkey rlHKEY_PERFORMANCE_DATA;
rl_hkey rlHKEY_CURRENT_CONFIG;
rl_hkey rlHKEY_DYN_DATA;

#define ID_HKEY_CLASSES_ROOT		0
#define ID_HKEY_CURRENT_USER		1
#define ID_HKEY_LOCAL_MACHINE		2
#define ID_HKEY_USERS 			3
#define ID_HKEY_PERFORMANCE_DATA 	4
#define ID_HKEY_CURRENT_CONFIG 		5
#define ID_HKEY_DYN_DATA 		6

//}}}

// {{{ Array of loaded hives 
//struct hive *hive[MAX_HIVES+1];
//char *pathmap[MAX_HIVES+1];

rl_hkey mounts[MAX_HIVES+1];

char *HKEY_CLASSES_ROOT_PATHS[] = {""};
char *HKEY_CURRENT_USER_PATHS[] = {""};
char *HKEY_LOCAL_MACHINE_PATHS[] = { "SAM", "system", "security", "software", "" };
char *HKEY_USERS_PATHS[] = {""};
char *HKEY_PERFORMANCE_DATA_PATHS[] = {""};
char *HKEY_CURRENT_CONFIG_PATHS[] = {""};
char *HKEY_DYN_DATA_PATHS[] = {""};
/*
char **fakepaths[] =
{
	HKEY_CLASSES_ROOT_PATHS,
	HKEY_CURRENT_USER_PATHS,
	HKEY_LOCAL_MACHINE_PATHS,
	HKEY_USERS_PATHS,
	HKEY_PERFORMANCE_DATA_PATHS,
	HKEY_CURRENT_CONFIG_PATHS,
	HKEY_DYN_DATA_PATHS
};
*/
// }}}
int no_hives = 0;
int dirty;
int libhive_inited = 0;

//char *win_errstr[] =  // {{{
//{
//	"This is the error you get when there was no error.",
//	"Buffer too small.  Need more buffer.",
//	"There are no more idems.",
//	"Generic error - please write real error.",
//	"Not found.",
//	"Open failed."
//}; // }}}

/* Forward Defs {{{ */

long tard_set( ulong data_len, ulong *buf_len, char *buf ); 
rl_hkey alloc_rl_meta( struct hive *h, const char *mp ); 
rl_hkey alloc_rl_hkey( struct hive *h ); 


long hkey_info( rl_hkey hkey ); 

#ifdef WITH_WIN32_PASSTHROUGH
rl_hkey alloc_rl_native( HKEY hkey );
#endif

void free_rl_hkey( rl_hkey hkey ); 
int guess_hive_type( struct hive *hdesc ); 
int findhive( char *p );
//}}}

//typedef unsigned long ulong;
//
void populate_all_meta();
void populate_meta(rl_hkey *rl, char *str, int id, int flag);

#ifdef WITH_WIN32_PASSTHROUGH
HKEY get_native_hkey( rl_hkey rl ) // {{{
{

	if( NK_NATIVE == rl->type ) return rl->native;

	ASSERT( NK_META == rl->type, "Got non-meta key");

	switch( rl->flag ) {
		case META_ROOT:
			switch( rl->id ) {
				case ID_HKEY_CLASSES_ROOT: 	return HKEY_CLASSES_ROOT;
				case ID_HKEY_CURRENT_USER: 	return HKEY_CURRENT_USER;
				case ID_HKEY_LOCAL_MACHINE: 	return HKEY_LOCAL_MACHINE;
				case ID_HKEY_USERS: 		return HKEY_USERS;
				case ID_HKEY_PERFORMANCE_DATA: 	return HKEY_PERFORMANCE_DATA;
				case ID_HKEY_CURRENT_CONFIG: 	return HKEY_CURRENT_CONFIG;
				case ID_HKEY_DYN_DATA: 		return HKEY_DYN_DATA;
				default:
					DIE("Unknown root HKEY");
			}
		case META_NATIVE:
			return rl->native;

		default:
			DIE("Got invalid flag value: %d", rl->flag);
	}
} // }}}
#endif

char *get_type_str( ulong type ) // {{{
{
	return (type <= REG_MAX ? val_types[type] : "(unknown)");
} // }}}

/** {{{ Mount/Umount functions **/
char *hl_mount_help() // {{{
{
	return "mount options:\n"
		"  hive_file                 guess mount point\n"
		"  hive_file=mount_location  specify mount point\n"
		"\n"
		"  examples: \n"
		"    /mnt/hda1/winnt/system32/config/software\n"
		"      Load software hive, guess where it goes\n"
		"      (/HKEY_LOCAL_MACHINE/software)\n"
		"    /mnt/hda1/winnt/system32/config/SAM=/HKEY_LOCAL_MACHINE/software/foo\n"
		"        Mount a SAM hive as /HKEY_LOCAL_MACHINE/software/foo\n"
		"\n"
		"  Note: See readme file on how to write using the Linux NTFS driver\n";

} // }}}
int hl_get_no_hives() // {{{
{
	return no_hives;
} // }}}
int mount_seek_gr( const char *s ) // {{{ find the offset of either the first mount path > s or end of elements
{
	//TODO: this could easily be binary search
	int i=0;
	for( i=0; i<no_hives; i++ ) {
		if( 0 < strcmp( s, mounts[i]->path ) ) return i;
	}
	return no_hives;
} // }}}
struct hive *hl_mounts_remove( int n ) // {{{ remove hive from mount table 
{
	struct hive *result;
	ASSERT( n >= 0 && n < no_hives, "hl_hive_remove(): n outside of bounds.");

	//result = hive[hno];

	result = mounts[n]->hd;

	memmove( mounts + n, mounts + n + 1, sizeof( rl_hkey ) * (no_hives - 1 - n ) );

	no_hives--;

	return result;

} // }}}
int hl_mounts_add( rl_hkey hkey ) // {{{ insert a properly-opened hive into mount table
{
	if( no_hives == MAX_HIVES ) {
		ERR_WIN( rlERROR_BROKEN_ERROR, "no_hives == MAX_HIVES, sorry" );
		return -1;
	}

	ASSERT( hkey->path != NULL, "hl_mounts_add(): attempt to add hkey with no path.");

	ulong n = mount_seek_gr( hkey->path );

	//hive[no_hives] = h;

	memmove( mounts + n + 1, mounts + n, sizeof( rl_hkey ) * (no_hives - n) );
	mounts[n] = hkey;

	/*
	if( 0 != htype ) {
		pathmap[no_hives] = hivepath[htype];
	}
	else {
		WARN("Uhhhhh...Nothing assigned.  TODO: fix this.");
		parse_regf_header( h, h->head );
	}
	*/

	no_hives++;

	return 0;
} // }}}
int meta_enum( rl_hkey hkey, ulong idx, vbuf *vb ) // {{{
{
	int i=0;
	int cnt=0;

	for(i=0; i<no_hives; i++) {
		if( mounts[i]->path == strstr( mounts[i]->path, hkey->path ) ) {
			int inlen = strlen( hkey->path );
			char *subpath = mounts[i]->path + inlen;
			if( *subpath == '\\' ) {
				while( *subpath == '\\' ) subpath++;
				if( idx == cnt ) {
					int splen;
					char *element_end = index( subpath, '\\' );
					if(!element_end)element_end = subpath + strlen( subpath );
					splen = element_end - subpath; 
					vbuf_strnset( vb, subpath, splen );
					return 0;
				}
				cnt++;
			}
		}
	}
	//WARN("meta_enum() calling ERR_WIN()");
	ERR_WIN( rlERROR_NO_MORE_ITEMS, "meta_enum() couldn't find anything at index %d.", idx );
	//WARN("back in meta_enum()...");

	return -1;
} // }}}
int hl_unload_hive( int hno ) // {{{ 
{
	struct hive *h = hl_mounts_remove( hno );
	int result = 0;

	ASSERT( NULL != h, "hl_hive_remove() returned NULL.\n");

	if( 0 != writeHive( h ) )
		result = -1;

	closeHive( h );

	return result;
} // }}}
long hl_mount( const char *fname, const char *mpoint, int mode ) // {{{ 
{
	struct hive *h = openHive(fname, mode);
	const char *mp = mpoint;

	if( mpoint[0] != '/' ) 
		WARN("You are trying to mount a hive file at a location not prefixed with a / (probably not what you want).");

	DB(DB_1, "Attempting to load hive file %s.", fname);

	if( NULL == h ) return rlERROR_OPEN_FAILED;

	if( mpoint[0] == '\0' ) {
		int htype = guess_hive_type( h );
		if( !hivepath[htype] ) {
			ERR_WIN( rlERROR_BROKEN_ERROR, 
				"no mount point specified for hive file %s and can not find default mount point.\n", fname );
			return -1;
		}

		mp = hivepath[htype];
		//fprintf(stderr, "hl_mount(): no mount point given for hive, using %s\n", mp );
	}

	DB(DB_0, "Attempting to mount hive file %s at %s.", fname, mp);

	rl_hkey hkey = alloc_rl_meta( h, mp );

	if( 0 != hl_mounts_add( hkey ) ) {
		closeHive( h );
		return -1;
	}

	return 0;
} // }}}
void hl_close_all() // {{{
{
	while( no_hives ) {
		hl_unload_hive( 0 );
	}
} // }}}
void flush_hives() // {{{
{
	int i;
	for( i=0; i<no_hives; i++) {
		writeHive( mounts[i]->hd );
	}
	/*
	for( i=0; i<no_hives; i++) {
		writeHive(hive[i]);
	}
	*/
} // }}}
int hl_mount_list( int argc, char **argv ) // {{{
{
	int i=0;
	static vbuf *fname = NULL;
	static vbuf *mpoint = NULL;
	char *p;

	if( !fname ) fname = vbuf_alloc(0);
	if( !mpoint ) mpoint = vbuf_alloc(0);

	for(i=0; i<argc; i++) {

		if( (p=esc_index( argv[i], '=')) ) {
			vbuf_strnset( fname, argv[i], p - argv[i] );
			vbuf_strset( mpoint, p + 1 );
		}
		else {
			vbuf_strset( fname, argv[i] );
			vbuf_strset( mpoint, "" );
		}

		if( 0 != hl_mount( fname->b, mpoint->b, 0 ) )
			return -1;
	}
	return 0;
} // }}}
int hl_mount_count() // {{{ //TODO: maybe hl_get_no_hives() just calls this function... :P
{
	return hl_get_no_hives(); 
} // }}}
int hl_mount_get_point( vbuf *vb, int no ) // {{{
{
	ASSERT( no < no_hives, "no >= no_hives" );

	rl_hkey hkey = hl_mount_get_hkey( no );
	vbuf_strset( vb, hkey->path );

	return 0;

} // }}}
rl_hkey hl_mount_get_hkey( int no ) // {{{
{
	return mounts[no];
} // }}}

// }}}
/** IN USE **/

void rl_hkey_dump(rl_hkey hk) // {{{
{
	WARN("rlhkey_dump:\n  type: %d\n  path: %s\n  id: %d\n  flag: %d\n  hd: %p\n  nk: %p", hk->type, hk->path, hk->id, hk->flag, hk->hd, hk->nk );

#ifdef WITH_WIN32_PASSTHROUGH
	WARN("  and native: %p", hk->native );
#endif
} // }}}
/* helpers - these rely on rlReg* functions, they just try to make life a bit easier */
struct valh *vh_open( rl_hkey hk, char *path ) // {{{
{ 
	struct valh *result;
	result = F_MALLOC( sizeof( struct valh ) );

	result->blen = 0;
	result->buf = NULL;

	if( rlERROR_MORE_DATA != rlRegQueryValueEx( hk, path, &result->type, NULL, &result->blen ) ) {
		FREE( result );
		return NULL;
	}

	result->buf = F_MALLOC( result->blen );

	if( rlERROR_SUCCESS != rlRegQueryValueEx( hk, path, &result->type, result->buf, &result->blen ) ) {
		FREE( result->buf );
		FREE( result );
		return NULL;
	}

	return result;
} // }}}
int vh_close( struct valh *vh ) // {{{ Commit changes made to vh to registry free()
{
	int result = 0;
	//TODO: I don't know if Reg{Set,Query,Whatever}ValueEx will recurse beyond the hkey or not
	//      If not then a dirname()/basename() messy horror will have to be associated with these
	//      helpers
	if( rlERROR_SUCCESS != rlRegSetValueEx( vh->hk, vh->name, 0, vh->type, vh->buf, vh->blen ) ) {
		result = -1;
	}
	vh_drop( vh );
	return result;
} // }}}
int vh_drop( struct valh *vh ) // {{{ Don't commit anything, just free()
{
	FREE( vh->buf );
	FREE( vh );
	return 0;
} // }}}
int vh_set_buf( struct valh *vh, void *buf, long len ) // {{{ resize vh buffer and copy buf into it
{
	if( len != vh->blen ) {
		if( buf == vh->buf ) {
			void *b = F_MALLOC( len );
			bzero(b, len );
			memcpy( b, buf, (len<vh->blen)?len:vh->blen );
			free( vh->buf );
			vh->buf = b;
			vh->blen = len;
		}
		else {
			vh->buf = F_REALLOC( vh->buf, len );
			vh->blen = len;
		}
	}
	memcpy( vh->buf, buf, len );
	return 0;
} // }}}
ulong vh_get_dword( struct valh *vh ) // {{{
{
	ulong result;
	ASSERT( vh->blen == sizeof( ulong ), "vh_get_dword(): data length is not correct." );
	ASSERT( vh->type == REG_DWORD, "vh_get_dword(): called on non-dword." );

	memcpy( &result, vh->buf, sizeof( ulong ));
	return result;
} // }}}
int vh_set_dword( struct valh *vh, ulong val ) // {{{
{
	ASSERT( 0 == vh_set_buf( vh, (char*)&val, sizeof( ulong )), "vh_set_word(): error in reg_set_buf()"); 
	vh->type = REG_DWORD;

	return 0;
} // }}}

/*
void normalize_path( char *p ) //{{{ in: "\\broken\\\\path/with////mixed/slashes"; out: "\\clean\\path\\"  
{
	int i=0;
	int blen = strlen(p);

	//change forward slashes to backslashes (because escape characters are great to puke everywhere)
	for( i=0; i<blen; i++ ) if( p[i] == '/' ) p[i] = '\\';

	//eat multiple backslashes
	for( i=0; i<blen; i++ ) {
		if( p[i] == '\\' ){
			int end=i;
			while( end < blen && p[end] == '\\' ) end++;
			if( end > i + 1) {
				memmove( &p[i]+1, &p[end], blen-i );
			}
		}
	}

	//WARN("normalize_path() out: %s", p );

} // }}}
*/
#ifdef WITH_WIN32_PASSTHROUGH
/*
long nativeval( ulong *type, char *buf, ulong *blen ) // {{{
{
	VBUF_STATIC( val8buf, 10 );
	VBUF_STATIC( val16buf, 10 );
	//WARN("nativeval(): %p %p %p\n", type, buf, blen );
	if( blen && buf && ( REG_SZ == *type || REG_EXPAND_SZ == *type ) ) {

		cleanutf16( val16buf, buf, *blen );

		//WARN("Bark? %d", *blen);
		//hexdump( val16buf->b, 0, val16buf->dlen, 1 );
		ASSERT( 0 == vb_utf16to8( val8buf, val16buf->b, val16buf->dlen), "utf16 to 8 conversion failed.");
		if( *blen < val8buf->dlen ) return rlERROR_MORE_DATA; 
		//ASSERT( *blen >= val8buf->dlen, "Not enough buffer.");
		memcpy( buf, val8buf->b, val8buf->dlen );
		*blen = val8buf->dlen;
	}

	return 0;
} // }}}
*/
long cleanutf16( vbuf *v, char *buf, ulong len ) // {{{ Windows UTF16 is a) zero-terminated, b) not-zero-terminated, c) zero-terminated with trailing garbage, this turns a,b,c -> a
{
	int i;
	int end = len;
	//check for zero-terminated with trailing garbage
	for( i=0; i<=len - 2; i+=2 ) {
		if( 0 == buf[i] && 0 == buf[i+1] ) {
			end = i + 2;
			break;
		}
	}

	vbuf_set( v, buf, end );

	//TODO: fail properly instead of crashing

	//WARN("Bark. %d", *blen);
	//	hexdump( val16buf->b, 0, val16buf->dlen, 1 );

	//zero-terminate if not already terminated
	if( v->dlen < 2 || v->b[ v->dlen - 1 ] != 0 || v->b[ v->dlen - 2 ] != 0 )
		vbuf_append( v, "\0\0", 2 );

	return 0;
} // }}}
long vb_winutf16to8( vbuf *out, char *buf, ulong blen ) // {{{
{
	VBUF_STATIC( tmp, 10 );

	cleanutf16( tmp, buf, blen );

	ASSERT( 0 == vb_utf16to8( out, tmp->b, tmp->dlen), "utf16 to 8 conversion failed.");

	return 0;
} // }}}
long bufpopulateutf16( vbuf *in, char *buf, ulong *blen ) // {{{
{
	VBUF_STATIC( val8dump, 10 );
	int r = rlERROR_SUCCESS;

	vb_winutf16to8( val8dump, in->b, in->dlen );

	if( *blen < val8dump->dlen && buf ) {
		r = rlERROR_MORE_DATA;
	} else {
		if( buf ) {
			memcpy( buf, val8dump->b, val8dump->dlen );
		}
	}

	if( blen ) {

		//WARN("%d %d %d", r, *blen, val8dump->dlen);
		*blen = val8dump->dlen;
	}

	//WARN("%p return %d", buf, r );
	return r;
} // }}}
long rlRegEnumValueNative( rl_hkey key, ulong index, char *vname, ulong *vnlen,  ulong *type, char *buf, ulong *blen) // {{{
{
	//TODO: this function is horrible
	long r;
	ulong mytype;
	VBUF_STATIC( utf16buf, 100 );
	VBUF_STATIC( utf8buf, 100 );
	VBUF_STATIC( val16dump, 10 );

	long vnlenw = MAX_VALUE_NAME_WCHR - 10;
	ulong *blin;
	ulong blinv;
	if( blen ) {
		blinv = *blen;
		blin = &blinv;
	} else {
		blin = NULL;
	}

	vbuf_resize( utf16buf, MAX_VALUE_NAME_WCHR * 2 );
	//WARN("rlRegEnumValueNative() starting... (blen == %d)", blen);

	r = RegEnumValueW( get_native_hkey( key ), index, (void*)utf16buf->b, &vnlenw, NULL, &mytype, buf, blen );

	// We convert REG_SZ to UTF8 (for now), we must find the size of the output to report required buffer length to caller
	if( blin && ( rlERROR_MORE_DATA == r || rlERROR_SUCCESS == r ) && ( REG_SZ == mytype || REG_EXPAND_SZ == mytype || REG_MULTI_SZ == mytype ) ) {
		int r2;
		long initialsize = 0;
		vnlenw = MAX_VALUE_NAME_WCHR - 10;
		if( rlERROR_SUCCESS != ( r2 = RegEnumValueW( get_native_hkey( key ), index, (void*)utf16buf->b, &vnlenw, NULL, &mytype, NULL, &initialsize ) ) ) 
			DIE("Expected ERROR_SUCCESS %d", initialsize);

		WARN("Setting initialsize: %d w/r=%d", initialsize, r );
		vbuf_resize( val16dump, initialsize );

		vnlenw = MAX_VALUE_NAME_WCHR - 10;
		if( rlERROR_SUCCESS != ( r2 = RegEnumValueW( get_native_hkey( key ), index, (void*)utf16buf->b, &vnlenw, NULL, &mytype, val16dump->b, &initialsize ) ) ) 
			DIE("Expected ERROR_SUCCESS, got %d %d w/init=%d", r2, rlERROR_MORE_DATA, initialsize );

		r = bufpopulateutf16( val16dump, buf, blin );

		if( blen ) *blen = *blin;

		WARN("r == %d", r );
	}


	ERR_WIN( r, "" );
	if( r != rlERROR_SUCCESS ) return r;
	
	// Everything below here deals with the value name

	utf16buf->dlen = vnlenw * 2; //count in 8-bit-bytes, not 16
	vbuf_check( utf16buf );
	if( vnlenw > 0 ) {
		vbuf_append( utf16buf, "\0\0", 2 );
		if( -1 == vb_utf16to8( utf8buf, utf16buf->b, utf16buf->dlen ) ) {
			WARN("vb_utf16to8() failed on following due to: %s", uerr_str( uerr_get() ) );
			hexdump( utf16buf->b, 0, utf16buf->dlen, 1 );
		}
	}
	else {
		vbuf_strset( utf8buf, "" );
		//WARN("ZERO-LENGTH VALUE NAME!");
		//utf8buf->dlen=0;
	}

	/*
	vnlen  vndata enough
	0	0	0
	0	0	1

	0	1	0 UNDEF
	0	1	1

	1	0	0 SET
	1	0	1 SET

	1	1	0 MORE_DATA & SET
	1	1	1 COPY & SET
	*/

	if( vnlen ) {
		if( *vnlen >= utf8buf->dlen ) {
			if( vname ) memcpy( vname, utf8buf->b, utf8buf->blen );
		} else {
			r = rlERROR_MORE_DATA;
		}


		*vnlen = utf8buf->dlen;
	}

	if( type ) *type = mytype;

	//r = nativeval( type, buf, blen );

	/*
	WARN("WTF: %d %p %p", r, buf, blen );
	if( 0 == r && buf && blen ) {
		WARN("Nativeval was called %d.", *blen);
		hexdump( buf, 0, *blen, 1 );
	}
	*/

	return r;
} // }}}
#endif
long rlRegEnumValue(rl_hkey key, ulong index, char *vname, ulong *vnlen,  ulong *type, char *buf, ulong *blen) // {{{
{
	struct vk_key *vk;
	int result1, result2;
	VBUF_STATIC( namebuf, 10 );
	VBUF_STATIC( valbuf, 10 );
	//WARN("rlRegEnumValue() starting, blen: %p", blen);
	switch( key->type ) {
		case NK_NORMAL: // {{{
			{
				ulong vk_name_len, vk_data_len;

				//WARN("That's not right...");
				if( index >= key->nk->no_values ) return rlERROR_NO_MORE_ITEMS;

				vk = nk_get_value( key->hd, key->nk, index );

				vk_data_len = vk_get_length( vk );
				vk_name_len = vk_get_name_len( vk );

				if( 0 == (result1 = tard_set( vk_name_len, vnlen, vname ) ) && NULL != vname ) {
					//*vnlen = vk_get_name( vk, vname, *vnlen );
					vb_vk_get_name( namebuf, vk );
					//WARN("vb_vk_get_name(): got %s", namebuf->b );
					if( *vnlen >= namebuf->dlen )
						memcpy( vname, namebuf->b, namebuf->dlen );
					else 
						DIE("I didn't have room to populate vname, I'm not sure if I should continue");
					*vnlen = namebuf->dlen;
				}


				if( 0 == (result2 = tard_set( vk_data_len, blen, buf ) ) && NULL != buf ) {
					ASSERT( 0 == vk_get_val2( key->hd, vk, valbuf ), "Could not get value.");
					memcpy( buf, valbuf->b, valbuf->dlen );
					//void *p = vk_get_val2( key->hd, vk );
					//memcpy( buf, p, vk_data_len );
				}

				//Yeah, I don't know about this...
				//WARN("WTF: %s", vname );
				ASSERT( 0 != strcmp( vname, "SQL Server\\/" ), "No you don't!" );
				return (result1>result2)?result1:result2;
			} // }}}
		case NK_META:
			//WARN("Hmmm...");
			//meta keys don't have values (or morals).
			return rlERROR_NO_MORE_ITEMS;
#ifdef WITH_WIN32_PASSTHROUGH
		case NK_NATIVE: 

			//WARN("Weee...");
			return rlRegEnumValueNative( key, index, vname, vnlen, type, buf, blen );
#endif
		default:
			DIE("Unknown key type.");
	}

	DIE("I should not execute.");

	return 0;
}
// }}}

#ifdef WITH_WIN32_PASSTHROUGH

#if 0
long rlRegEnumKeyExNative(rl_hkey hkey, ulong index, char *name, ulong *len, char *cname,  // {{{TRUNC
                                   ulong *cnlen, FILETIME *ft) 
{
	unsigned long l = 0;
	for( l = 0; l < 0xffffffffUL; l++ ) {
		char *c = winstrerror( l );
		printf("%07ld: %s\n", l, winstrerror( l ) );
	}
	DIE("Done.");
} // }}}
#endif
#if 1
long rlRegEnumKeyExNative(rl_hkey hkey, ulong index, char *name, ulong *len, char *cname,  // {{{TRUNC
                                   ulong *cnlen, FILETIME *ft) 
{
	VBUF_STATIC( utf16name, 10 );
	VBUF_STATIC( utf16class, 10 );

	VBUF_STATIC( utf8name, 10 );
	VBUF_STATIC( utf8class, 10 );
	ulong inlen=0;
	ulong iclen=0;

	TCHAR    achKey[MAX_KEY_LENGTH_WCHR];   // buffer for subkey name
	DWORD    cbName;                   // size of name string 
	TCHAR    achClass[MAX_PATH] = TEXT("");  // buffer for class name 
	DWORD    cchClassName = MAX_PATH;  // size of class string 
	DWORD    cSubKeys=0;               // number of subkeys 
	DWORD    cbMaxSubKey;              // longest subkey size 
	DWORD    cchMaxClass;              // longest class string 
	DWORD    cValues;              // number of values for key 
	DWORD    cchMaxValue;          // longest value name 
	DWORD    cbMaxValueData;       // longest value data 
	DWORD    cbSecurityDescriptor; // size of security descriptor 
	FILETIME ftLastWriteTime;      // last write time 

	DWORD i, retCode; 

	TCHAR  achValue[MAX_VALUE_NAME_WCHR]; 
	DWORD cchValue = MAX_VALUE_NAME_WCHR; 

	//WARN("rlRegEnumKeyExNative() starting...");
 
    // Get the class name and the value count. 
	/*
    retCode = RegQueryInfoKey(
        get_native_hkey( hkey ),                    // key handle 
        achClass,                // buffer for class name 
        &cchClassName,           // size of class string 
        NULL,                    // reserved 
        &cSubKeys,               // number of subkeys 
        &cbMaxSubKey,            // longest subkey size 
        &cchMaxClass,            // longest class string 
        &cValues,                // number of values for this key 
        &cchMaxValue,            // longest value name 
        &cbMaxValueData,         // longest value data 
        &cbSecurityDescriptor,   // security descriptor 
        &ftLastWriteTime);       // last write time 
	*/
 

	// this doesn't work
	vbuf_resize( utf16name, MAX_KEY_LENGTH_WCHR * 2 );
	inlen = MAX_KEY_LENGTH_WCHR;
	//WARN("Going in...");
	long r = RegEnumKeyExW( get_native_hkey( hkey ), index, (void*)utf16name->b, &inlen, NULL, NULL, &iclen, ft );
	ERR_WIN( r, "" );
	//WARN("win32 call returns: %s", winstrerror( r ) );

	if( r == rlERROR_NO_MORE_ITEMS ) return r;

	utf16name->dlen = inlen * 2;
	vbuf_check( utf16name );

	vbuf_append( utf16name, "\0\0", 2 );

	//ASSERT( r == rlERROR_SUCCESS, "We failed to achieve the error of success.");
	//ASSERT( r == rlERROR_MORE_DATA, "Got something weird back.");
	//WARN("Got %d and %d at %ld", inlen, iclen, index );
	//vbuf_resize( utf16name, inlen );

	//WARN("here.");
	vbuf_resize( utf16class, iclen );

	//TODO: set class length to MAX, whatever that is.

	//r = RegEnumKeyExW( get_native_hkey( hkey ), index, utf16name->b, &inlen, NULL, utf16class->b, &iclen, ft );
	//hexdump( utf16name->b, 0, inlen, 1 );
	//hexdump( utf16class->b, 0, iclen, 1 );

	//WARN("here.");

	if( -1 == vb_utf16to8( utf8name, utf16name->b, utf16name->dlen ) ) {
			WARN("vb_utf16to8() failed on following due to: %s", uerr_str( uerr_get() ) );
			hexdump( utf16name->b, 0, utf16name->dlen, 1 );
	}

	//WARN("here2: %s", utf8name->b);
	if( name && len && * len >= utf8name->dlen  ) {
		memcpy( name, utf8name->b, utf8name->dlen  );
		r = rlERROR_SUCCESS;
	}
	if( len && *len < utf8name->dlen )
		r = rlERROR_MORE_DATA;

	if( len )
		*len =  utf8name->dlen ;

	//DIE("Got: %d, %s", r, winstrerror( r ) );

	//long r = RegEnumKeyExW( get_native_hkey( hkey ), index, name, len, NULL, cname, cnlen, ft );

	//WARN("End function heap verify.");
	//heap_verify();
	return r;
} // }}}
#endif
#endif
long rlRegEnumKeyEx(rl_hkey hkey, ulong index, char *name, ulong *len, char *cname,  // {{{TRUNC
                                   ulong *cnlen, FILETIME *ft) 
{
//TODO: this pass-a-null-buffer-and-pointer-to-a-long-and-get-required-buffer-length-back thing doesn't work for RegEnumKeyEx().  So, for compatibility, I need to make this function even more dysfunctional than it already is.
	struct nk_key *nk;
	int result1 = rlERROR_SUCCESS, result2=rlERROR_SUCCESS;
	VBUF_STATIC( vb, 10 );
	VBUF_STATIC( namebuf, 10 );

	switch( hkey->type) {
		case NK_NORMAL: { // {{{
				ASSERT( index <= nk_get_no_subkeys( hkey->nk ), "index is out of bounds" );
				ulong nk_name_len, nk_class_len;
				if( index >= nk_get_no_subkeys( hkey->nk ) ) return rlERROR_NO_MORE_ITEMS;

				nk = nk_get_subkey( hkey->hd, hkey->nk, index );

				if( 0 != nk_vb_get_name( nk, namebuf ) ) {
					return rlERROR_BROKEN_ERROR;
				}

				nk_name_len = namebuf->dlen;
				//nk_name_len = nk_get_name_len( nk ) + 1;
				nk_class_len = nk_get_class_len( nk ) + 1;

				if( 0 == (result1 = tard_set( nk_name_len, len, name ) ) && NULL != name ) {
					//nk_get_name( nk, name, *len );
					ASSERT( *len >= nk_name_len, "Avoided buffer overflow. :P");
					memcpy( name, namebuf->b, namebuf->dlen );
				}

				if( NULL != cnlen ) {
					if( 0 == (result2 = tard_set( nk_class_len, cnlen, cname ) ) && NULL != cname ) {
						nk_get_classname( hkey->hd, nk, cname, *cnlen );
					}
				}

				//Yeah, I don't know about this...

				//TODO: this stuff is useless now
				//WARN("WTF: X %s X %s X", name,"SQL Server\\\\\\/" );
				ASSERT( 0 != strcmp( name, "SQL Server\\\\\\/" ), "No you don't!" );
				//WARN("Something is broken: >%s<", name );
				return (result1>result2)?result1:result2;
			} // }}}
		case NK_META: { // {{{
				//WARN("rlRegEnumKeyEx(): Meta.");
				if( 0 != meta_enum( hkey, index, vb ) ) {
					ASSERT( uerr_get()->errclass == UERR_WIN, "Uhhh...that's weird.");

#ifdef WITH_WIN32_PASSTHROUGH
					//WARN("rlRegEnumKeyEx(): trying passthrough stuff.");
					long r;
					r = rlRegEnumKeyExNative( hkey ,index, name, len, cname, cnlen, ft );
					//r = RegEnumKeyExW( get_native_hkey(hkey), index, name, len, NULL, cname, cnlen, ft );

					//WARN("RegEnumKeyExW() returns %ld, %ld is no more", r, ERROR_NO_MORE_ITEMS );
					return r;
#endif
					return uerr_get()->errnumber;
				}


				if( 0 == (result1 = tard_set( vb->dlen - 1, len, name ) ) && NULL != name ) {
					strcpy( name, vb->b );
				}

				//TODO: I guess I should do the class name stuff here too...

				return (result1>result2)?result1:result2;
			} // }}}
#ifdef WITH_WIN32_PASSTHROUGH
		case NK_NATIVE: 
			return rlRegEnumKeyExNative( hkey, index, name, len, cname, cnlen, ft );
#endif
		default:
			DIE("Unknown key type: %d", hkey->type );
	}

	DIE("I should not execute.");

	return 0;

}
// }}}
/*
long rlRegOpenHiveFile(rl_hkey *result, const char *fname, const char *keypath, int mode ) // {{{
{
	struct nk_key *rootkey;

	if( 0 != load_hive( fname, mode ) ) {
		return rlERROR_OPEN_FAILED;
	}
	*result = alloc_rl_hkey( hive[no_hives-1] );

	fprintf(stderr, "Line executed properly.\n");
	rootkey = get_root_key( (*result)->hd );

	//TODO: I guess I should return an error value here rather than crashing and burning
	//...but crashing and burning is more fun
	ASSERT( NULL != ((*result)->nk = nk_trav_path( (*result)->hd, rootkey, keypath ) ), "nk_trav_path() returned NULL.\n" ); //TODO: return an error

	return 0;

} */
// }}}
long rlRegQueryValueEx(rl_hkey hkey, const char *vname, ulong *type, char *buf, ulong *blen) // {{{
{ //TODO: this should return SUCCESS if the given blen is zero and buf == NULL
	//TODO: this function is horrible
	struct vk_key *vk;
	int result;
	ulong vk_data_len;
	VBUF_STATIC( utf16buf, 10 );
	VBUF_STATIC( valbuf, 10 );
	VBUF_STATIC( val16dump, 10 );
	VBUF_STATIC( val8dump, 10 );

	ulong *blin;
	ulong blinv;
	if( blen ) {
		blinv = *blen;
		blin = &blinv;
	} else {
		blin = NULL;
	}

	/*
	if( blen ) {
		WARN("WTF: %d", *blen );
	}
	else {
		WARN("DEBWH?");
	}
	*/

	switch( hkey->type ) {
		case NK_NORMAL: { // {{{
//#warning   "You need to prevent NULL from going through here after you check if it can handle it"

			//if( NULL != index( vname, '\\' ) )
			//	WARN("rlRegQueryValueEx(): executing without traversal with odd name %s.", vname);
			int val_no = nk_find_value( hkey->hd, hkey->nk, vname );
			if( -1 == val_no || NULL == ( vk = nk_get_value( hkey->hd, hkey->nk, val_no ) ) ) {
			//if( NULL == (vk = vk_trav_path( hkey->hd, hkey->nk, vname )) ) {
				ERR_WIN(rlERROR_NOT_FOUND, "Value named \"%s\" not found.", vname );
				return rlERROR_NOT_FOUND;
			}

			vk_data_len = vk_get_length2( hkey->hd, vk );
			//WARN("rlRegQueryValueEx(): vk_data_len == %d", vk_data_len);

			if( 0 == (result = tard_set( vk_data_len, blen, buf ) ) && NULL != buf ) {
				ASSERT( 0 == vk_get_val2( hkey->hd, vk, valbuf ), "Could not get value.");
				ASSERT( *blen >= valbuf->dlen, "I'm trying to overrun a buffer.");
				memcpy( buf, valbuf->b, valbuf->dlen );
				/*
				void *p;
				if( NULL == ( p = vk_get_val( hkey->hd, vk ) ) ) {
					return rlERROR_NOT_FOUND;
				}
				memcpy( buf, p, vk_data_len );
				*/
				/*
				if( vk_data_len > 1000 ) {
					char badp[] = { 0x64, 0x62, 0x02 };
					ASSERT( 0 != memcmp( p, badp, 3 ), "uh-oh." )
					fprintf(stderr, "PATH: %s\n", vname );
					hexdump( p, 0, vk_data_len, 1 );
				}
				*/
			}

			if( NULL != type ) *type = vk_get_type( hkey->hd, vk );

			return result;
		} // }}}
		case NK_META: { // {{{
			//build path (meta+skname)
			char combpath[SZ_MAX], *cp;
			int hiveno;
			long rcode;
			rl_hkey rootkey;

			strncpy( combpath, hkey->path, SZ_MAX );
			strncat( combpath, vname, SZ_MAX );
			normalize_path( combpath, "\\/" );
			if( 0 > (hiveno = findhive( combpath )) ) return rlERROR_NOT_FOUND;
			cp = combpath + strlen( mounts[hiveno]->path );

			rootkey = alloc_rl_hkey( mounts[hiveno]->hd );
			rootkey->nk = get_root_key( mounts[hiveno]->hd );
			WARN("Recursing");
			rcode = rlRegQueryValueEx( rootkey, cp, type, buf, blen );
			free_rl_hkey( rootkey );
			return rcode;
		} // }}}
#ifdef WITH_WIN32_PASSTHROUGH
		case NK_NATIVE: { // {{{
			long r;
			ulong type1 = 0;
			ASSERT( vname, "vname can not be null.");
			vb_utf8to16( utf16buf, (void*)vname, strlen( vname ) + 1 );

			r = RegQueryValueExW( get_native_hkey( hkey ), (void*)utf16buf->b, NULL, &type1, buf, blen );
			if( type ) *type = type1;

			/*
			if( blen ) {
				WARN("rlRegQueryValueEx() w/ blen %d", *blen );
			}
			else {
				WARN("no blen");
			}
			*/

			if( blin  && ( rlERROR_MORE_DATA == r || rlERROR_SUCCESS == r ) && ( REG_SZ == type1 || REG_EXPAND_SZ == type1 || REG_MULTI_SZ == type1 ) ) {
				long initialsize = 0;
				int r2;
				if( rlERROR_SUCCESS != ( r2 = RegQueryValueExW( get_native_hkey( hkey ), (void*)utf16buf->b, NULL, &type1, NULL, &initialsize ) ) ) 
					DIE("Expected ERROR_SUCCESS");

				vbuf_resize( val16dump, initialsize );

				if( rlERROR_SUCCESS != ( r2 = RegQueryValueExW( get_native_hkey( hkey ), (void*)utf16buf->b, NULL, &type1, val16dump->b, &initialsize ) ) ) 
					DIE("Expected ERROR_SUCCESS");

				//WARN("Calling pop w/ buf %p, blin %d", buf, blin );
				r = bufpopulateutf16( val16dump, buf, blin );

				if( blen ) 
					*blen = *blin;


				/*
				vb_winutf16to8( val8dump, val16dump->b, initialsize );

				if( *blen < val8dump->blen ) {
					r = rlERROR_MORE_DATA;
				} else {
					memcpy( buf, val8dump->b, *blen );
				}

				if( blen ) *blen = val8dump->dlen;
				*/

				//WARN("B r == %d, %d", r, *blen );
			}


			/*
	WARN("WTF: %d %p %p", r, buf, blen );
	if( 0 == r && buf && blen ) {
		WARN("Nativeval was called %d.", *blen);
		hexdump( buf, 0, *blen, 1 );
	}
	*/

			if( 0 != r ) ERR_WIN(r, "" );
			return r;
		} // }}}
#endif
		default:
			DIE("rlRegQueryValueEx(): unknown default case: %d", hkey->type);
	}

	return 0;
} // }}}
#ifdef WITH_WIN32_PASSTHROUGH
long rlRegOpenKeyExNative( rl_hkey hkey, const char *skname, ulong options, REGSAM rsam, rl_hkey *result) // {{{
{
	long r;
	//WARN("rlRegOpenKeyExNative() starting... %s, %ld:", skname, rsam);
	HKEY native_result;
	VBUF_STATIC( utf16buf, 100);
	/*
	switch( hkey->flag ) {
		case META_ROOT: 
		case META_NATIVE:
		*/
			ASSERT( NULL != skname, "skname can not be NULL.");
			vb_utf8to16( utf16buf, (void*)skname, strlen( skname ) + 1 );

			//hkey_info( hkey );
			//WARN("rlRegOpenKeyExNative() trying to open %s\n", skname );
			//vbuf_append( utf16buf, "\0\0", 2 );
			//hexdump( utf16buf->b, 0, utf16buf->dlen, 1 );
			//vbuf_dump( utf16buf );
			//WARN("native key open...");
			r = RegOpenKeyExW( get_native_hkey( hkey ), (void*)utf16buf->b, options, rsam, &native_result );

			//if( 0 != r ) WARN("I got result: %d, %s\n", r, winstrerror(r) );
			//ERR_WIN( r, winstrerror(r) );
			ERR_WIN( r, "" );
			//WARN("Error set: %s", uerr_str( uerr_get() ) );
			if( 0 != r ) return r;
			//if( 0 != r ) DIE("I got result: %d, %s\n", r, winstrerror(r) );
			/*
			break;
		default:
			DIE("Unknown case: %d", hkey->flag);
	}
	*/

	*result = alloc_rl_native( native_result );

	//WARN("rlRegOpenKeyExNative() is returning.");

	return r;
} // }}}
#endif
long hkey_info( rl_hkey hkey ) // {{{ Debugging: dump some basic hkey info
{
	VBUF_STATIC( vb, 6000 );
	ulong type;
	ulong index=0;
	ulong len;
	ulong buflen=50;
	char buf[buflen];
	WARN("hkey_info() running...");

	WARN("Subkeys (vb->blen==%d):", vb->blen);
	len = vb->blen;

	while( 0 == rlRegEnumKeyEx( hkey, index, vb->b, &len, NULL, NULL, NULL ) ) {
		WARN(" i=%d:", index );
		index++;
		hexdump( vb->b, 0, len, 1 );
		len = vb->blen;
	}
	WARN("Values (vb->blen==%d):", vb->blen);

	len = vb->blen;
	WARN("len going in: %d", len );
	while( 0 == rlRegEnumValue( hkey, index, vb->b, &len, &type, buf, &buflen ) ) {
		WARN(" i=%d, len=%d, type=%d, hexdump:", index, len, type );
		index++;
		hexdump( vb->b, 0, len, 1 );
		hexdump( buf, 0, buflen, 1 );
		len = vb->blen;
	}

}  // }}}
long rlRegOpenKeyEx(rl_hkey hkey, const char *skname, ulong options, REGSAM rsam, rl_hkey *result) // {{{
{
	skname = (skname != NULL)?skname:"";

	switch( hkey->type ) {
		case NK_NORMAL:
		{
			//int skidx = nk_find_subkey( hkey->hd, hkey->nk, skname );
			*result = alloc_rl_hkey( hkey->hd );

			//parse_nk_nice( hkey->hd, hkey );

			if( NULL == ((*result)->nk = nk_trav_path( hkey->hd, hkey->nk, skname ) )) {
				//TODO: error from nk_trav_path
				ERR_WIN( rlERROR_NOT_FOUND, "nk_trav_path() was unsuccessful opening \"%s\"", skname );
				return rlERROR_NOT_FOUND;
			}

			//if( -1 == skidx ) DIE("Couldn't open %s.", skname); //TODO: shouldn't die, should return error //return -1;

			//(*result)->nk = nk_get_subkey( hkey->hd, hkey->nk, skidx );
			break;
		}
		case NK_NATIVE: {
#ifdef WITH_WIN32_PASSTHROUGH
			long r = rlRegOpenKeyExNative( hkey, skname, options, rsam, result );
			//WARN("Returning from rlRegOpenKeyEx()");
			return r;
#else
			DIE("Key type is native without WITH_WIN32_PASSTHROUGH");
#endif
		}
		case NK_META: { // {{{
			//WARN("rlRegOpenKeyEx(): found META");
			//build path (meta+skname)
			char combpath[SZ_MAX], *cp;
			int hiveno;
			long rcode;
			rl_hkey rootkey;

			if( 0 == strcmp( skname, "" ) ) {
				//WARN("The cow said moo.");
				*result = hkey;
				return rlERROR_SUCCESS;
			}

			strncpy( combpath, hkey->path, SZ_MAX );
			strncat( combpath, "\\", SZ_MAX );
			strncat( combpath, skname, SZ_MAX );
			normalize_path( combpath, "\\" );
			if( 0 > (hiveno = findhive( combpath )) ) {
#ifdef WITH_WIN32_PASSTHROUGH
				return rlRegOpenKeyExNative( hkey, skname, options, rsam, result );
#else
				//TODO: error from findhive()
				ERR_WIN( rlERROR_NOT_FOUND, "findhive() was unsuccessful.  Are the correct hives loaded?");
				return rlERROR_NOT_FOUND;

#endif
			}
			//else { WARN("You branched out") }
			cp = combpath + strlen( mounts[hiveno]->path );

			rootkey =  alloc_rl_hkey( mounts[hiveno]->hd );
			rootkey->nk = get_root_key( mounts[hiveno]->hd );
			rcode = rlRegOpenKeyEx( rootkey, cp, 0, rsam, result );
			free_rl_hkey( rootkey );
			return rcode;
		} // }}}
		default:
			DIE("rlRegOpenKeyEx(): unknown default case: %d", hkey->type );
	}

	return 0;
}
// }}}
int hkey_get_path(rl_hkey hk, char *path, int maxlen) // {{{
{
	return nk_get_path( hk->hd, hk->nk, path, maxlen );
} // }}}
long rlRegDeleteKey(rl_hkey key, const char *skname) // {{{
{
#ifdef WITH_WIN32_PASSTHROUGH
	int r;
	if( NK_NATIVE == key->type ) {
		VBUF_STATIC(utf16buf, 100 );
		vb_utf8to16( utf16buf, (char*)skname, strlen( skname ) + 1 );
		//WARN("CONSIDERING CALLING NATIVE RegDeleteKey().");
		//hexdump( utf16buf->b, 0, utf16buf->dlen, 1 );
		r = RegDeleteKeyW( get_native_hkey( key ), utf16buf->b );
		ERR_WIN( r, "" );
		return r;
		//return 0;
		
	}
#endif
	if( key->type != NK_NORMAL ) {
		ERR_WIN( rlERROR_NOT_FOUND, "Key is not real.");
		return rlERROR_NOT_FOUND;
	}
	if( 0 == nk_delete_subkey( key->hd, key->nk, skname ) ) {
		return rlERROR_SUCCESS;
	}

	return rlERROR_BROKEN_ERROR;
} // }}}
long rlRegDeleteValue(rl_hkey key, const char *vname) // {{{
{
	switch( key->type ) {
		case NK_NORMAL: // {{{
		{
			int val_no = nk_find_value( key->hd, key->nk, vname );
			if( -1 == val_no ) {
				//TODO: error is gross
				ERR_WIN( rlERROR_BROKEN_ERROR, "rlRegDeleteValue(): Couldn't find value \"%s\"", vname );
				return rlERROR_BROKEN_ERROR;
			}
			if( 0 == nk_delete_value( key->hd, key->nk, val_no) )
				return rlERROR_SUCCESS;

			//TODO: fix error
			ERR_WIN(rlERROR_BROKEN_ERROR, "rlRegDeleteValue(): Couldn't delete value %s", vname );
			return rlERROR_BROKEN_ERROR;
		} // }}}
#ifdef WITH_WIN32_PASSTHROUGH
		case NK_NATIVE: // {{{
		{
			VBUF_STATIC( utf16buf, 100 );
			vb_utf8to16( utf16buf, (char*)vname, strlen( vname ) +  1);

			long r = RegDeleteValueW( get_native_hkey( key ), (void*)utf16buf->b );
			ERR_WIN( r, "" );
			return r;
			break;
		} // }}}
#endif
		default:
			DIE("Unknown key-type: %d", key->type );
	}
} // }}}

/** CURRENT **/

long rlRegQueryInfoKey(rl_hkey key, char *clname, ulong *cllen, ulong *subkeys, ulong *maxsklen, ulong
		*maxcllen, ulong *nvalues, ulong *maxvalnamelen, ulong *maxvallen, ulong *securitydesc, FILETIME *lastwrite) // {{{
{
	//TODO: This needs some work
	//WARN("Foo!");
	//WARN("KTY: %d, %d", key->type, key->flag );
#ifdef WITH_WIN32_PASSTHROUGH
	if( key->type && key->flag == META_NATIVE || key->type == NK_NATIVE ) { 
		//WARN("Highly untested.");
		//TODO: highly untested
		int r = RegQueryInfoKeyW( get_native_hkey( key ), (void*)clname, NULL, cllen, subkeys, maxsklen, maxcllen, nvalues, maxvalnamelen, maxvallen,
				securitydesc, lastwrite );

		ERR_WIN( r, "" );
		return r;
	}
#endif
	
	if( NULL != subkeys ) {
		*subkeys = nk_get_no_subkeys( key->nk );
	}

	return rlERROR_SUCCESS;
}
// }}}

long rlRegSetValueEx( rl_hkey hkey, const char *vname, ulong reserved, ulong type, const char* buf, ulong blen) // {{{
{
	//TODO: should there be any sort of path traversal here?  I'm not sure win32 api does that
	struct vk_key *vk;
	VBUF_STATIC( utf16valbuf, 10 );

	switch( hkey->type ) {
		case NK_NORMAL: // {{{
		{
//#warning   "You need to prevent NULL from going through here after you check if it can handle it"
			//if the value already exists then we delete it
			int voff;

			/*
			if( NULL != (vk = vk_trav_path( hkey->hd, hkey->nk, vname )) ) {
			}
			*/

			if( 0 <= ( voff = nk_find_value( hkey->hd, hkey->nk, vname ) ) ) {
				if( 0 != nk_delete_value( hkey->hd, hkey->nk, voff ) ) {
					//TODO: error from nk_delete_value()
					ERR_WIN( rlERROR_BROKEN_ERROR, "rlRegSetValueEx(): attempt to delete existing value failed.");
					return rlERROR_BROKEN_ERROR;
				}
			}

			return nk_set_value( hkey->hd, hkey->nk, vname, type, buf, blen );
		} // }}}
		case NK_META: // {{{
		{ 
			//WARN("rlRegOpenKeyEx(): found META");
			//build path (meta+skname)
			//TODO: this probably should not exist, path traversal in this function probably should not be
			char combpath[SZ_MAX], *cp;
			int hiveno;
			long rcode;
			rl_hkey rootkey;

			strncpy( combpath, hkey->path, SZ_MAX );
			strncat( combpath, vname, SZ_MAX );
			normalize_path( combpath, "\\/" );
			if( 0 > (hiveno = findhive( combpath )) ) return rlERROR_NOT_FOUND;
			cp = combpath + strlen( mounts[hiveno]->path );

			rootkey =  alloc_rl_hkey( mounts[hiveno]->hd );
			rootkey->nk = get_root_key( mounts[hiveno]->hd );
			rcode = rlRegSetValueEx( rootkey, cp, 0, type, buf, blen );
			free_rl_hkey( rootkey );
			return rcode;
		} // }}}
#ifdef WITH_WIN32_PASSTHROUGH
		case NK_NATIVE: // {{{
		{
			VBUF_STATIC( utf16buf, 100 );
			vb_utf8to16( utf16buf, (char*)vname, strlen( vname ) + 1 );
			long r;
			if( type == REG_SZ || type == REG_EXPAND_SZ || type == REG_MULTI_SZ ) {
				vb_utf8to16( utf16valbuf, buf, ( blen ) );

				r = RegSetValueExW( get_native_hkey( hkey ), (void*)utf16buf->b, reserved, type, utf16valbuf->b, utf16valbuf->dlen );
			} else {
				r = RegSetValueExW( get_native_hkey( hkey ), (void*)utf16buf->b, reserved, type, buf, blen );
			}
			ERR_WIN( r, "" );
			return r;
		} // }}}
#endif
		default:
			DIE("unknown default case.");
	}

	return rlERROR_SUCCESS;
} // }}}

long rlRegCreateKeyEx(rl_hkey hkey, const char *skname, const char *r_class, ulong options, REGSAM desired, 
                                     SECURITY_ATTRIBUTES *security, rl_hkey *result, ulong *disposition) // {{{
{
	skname = (skname != NULL)?skname:"";

	ASSERT( result, "result is NULL.  I don't know what to do");

	switch( hkey->type ) {
		case NK_NORMAL: // {{{
		{
			//int skidx = nk_find_subkey( hkey->hd, hkey->nk, skname );
			*result = alloc_rl_hkey( hkey->hd );

			//parse_nk_nice( hkey->hd, hkey );

			if( NULL == ((*result)->nk = nk_create( hkey->hd, skname, hkey->nk ) )) {
				//TODO: error from nk_create
				//ERR_WIN( rlERROR_BROKEN_ERROR, "Couldn't create \"%s\"", skname );
				return rlERROR_BROKEN_ERROR;
			}

			//if( -1 == skidx ) DIE("Couldn't open %s.", skname); //TODO: shouldn't die, should return error //return -1;

			//(*result)->nk = nk_get_subkey( hkey->hd, hkey->nk, skidx );
			break;
		} // }}}
		case NK_META: // {{{
		{
			//WARN("rlRegOpenKeyEx(): found META");
			//build path (meta+skname)
			char combpath[SZ_MAX], *cp;
			int hiveno;
			long rcode;
			rl_hkey rootkey;

			if( 0 == strcmp( skname, "" ) ) {
				*result = hkey;
				return rlERROR_SUCCESS;
			}

			strncpy( combpath, hkey->path, SZ_MAX );
			strncat( combpath, "\\", SZ_MAX );
			strncat( combpath, skname, SZ_MAX );
			normalize_path( combpath, "\\/" );
			if( 0 > (hiveno = findhive( combpath )) ) {
				//TODO: findhive error
				eprintf(UERR_WIN, rlERROR_NOT_FOUND, "Couldn't locate hive %s", combpath );
				return rlERROR_NOT_FOUND;
			}
			cp = combpath + strlen( mounts[hiveno]->path );

			rootkey =  alloc_rl_hkey(mounts[hiveno]->hd);
			rootkey->nk = get_root_key( mounts[hiveno]->hd );
			rcode = rlRegCreateKeyEx( rootkey, cp, r_class, options, desired, security, result, disposition );
			free_rl_hkey( rootkey );
			return rcode;
		} // }}}
#ifdef WITH_WIN32_PASSTHROUGH
		case NK_NATIVE: // {{{
		{
			VBUF_STATIC( utf16, 100 );
			vb_utf8to16( utf16, (char*)skname, strlen( skname ) + 1 );
			HKEY nresult;
			//TODO: I don't deal with class stuff here at all.
			int r = RegCreateKeyExW( get_native_hkey( hkey ), (void*)utf16->b, 0, NULL, options, desired, security, &nresult, disposition );
			ERR_WIN( r, "" );
			if( rlERROR_SUCCESS != r ) return r;

			if( result ) {
				*result = alloc_rl_native( nresult );
			}

			return r;
			break;
		} // }}}
#endif
		default:
			DIE("rlRegCreateKeyEx(): hkey->type == %d.", hkey->type);
	}

	return 0;
}
// }}}

/** INTERNAL **/
void libhive_init() // {{{
{ 
	if( 0 == libhive_inited ) {
		populate_all_meta();
		libhive_inited = 1;
	}
} // }}}
void populate_all_meta() // {{{ populate HKEY_*
{
	populate_meta( &rlHKEY_CLASSES_ROOT, "\\HKEY_CLASSES_ROOT", ID_HKEY_CLASSES_ROOT, META_ROOT);
	populate_meta( &rlHKEY_CURRENT_USER, "\\HKEY_CURRENT_USER", ID_HKEY_CURRENT_USER, META_ROOT);
	populate_meta( &rlHKEY_LOCAL_MACHINE, "\\HKEY_LOCAL_MACHINE", ID_HKEY_LOCAL_MACHINE, META_ROOT);
	populate_meta( &rlHKEY_USERS, "\\HKEY_USERS", ID_HKEY_USERS, META_ROOT);
	populate_meta( &rlHKEY_PERFORMANCE_DATA, "\\HKEY_PERFORMANCE_DATA", ID_HKEY_PERFORMANCE_DATA, META_ROOT);
	populate_meta( &rlHKEY_CURRENT_CONFIG, "\\HKEY_CURRENT_CONFIG", ID_HKEY_CURRENT_CONFIG, META_ROOT);
	populate_meta( &rlHKEY_DYN_DATA, "\\HKEY_DYN_DATA", ID_HKEY_DYN_DATA, META_ROOT);
} // }}}
void populate_meta(rl_hkey *rl, char *str, int id, int flag) //  {{{
{
	//TODO: free this stuff during shutdown
	*rl = F_MALLOC( sizeof( struct nk_desc ) );
	(*rl)->type = NK_META;
	(*rl)->path = str;
	(*rl)->id = id;
	(*rl)->flag = flag;
	//WARN("populate_meta() is untested!");
} // }}}
int pathscore( char *child, char *parent ) // {{{ in: 2 paths; out: number of characters that match, coming from the left
{
	int pl = strlen( parent );
	int cl = strlen( child );
	if( pl > cl ) return -1;
	if( 0 != strncmp( child, parent, strlen( parent ) ) ) return -1;
	return pl;
} // }}}
int findhive( char *p ) // {{{ in: path; out: hive number where path lives or -1
{
	int i, max=0, best=-1;
	//WARN("findhive(): untested");
	for( i=0; i<no_hives; i++ ) {
		int s = pathscore( p, mounts[i]->path );
		if( s > max ) {
			max = s;
			best = i;
		}
	}

	if( -1 == best ) {
		ERR_WIN( rlERROR_NOT_FOUND, "Couldn't find hive." );
	}

	return best;
} // }}}
int guess_hive_type( struct hive *hdesc ) // {{{ try to guess what type of hive hdesc refers to 
{
	int result = HTYPE_UNKNOWN;
	char *hname = regf_get_name( hdesc->head );

	//So, let's guess what kind of hive this is, based on keys in its root
	
	if( 0 == strcmp( hname, "stemRoot\\System32\\Config\\DEFAULT" )) return HTYPE_DEFAULT;

	//TODO: is the path part of the constant or does it change from user to user?
	if( 0 == strcmp( hname, "ettings\\Administrator\\ntuser.dat")) return HTYPE_USER;
	if( 0 == strcmp( hname, "\\SystemRoot\\System32\\Config\\SAM" )) return HTYPE_SAM;
	if( 0 == strcmp( hname, "emRoot\\System32\\Config\\SECURITY" )) return HTYPE_SECURITY;
	if( 0 == strcmp( hname, "emRoot\\System32\\Config\\SOFTWARE" ) ) return HTYPE_SOFTWARE;
	if( 0 == strcmp( hname, "SYSTEM" ) ) return HTYPE_SYSTEM;
	if( 0 == strcmp( hname, "1\\WINNT\\SYSTEM32\\CONFIG\\userdiff" )) return HTYPE_USERDIFF;
	if( 0 == strcmp( hname, "\\Microsoft\\Windows\\UsrClass.dat")) return HTYPE_USRCLASS;

	if (trav_path(hdesc, 0, "\\SAM", 0)) result = HTYPE_SAM;
	else if (trav_path(hdesc, 0, "\\ControlSet", 0)) result = HTYPE_SYSTEM;
	else if (trav_path(hdesc, 0, "\\Policy", 0)) result = HTYPE_SECURITY;
	else if (trav_path(hdesc, 0, "\\Microsoft", 0)) result = HTYPE_SOFTWARE;

	return result;
}
// }}}

long tard_set( ulong data_len, ulong *buf_len, char *buf ) { // {{{ All of the functions that take buffers will return ERROR_MORE_DATA if the buffers aren't big enough

	long result = 0;
	if( NULL != buf_len ) {
		long blin = *buf_len;
		*buf_len = data_len;

		if( buf == NULL ) return rlERROR_SUCCESS;

		if( data_len > blin ) {
			ERR_WIN( rlERROR_MORE_DATA, "" );
			return rlERROR_MORE_DATA;
		}
	}

	return result;

}
// }}}

void free_rl_hkey( rl_hkey hkey ) { // {{{ deallocate an rl_hkey
	//TODO: Should root key handles be "closable"?
	if( hkey->flag == META_ROOT ) return;
	//ASSERT( hkey->flag != META_ROOT, "Attempt to close root key");
	if( hkey->hd ) {
		ASSERT( (hkey->hd->refcount > 0), "Attempt to free an rl_key with refcount > 0");
		hkey->hd->refcount--;
		hkey->hd = NULL;

	}

	if( hkey->path ) free( hkey->path );
#ifdef WITH_WIN32_PASSTHROUGH

	if( hkey->native ) {
		//WARN("Closing native key...");
		RegCloseKey( hkey->native );
	}
#endif

	FREE( hkey );
} // }}}

rl_hkey alloc_rl_meta( struct hive *h, const char *mp ) // {{{ allocate a new rl_hkey for a hive, increment reference count
{
	rl_hkey result = alloc_rl_hkey( h );
	result->path = F_MALLOC( strlen( mp ) + 1 );
	result->type = NK_META;
	result->flag = META_NORMAL;
	strcpy( result->path, mp );
	normalize_path( result->path, "\\/" );
	//WARN("return hkey %p", result );
	return result;
} // }}}
rl_hkey alloc_rl_hkey( struct hive *h ) // {{{ allocate a new rl_hkey for a hive, increment reference count
{
	rl_hkey result;
	result = F_MALLOC( sizeof( struct nk_desc ) );
	result->type = NK_NORMAL;
	result->hd = h;
	result->path = NULL;
	h->refcount++;
	//WARN("return hkey %p", result );
	return result;
} // }}}

#ifdef WITH_WIN32_PASSTHROUGH
rl_hkey alloc_rl_native( HKEY hkey ) { // {{{

	rl_hkey result;
	result = F_MALLOC( sizeof( struct nk_desc ) );
	result->type = NK_NATIVE;
	result->hd = NULL;
	result->path = NULL;
	result->flag = 0;
	result->native = hkey;
	//WARN("return hkey %p", result );

	return result;
} // }}}
#endif

//DWORD rlFormatMessage( DWORD dwFlags, LPCVOID lpSource, DWORD dwMessageId, DWORD dwLanguageId, LPTSTR lpBuffer, DWORD nSize, va_list* Arguments) // {{{
//{
//	/* Oh.  My.  God.  WTF was the person who came up with this API
//	 * function smoking? */
//
//	/* I'm implementing this as quickly and half-assedly as possible.  If
//	 * you need missing features then you'll have to write them yourself */
//
//	ASSERT( dwFlags == (rlFORMAT_MESSAGE_ALLOCATE_BUFFER | rlFORMAT_MESSAGE_FROM_SYSTEM),
//			"FormatMessage() doesn't implement whatever it is that you are trying to do." );
//
//	ASSERT( dwMessageId < MAX_WIN_ERROR, "FormatMessage(): Got a dwMessageId I don't know about." );
//
//	*((void**)lpBuffer) = (void*)F_MALLOC( strlen( win_errstr[dwMessageId] ) );
//
//	strcpy( *(void**)lpBuffer, win_errstr[dwMessageId] );
//
//	return strlen( lpBuffer ) + 1;
//} // }}}

/** UNSORTED **/
long rlRegQueryMultipleValues(rl_hkey hkey, VALENTW *val_list, ulong nvals,
		char *buf, ulong *totsize)
{
	STUB;
}

/* Declarations for functions that are the same in Win16 and Win32 */

long rlRegCloseKey(rl_hkey key) // {{{
{
	ASSERT( NULL != key, "Attempt to close NULL key");
	//WARN("I am going to close the following:");
	//rl_hkey_dump( key );
	//WARN("and here I go...");
	free_rl_hkey( key );
	//WARN("...done");
	return 0;
} // }}}
long rlRegFlushKey(rl_hkey key) // {{{
{
	STUB;
} // }}}
//These depend on some sort of registry patch files (same as output from regedit?) {{{

//long rlRegReplaceKey(rl_hkey, const char *, const char *, const char *);
//long rlRegRestoreKey(rl_hkey, const char *, ulong);
//long rlRegLoadKey(rl_hkey, const char *, const char *);
//long rlRegSaveKey(rl_hkey, const char *, LPSECURITY_ATTRIBUTES);

//}}}
//I'll do these later (maybe) {{{

//long rlRegGetKeySecurity(rl_hkey, SECURITY_INFORMATION, PSECURITY_DESCRIPTOR, LPulong);
//long rlRegUnLoadKey(rl_hkey, const char *);
//long rlRegSetKeySecurity(rl_hkey, SECURITY_INFORMATION, PSECURITY_DESCRIPTOR);
//long rlRegConnectRegistry(const char *, rl_hkey, PHKEY);
//long rlRegOpenCurrentUser(REGSAM, PHKEY);
// }}}
//these functions have Ex counterparts.  I'm going to ignore these for now... {{{

//long rlRegEnumKey(rl_hkey, ulong, char *, ulong);
//long rlRegCreateKey(rl_hkey key, const char *, PHKEY);
//long rlRegOpenKey(rl_hkey key, const char *skname, rl_hkey *result);
//long rlRegQueryValue(rl_hkey, const char *, char *, long *);
//long rlRegSetValue(rl_hkey, const char *, ulong, const char *, ulong);
//}}}
// }}}

// vim600: set foldlevel=0 foldmethod=marker:
