/* {{{ preamble
 * ntreg.c
 *
 * NT Registry Hive access library, constants & structures
 *
 * Authors:	Sean Loaring
 *		Petter Nordahl-Hagen
 *
 * Copyright (c) 2005-2006 Sean Loaring
 * Copyright (c) 1997-2004 Petter Nordahl-Hagen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 }}} */

// {{{ **** Includes
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <syslog.h>
#include <signal.h>

#include "ntreg.h"
#include "../misc/tools.h"
#include "../misc/globals.h"
#include "../misc/uerr.h"
#include "winerr.h"
// }}}
// {{{ **** Defines and macros
#define REGF_HEADER_SIZE 0x1000 /* always 4k? */

#define DELIM '\\'

#define LH_STRICT 1
#define LOAD_DEBUG 1

#define debug(x) {x;}
//#define debug(x) {;}

#define NK_TYPE_ISO8859 0x20
#define NK_TYPE_UTF16   0x00

// }}}
// {{{ **** Some constants
const char ntreg_version[] = "ntreg lib routines, v0.91 040105, (c) Petter N Hagen";

char *val_types[REG_MAX+1] = {
	"REG_NONE", "REG_SZ", "REG_EXPAND_SZ", "REG_BINARY", "REG_DWORD",       /* 0 - 4 */
	"REG_DWORD_BIG_ENDIAN", "REG_LINK",                                     /* 5 - 6 */
	"REG_MULTI_SZ", "REG_RESOUCE_LIST", "REG_FULL_RES_DESC", "REG_RES_REQ", /* 7 - 10 */
};
// }}}
// {{{ **** Globals that should probably not exist
int *watch;
char ddbuf[1024]; //debug crap to delete

//TODO: this should probably be in hivelib.c/h
char* hivepath[] = {
	NULL,
	"\\HKEY_LOCAL_MACHINE\\SAM",
	"\\HKEY_LOCAL_MACHINE\\system",
	"\\HKEY_LOCAL_MACHINE\\SECURITY",
	"\\HKEY_LOCAL_MACHINE\\software",
	"\\HKEY_CURRENT_USER\\",
	NULL,
	"\\HKEY_USERS\\SID-RID_Classes",
	NULL
};
// }}}
// {{{ **** Forward defs
//

// LX
struct nk_key *lx_get_nk( struct hive *hdesc, lx_key *lx, int off );
long lxfindsubkey( struct hive *hdesc, lx_key *lx, const char *name );
lx_key *lx_alloc( struct hive *hdesc, void *p, int id, int count ); 
int lx_remove_entry( struct hive *hdesc, lx_key *lx, int no );
struct nk_key *lx_get_subkey( struct hive *hdesc, lx_key *lx, int off );
lx_key *lx_get_lx( struct hive *hdesc, lx_key *lx, int off, int *off_out );

void *lx_get_raw_slot( lx_key *lx, int off ); // {{{ return the offset to a slot in the array

//LF, LH, LI, RI
void lf_hash( const char *name, char* buf );
void lh_hash_dec( vbuf *out, unsigned long hash ); // Display lh_hash in base 37
unsigned long lh_hash( const char *name ); 
unsigned long lh_hash2( const char *name ); 
lx_key *ri_find_lx( struct hive *hdesc, struct ri_key *ri, int off, int *lx_off, int *lxp_off ); 
lx_key *ri_get_lx( struct hive *hdesc, struct ri_key *ri, int off );

//NK
void nk_vb_get_rawname( struct nk_key *nk, vbuf *out );
struct nk_key* nk_get_parent( struct hive *hdesc, struct nk_key *nk );
int nk_get_path(struct hive *hdesc, struct nk_key *nk, char *path, int maxlen);
lx_key *nk_get_lx( struct hive *hdesc, struct nk_key *nk );
struct nk_key *nk_create( struct hive *hdesc, const char *name, struct nk_key *parent );

//HDESC
void hdesc_free( struct hive *hdesc ); 
void hive_avail_catalog( struct hive *hdesc, struct hbin_page *p, struct hbin_block *b );

//GENERIC POINTER
char *display_ptr_offset( struct hive *hdesc, void *p );
void check_wtf( struct hive *hdesc, HBB *hbb );
//struct hbin_block* hive_get_first_hbinb( struct hive *hdesc ); 

//VK
void vk_delete( struct hive *hdesc, struct vk_key *vk ); 

//HBIN
void parse_hbin_page( struct hive *hdesc, struct hbin_page *p ); 
long hbinb_size( struct hbin_block *p ); 
struct hbin_block* hbinp_first_hbinb( struct hbin_page *p ); 
struct hbin_page* hive_get_first_hbinp( struct hive *hdesc ); 
struct hbin_page* hbinp_get_next_hbinp( struct hive *hdesc, struct hbin_page *p );
struct hbin_block* hbinb_get_next( struct hive *hdesc, struct hbin_page *p, struct hbin_block *b ); 
struct hbin_block* hbinb_get_prev( struct hive *hdesc, struct hbin_page *p, struct hbin_block *b ); 
struct hbin_block* get_hbin_block( struct hive *hdesc, void *p );
struct hbin_page* hive_get_last_hbinp( struct hive *hdesc );


//UNSORTED

//int alloc_val_data(struct hive *hdesc, int vofs, char *path, int size);
//void del_vk(struct hive *hdesc, int i_vkofs);
//int gettoken( char *tok, char **path );
lx_key* pull_lxkey( struct hive *hdesc, void *p ); 

// PARSE
void parse_nk_struct( struct hive *hdesc, struct nk_key *key );
void parse_li_struct(struct hive *hdesc, struct li_key *key);
void parse_vk_struct(struct hive *hdesc, struct vk_key *key);
void parse_vlist(struct hive *hdesc, struct v_list *vl, int no);
void parse_lf_struct(struct hive *hdesc, struct lf_key *key);
void parse_sk_struct(struct hive *hdesc, struct sk_key *key);
void parse_lh_struct(struct hive *hdesc, struct lh_key *key);
void parse_ri_struct(struct hive *hdesc, struct ri_key *key);

// }}}

char *id_to_str( unsigned long id ) // {{{
{
	switch( id ) {
		case ID_REGF_HEADER:
			return "ID_REGF_HEADER";
		case ID_HBIN_PAGE:
			return "ID_HBIN_PAGE";
		case ID_SK_KEY:
			return "ID_SK_KEY";
		case ID_LF_KEY:
			return "ID_LF_KEY";
		case ID_LH_KEY:
			return "ID_LH_KEY";
		case ID_LI_KEY:
			return "ID_LI_KEY";
		case ID_RI_KEY:
			return "ID_RI_KEY";
		case ID_VK_KEY:
			return "ID_VK_KEY";
		case ID_NK_KEY:
			return "ID_NK_KEY";
		default:
			return "UNKNOWN ID";
	}
}
// }}}

// {{{ **** Functions that attempt to contain horrible pointer arithmetic
fileoff get_file_offset( struct hive *hdesc, void *p ) // {{{
{
	return (char*)p - hdesc->buf;
}
// }}}

heapoff get_heap_offset( struct hive *hdesc, void *p ) { return (char*)p - hdesc->heap; }
void *get_ptr_off_f( struct hive *hdesc, long off ) { return (char*)hdesc->buf + off; }
void *get_ptr_off_h( struct hive *hdesc, long off )
{
	ASSERT( off >= 0, "Attempt to get heap pointer with negative offset" );
	//ASSERT( off < hdesc->head->lastblock, "get_ptr_off_h(): offset is outside of valid range.");
	void *result = (void*)hdesc->heap + off;
	//printf("get_ptr_off_h(): pulling type %x == %s\n", 0xFFFF & *(int*)(result + 4), id_to_str( 0xFFFF & (unsigned long)*(int*)(result + 4) ) );
	return result;
}
// }}}

// {{{ **** Get struct from pointer functions

struct hbin_block* get_hbin_block( struct hive *hdesc, void *p ) // {{{
{
	struct hbin_block *result = p;

	/*
	static char bm[] = { 0x64, 0x62, 0x02 };
	ASSERT( 0 != memcmp( bm, p, 3 ), "Here it is?" );
	*/

	//TODO: sanity checks would be nice
	//if( gverbose ) parse_nk( hdesc, (char*)p - (char*)hdesc->buf, 0 );
	return result;
} //}}}
struct vk_key* check_vk_ptr( struct hive *hdesc, void *p ) // {{{
{
	struct vk_key *result = p;

	ASSERT( hdesc, "That's strange.");

	if( result->id != ID_VK_KEY ) 
		DIE( "check_vk_ptr(): Invalid VK_KEY record at 0x%lx (id==0x%x, %s)\n",
			get_file_offset( hdesc, p ), result->id, id_to_str( result->id ) );

	return result;
}
// }}}
struct nk_key* check_nk_ptr( struct hive *hdesc, void *p ) // {{{
{
	struct nk_key *result = p;

	if( result->id == ID_NK_KEY ) 
		return result;
	else {
		DIE("check_nk_ptr(): Invalid NK_KEY record at 0x%lx\n",
			get_file_offset( hdesc, p ) );

		return NULL;
	}
}
// }}}
struct lf_key* check_lf_ptr( struct hive *hdesc, void *p ) // {{{
{
	struct lf_key *result = p;

	if( result->head.id == ID_LF_KEY )
		return result;
	else {
		DIE("Invalid LF_KEY record at 0x%lx\n", get_file_offset( hdesc, p ) );
		return NULL;
	}
}
// }}}
struct lh_key* check_lh_ptr( struct hive *hdesc, void *p ) // {{{
{
	struct lh_key *result = p;

	if( result->head.id == ID_LH_KEY ) {
		return result;
	}
	else {
		DIE("Invalid LH_KEY record at 0x%lx\n", get_file_offset( hdesc, p ) );
		return NULL;
	}
}
// }}}
struct sk_key* check_sk_ptr( struct hive *hdesc, void *p ) // {{{
{
	struct sk_key *result = p;

	if( result->id == ID_SK_KEY )
		return result;
	else {
		DIE("Invalid SK_KEY record at 0x%lx\n", get_file_offset( hdesc, p ) );
		return NULL;
	}
}
// }}}
struct ri_key* check_ri_ptr( struct hive *hdesc, void *p ) // {{{
{
	struct ri_key *result = p;

	if( result->head.id == ID_RI_KEY ) {
		if( gverbose ) parse_ri_struct ( hdesc, result );
		return result;
	}
	else {
		DIE("Invalid RI_KEY record at 0x%lx\n", get_file_offset( hdesc, p ) );
		return NULL;
	}
}
// }}}
struct li_key* check_li_ptr( struct hive *hdesc, void *p ) // {{{
{
	struct li_key *result = p;

	if( result->head.id == ID_LI_KEY ) {
		if( gverbose )parse_li_struct( hdesc, result );
		return result;
	}
	else
	{
		DIE("Invalid LI_KEY record at 0x%lx\n", get_file_offset( hdesc, p ) );
		return NULL;
	}
}
// }}}
int hbinp_is_valid( struct hbin_page *p ) // {{{
{
	return p->id == ID_HBIN;
} // }}}
struct hbin_page* check_hbin_page( struct hive *hdesc, void *p ) // {{{
{
	struct hbin_page *result = p;

	if( hbinp_is_valid( p ) )
		return result;
	else
	{
		DIE("Invalid HBIN_PAGE record at F:%#lx\n");
		return NULL;
	}
}
// }}}
struct hbinp_stats // {{{
{
	int used;
	int free;
}; // }}}
void hive_scan( struct hive *hdesc ) // {{{ doesn't do much right now
{
	static long lastfree = -1;
	struct hbin_page *p = hive_get_first_hbinp( hdesc );
	struct hbinp_stats stats;
	int cnt=0;
	int pf;
	long free=0;
	long used=0;
	DB( DB_INFO, "hive_scan() starting...");

	DB( DB_INFO, "Checking hbin layout sanity...");
	while( p ) {
		/*
		fprintf(stderr, "Page: %d 0x%lx/0x%lx %lx\n", cnt,
				get_heap_offset( hdesc, p ),
				hdesc->head->lastblock,
				p->ofs_next);
		parse_hbin_page( hdesc, p );
		*/
		ASSERT( 0 <= (pf = hbinp_scan( hdesc, p, &stats )), "Broken page F:0x%#x",
				get_file_offset( hdesc, p ) );

		used += stats.used;
		free += stats.free;
		p = hbinp_get_next_hbinp( hdesc, p );
		cnt++;
	}
	DB( DB_INFO, "hive_scan() done.");
	DB( DB_INFO, "Total Used:   % #10x", used);
	DB( DB_INFO, "Total Free:   % #10x", free );
	DB( DB_INFO, "Total:        % #10x", free + used );
	DB( DB_INFO, "Percent Free: % 10.2f", (100*(float)free)/(float)(free+used) );
	if( lastfree >= 0 ) {
		DB( DB_INFO, "Free Delta:   %#10x", lastfree - free );
	}

	lastfree = free;

} // }}}
int hbinp_scan( struct hive *hdesc, struct hbin_page *hbp, struct hbinp_stats *stats ) // {{{ returns: -1: broken, 0...: free space (it currently doesn't return -1, the program just dies)
{
	struct hbin_page *p = hbp;

	bzero( stats, sizeof( struct hbinp_stats ) );
	ASSERT( hbinp_is_valid( p ), "Page is not valid.");

	//struct hbin_page *p2 = hbinp_get_next_hbinp( hdesc, p );
	struct hbin_block *b = hbinp_first_hbinb( p );
	int r=0;
	int cnt=0;

	//ASSERT( p2, "wtf? %lx", p->ofs_next);

	do {
		int size = hbinb_size( b );
		cnt++;
//			WARN("CHECK...");
		ASSERT( (char*)p + p->ofs_next >= (char*)b + size, "Block extends into next page %d F:0x%08lx",
				cnt,
				get_file_offset( hdesc, b ) ); 

		if( hbinb_is_allocd( b ) ) 
			stats->used += hbinb_size( b );
		else
			stats->free += hbinb_size( b );
#if 1
		//fprintf(stderr, "%c", (hbinb_is_allocd( b ))?'o':'.');
		/*
		WARN("Wee: %p/%c %x",
				b, (hbinb_is_allocd( b ))?'y':'n',
				size);
				*/
		/*
		(char*)b + size, (char*)p2,
				(char*)p2 - (char*)b + size 
				);
				*/
#endif

		if( !hbinb_is_allocd( b ) ) r+= hbinb_size( b );
		if( (char*)p + p->ofs_next == (char*)b + size ) {
			//block terminates on start of next page, is good
			return r;
		}

		b = hbinb_get_next( hdesc, p, b );
	}
	while( b );

	DIE("How did I get here?");

} // }}}

// }}}

// {{{ *** Functions that dump structs and things to nice hex and ascii (called parse_* for some reason)
/* The following routines are mostly for debugging, I used it
 * much during discovery. the -t command line option uses it,
 * also the 'st' and 's' from the editor & hexdebugger.
 * All offsets shown in these are unadjusted (ie you must add
 * headerpage (most often REGF_HEADER_SIZE) to get file offset)
 */

//TODO: I don't like this macro
#define D_OFFS(o) ( (void *)&(key->o)-(void *)key )
int parse_vk_name( struct hive *hdesc, struct nk_key *nk, int voff ) // {{{ Verbosely print the name of a value key
{
	int i;
	struct vk_key *vk = nk_get_value( hdesc, nk, voff );
	for(i = 0; i < vk->len_name; i++) putc(vk->keyname[i], stderr);
	fprintf(stderr, "\n  ");
	return 0;
} // }}}
int parse_nk_nice( struct hive *hdesc, struct nk_key *nk ) // {{{
{
	int i=0,k=0;
	fprintf(stderr, "=== NK Key Record ===\n");

	parse_nk_struct( hdesc, nk );

	if( nk->no_subkeys == 0 ) {
		fprintf(stderr, "--NO SUBKEYS--\n");
	}
	else {
		fprintf(stderr, "--Subkeys (%ld)--\n  ", nk->no_subkeys );

		for( k=0; k<nk->no_subkeys; k++)
		{
			struct nk_key *key = nk_get_subkey( hdesc, nk, k );
			for(i = 0; i < key->len_name; i++) putc(key->keyname[i], stderr);
			fprintf(stderr, "\n  ");
			fflush(stderr);
		}
	}



	if( nk->no_values == 0 ) {
		fprintf( stderr, "\n--NO VALUES--\n");
	}
	else {
		fprintf(stderr, "\n--Values--\n  " );
		for( k=0; k<nk->no_values; k++) {

			parse_vk_name( hdesc, nk, k );
		}
	}


	fprintf(stderr, "=== END NK Key Record ===\n");

	return 0;
}
// }}}
void parse_nk_struct( struct hive *hdesc, struct nk_key *key ) // {{{
{
	int i;

	if( key->id != ID_NK_KEY ) raise( SIGSEGV );


	fprintf(stderr, "== nk at offset 0x%0x\n", (void*)key - (void*)hdesc->buf);

#define D_OFFSB(o) ( (void *)&(key->o)-(void *)key )

	fprintf(stderr, "%04x   type              = 0x%02x %s\n", D_OFFSB(type)  , key->type,
				   (key->type == KEY_ROOT ? "ROOT_KEY" : "") );
	fprintf(stderr, "%04x   timestamp skipped\n", D_OFFSB(timestamp) );
	fprintf(stderr, "%04x   parent key offset = 0x%0lx\n", D_OFFSB(ofs_parent) ,key->ofs_parent);
	fprintf(stderr, "%04x   no_subkeys        = %ld\n", D_OFFSB(no_subkeys),key->no_subkeys);
	fprintf(stderr, "%04x   lf-record offset  = 0x%0lx\n",D_OFFSB(ofs_lf),key->ofs_lf);
	fprintf(stderr, "%04x   number of values  = %ld\n", D_OFFSB(no_values),key->no_values);
	fprintf(stderr, "%04x   val-list offset   = 0x%0lx\n",D_OFFSB(ofs_vallist),key->ofs_vallist);
	fprintf(stderr, "%04x   sk-record offset  = 0x%0lx\n",D_OFFSB(ofs_sk),key->ofs_sk);
	fprintf(stderr, "%04x   classname offset  = 0x%0lx\n",D_OFFSB(ofs_classnam),key->ofs_classnam);
	fprintf(stderr, "%04x   *unused?*         = 0x%0lx\n",D_OFFSB(dummy4),key->dummy4);
	fprintf(stderr, "%04x   name length       = %d\n", D_OFFSB(len_name),key->len_name);
	fprintf(stderr, "%04x   classname length  = %d\n", D_OFFSB(len_classnam),key->len_classnam);

	fprintf(stderr, "%04x   Key name: <",D_OFFSB(keyname) );
	for(i = 0; i < key->len_name; i++) putc(key->keyname[i], stderr);
	fprintf(stderr, ">\n== End of key info.\n");

}
// }}}
void parse_vk_struct(struct hive *hdesc, struct vk_key *vk ) /* {{{ Parse the vk datablock */
{
	int i;

#define D_VOFFS(o) ( (void *)&(vk->o)-(void *)vk )

	fprintf( stderr, "== vk at offset %0lx\n", get_file_offset( hdesc, vk ) );

	fprintf( stderr, "%04x   name length       = %d (0x%0x)\n", D_VOFFS(len_name),
				     vk->len_name, vk->len_name  );
	fprintf( stderr, "%04x   length of data    = %ld (0x%0lx)\n", D_VOFFS(len_data),
				     vk->len_data, vk->len_data  );
	fprintf( stderr, "%04x   data offset       = 0x%0lx\n",D_VOFFS(ofs_data),vk->ofs_data);
	fprintf( stderr, "%04x   value type        = 0x%0lx  %s\n", D_VOFFS(val_type), vk->val_type,
		 (vk->val_type <= REG_MAX ? val_types[vk->val_type] : "(unknown)") ) ;

	fprintf( stderr, "%04x   flag              = 0x%0x\n",D_VOFFS(flag),vk->flag);
	fprintf( stderr, "%04x   *unused?*         = 0x%0x\n",D_VOFFS(dummy1),vk->dummy1);

	fprintf( stderr, "%04x   Key name: <",D_VOFFS(keyname) );
	for(i = 0; i < vk->len_name; i++) putc(vk->keyname[i], stderr );
	fprintf( stderr, ">\n== End of key info.\n");

}
// }}}
void parse_vlist(struct hive *hdesc, struct v_list *vl, int no) // {{{
{
	int i;

	fprintf( stderr, "== vlist F:%0lx\n", get_file_offset( hdesc, vl ) );
	for(i=0;i<no;i++) {
		fprintf(stderr, "  F:%lx\n", vl[i] );
	}
} // }}}
void parse_sk_struct(struct hive *hdesc, struct sk_key *sk) // {{{
{
#define D_OFFSC(o) ( (void *)&(sk->o)-(void *)sk )

	fprintf( stderr, "== sk at offset %0lx\n", get_heap_offset( hdesc, sk ) );

	fprintf( stderr, "%04x   *unused?*         = %d\n"   , D_OFFSC(dummy1),     sk->dummy1    );
	fprintf( stderr, "%04x   Offset to prev sk = 0x%0lx\n", D_OFFSC(ofs_prevsk), sk->ofs_prevsk);
	fprintf( stderr, "%04x   Offset to next sk = 0x%0lx\n", D_OFFSC(ofs_nextsk), sk->ofs_nextsk);
	fprintf( stderr, "%04x   Usage counter     = %ld (0x%0lx)\n", D_OFFSC(no_usage),
						    sk->no_usage,sk->no_usage);
	fprintf( stderr, "%04x   Security data len = %ld (0x%0lx)\n", D_OFFSC(len_sk),
						    sk->len_sk,sk->len_sk);

	fprintf( stderr, "== End of key info.\n");
} // }}}
void parse_lx_struct( struct hive *hdesc, lx_key *lxkey ) // {{{
{
	switch( lxkey->head.id ) {
		case ID_LI_KEY:
			parse_li_struct( hdesc, &lxkey->likey );
			break;
		case ID_LF_KEY:
			parse_lf_struct( hdesc, &lxkey->lfkey );
			break;
		case ID_LH_KEY:
			parse_lh_struct( hdesc, &lxkey->lhkey );
			break;
		case ID_RI_KEY:
			parse_ri_struct( hdesc, &lxkey->rikey );
			break;
		default:
			DIE("parse_lx_struct(): Unknown ID 0x%x (%s)\n", lxkey->head.id, id_to_str( lxkey->head.id ) );
	}
}
// }}}
void parse_lf_struct( struct hive *hdesc, struct lf_key *key ) // {{{ Parse the lf datablock (>4.0 'nk' offsets lookuptable)
{
	int i;
	VBUF_STATIC( vbhex, 10 );

	fprintf(stderr, "== lf at offset %0lx\n", get_file_offset( hdesc, key ));

	fprintf(stderr, "%04x   number of keys    = %d\n", D_OFFS(head.no), key->head.no  );


	for(i = 0; i < key->head.no; i++) {

		vbuf_hexdump( vbhex, key->hash[i].name, 0, 4, 1 );
		fprintf(stderr, "%04x      %3d   Offset: 0x%08lx  - <%c%c%c%c>:%s\n", 
			D_OFFS(hash[i].ofs_nk), i,
			key->hash[i].ofs_nk,
			key->hash[i].name[0],
			key->hash[i].name[1],
			key->hash[i].name[2],
			key->hash[i].name[3],
			vbhex->b);
	}

	fprintf(stderr, "== End of key info.\n");
}
// }}}
void parse_lf_names( struct hive *hdesc, struct lf_key *key ) // {{{ Parse the lf datablock (>4.0 'nk' offsets lookuptable)
{
	int i;
	struct nk_key *nk;

	VBUF_STATIC( raw, 10 );
	VBUF_STATIC( hd, 10 );

	fprintf(stderr, "== lf at offset %0lx\n", get_file_offset( hdesc, key ));

	fprintf(stderr, "%04x   number of keys    = %d\n", D_OFFS(head.no), key->head.no  );

	for(i = 0; i < key->head.no; i++) {
		nk = lx_get_subkey( hdesc, (lx_key*)key, i );
		nk_vb_get_rawname( nk, raw );
		vbuf_hexdump( hd, raw->b, 0, raw->dlen, 1 );
		fprintf(stderr, "%04x      %3d   Offset: 0x%08lx  - <%c%c%c%c> %s\n", 
			D_OFFS(hash[i].ofs_nk), i,
			key->hash[i].ofs_nk,
			key->hash[i].name[0],
			key->hash[i].name[1],
			key->hash[i].name[2],
			key->hash[i].name[3],
			hd->b );

	}

	fprintf(stderr, "== End of key info.\n");
}
// }}}
void parse_lh_struct( struct hive *hdesc, struct lh_key *lh ) // {{{ Parse the lh datablock (WinXP offsets lookuptable)
{
	/* 
	* The hash is most likely a base 37 conversion of the name string
	*/

	int i;
	fprintf(stderr, "== lh at offset %0lx\n", get_heap_offset( hdesc, lh ) );

	fprintf(stderr, "%04x   number of keys    = %d\n", (void*)&lh->head.no - (void*)lh, lh->head.no  );

	for(i = 0; i < lh->head.no; i++) {
		fprintf(stderr, "%04x      %3d   Offset: 0x%0lx  - <hash: %08lx>\n", 
		(void*)&hash[i].ofs_nk - (void*)lh,
		i,
		lh->hash[i].ofs_nk,
		lh->hash[i].hash );
	}

	fprintf(stderr, "== End of key info.\n");
}
// }}}
void parse_li_struct(struct hive *hdesc, struct li_key *key) // {{{
{
	int i;

	fprintf(stderr, "== li at offset 0x%lx\n", get_file_offset( hdesc, key ) );

#define D_OFFS(o) ( (void *)&(key->o)-(void *)key )

	fprintf(stderr, "%04x   number of keys    = %d\n", D_OFFS(head.no), key->head.no  );

	for(i = 0; i < key->head.no; i++) {
		fprintf(stderr, "%04x      %3d   nk Offset: 0x%0lx\n", 
			D_OFFS(hash[i].ofs_nk), i,
			key->hash[i].ofs_nk);
	}
	fprintf(stderr, "== End of key info.\n");

}
// }}}
void parse_ri_struct(struct hive *hdesc, struct ri_key *key) // {{{
{
	int i;

	fprintf(stderr, "== ri at offset 0x%lx\n", get_file_offset( hdesc, key ) );

#define D_OFFS(o) ( (void *)&(key->o)-(void *)key )

	fprintf(stderr, "%04x   number of ri ptrs    = %d\n", D_OFFS(head.no), key->head.no  );

	for(i = 0; i < key->head.no; i++) {
		fprintf(stderr, "%04x      %3d   li Offset: 0x%0lx\n", 
			D_OFFS(hash[i]), i,
			key->hash[i]);
	}
	fprintf(stderr, "== End of key info.\n");

}
// }}}
void parse_regf_header( struct hive *hdesc, struct regf_header *head ) // {{{
{
	fprintf( stderr, "== regf_header\n");
	fprintf( stderr, "id(0x66676572):         0x%lx\n", head->id );
	fprintf( stderr, "unknown1:               0x%lx\n", head->unknown1 );
	fprintf( stderr, "unknown2(0):            0x%lx\n", head->unknown2 );
	fprintf( stderr, "timestamp:              skipping\n" );
	fprintf( stderr, "unknown3(1):            0x%lx\n", head->unknown3 );
	fprintf( stderr, "unknown4(version?, 3):  0x%lx\n", head->unknown4 );
	fprintf( stderr, "unknown5(0):            0x%lx\n", head->unknown5 );
	fprintf( stderr, "unknown6(1):            0x%lx\n", head->unknown6 );
	fprintf( stderr, "ofs_rootkey:            0x%lx\n", head->ofs_rootkey );
	fprintf( stderr, "lastblock:              0x%lx\n", head->lastblock );
	fprintf( stderr, "unknown7(1):            0x%lx\n", head->unknown7 );
	fprintf( stderr, "name:                   %s\n", regf_get_name( head ) );
	fprintf( stderr, "checksum:               0x%lx\n", head->checksum );
}
// }}}

char *regf_get_name( struct regf_header *head ) // {{{
{
	VBUF_STATIC( vb, 10 );
	VBUF_STATIC( r, 10 );
	VBUF_STATIC( utf8buf, 10 );

	vbuf_set( vb, head->name, sizeof( head->name ) );

	if( -1 == vb_utf16to8( utf8buf, vb->b, 0x40 ) ) return NULL;

	return utf8buf->b;
} // }}}

int hbinbh_is_allocd( struct hbb_head *h ) // {{{
{
	if (h->size < 0)
		return 1;
	else
		return 0;
} // }}}
long hbinbh_size( struct hbb_head *h ) // {{{ //retrieve the size of hbin block
{
	if (h->size < 0)
		return -h->size;
	else
		return h->size;
} // }}}
int hbinb_is_allocd( struct hbin_block *p ) // {{{
{
	if (p->size < 0)
		return 1;
	else
		return 0;
} // }}}
long hbinb_size( struct hbin_block *p ) // {{{ //retrieve the size of hbin block
{
	if (p->size < 0)
		return -p->size;
	else
		return p->size;
} // }}}
long hbinb_dlen( struct hbin_block *p ) // {{{ //retrieve length of data space in hbin block
{
	return hbinb_size( p ) - sizeof( struct hbb_head );
} // }}}
void parse_hbin_head_short( struct hive *hdesc, struct hbin_block *p ) // {{{
{
	fprintf( stderr, "== HBIN Block (%s) ", display_ptr_offset( hdesc, p ) );
	fprintf( stderr, "%s ", hbinb_is_allocd( p )?"A":"x" );
	fprintf( stderr, "Size: (%ld/0x%lx)\n", hbinb_size( p ), hbinb_size( p ) );
	//fprintf( stderr, "Contains:               %s\n", id_to_str( *(unsigned short*)&p->data ) );
} // }}}
void hbinbh_vdump( vbuf *vb, struct hive *hdesc, struct hbb_head *hbbh ) // {{{
{
	vbuf_printf( vb, "== HBIN Block (%s)\n", display_ptr_offset( hdesc, hbbh ) );
	vbuf_printfa( vb, "Allocated:              %d\n", hbinbh_is_allocd( hbbh ) );
	vbuf_printfa( vb, "Size:                   (%ld/0x%lx)\n", hbinbh_size( hbbh ), hbinbh_size( hbbh ) );
} // }}}
void hbinb_vdump( vbuf *vb, struct hive *hdesc, struct hbin_block *b ) // {{{
{
	hbinbh_vdump( vb, hdesc, (struct hbb_head*)b );

	vbuf_printfa( vb, "Contains:               %s\n", id_to_str( *(unsigned short*)&b->data ) );
	vbuf_hexdump( vb, b->data, 0, hbinb_dlen( b ), 1 );
} // }}}
void parse_hbin_head( struct hive *hdesc, struct hbin_block *p ) // {{{
{
	fprintf( stderr, "== HBIN Block (%s)\n", display_ptr_offset( hdesc, p ) );
	fprintf( stderr, "Allocated:              %d\n", hbinb_is_allocd( p ) );
	fprintf( stderr, "Size:                   (%ld/0x%lx)\n", hbinb_size( p ), hbinb_size( p ) );
	fprintf( stderr, "Contains:               %s\n", id_to_str( *(unsigned short*)&p->data ) );
} // }}}
void parse_hbin_block( struct hive *hdesc, struct hbin_block *p ) // {{{
{
	parse_hbin_head( hdesc, p );
	if( hbinb_size( p ) > 100000 ) 
		fprintf(stderr, "suspect block, not printing.\n");
	else 
		switch( *(unsigned short*)&p->data ) { 
			case ID_VK_KEY:
				parse_vk_struct( hdesc, (void*)p->data );
				break;
			default:
				hexdump( p->data, 0, hbinb_dlen( p ), 1 );
		}
}
// }}}
void parse_hbin_page( struct hive *hdesc, struct hbin_page *p ) // {{{
{
	VBUF_STATIC( hexbuf, 10 );
	vbuf_hexdump( hexbuf, p->dummy1, 0, 14, 1 );
	fprintf( stderr, "== HBIN Page Header (at HEAP:0x%lx)\n", get_heap_offset( hdesc, p ) );
	fprintf( stderr, "id(0x6e696268):         0x%lx\n", p->id );
	fprintf( stderr, "ofs_from1:              0x%lx\n", p->ofs_from1 );
	fprintf( stderr, "ofs_next:               0x%lx\n", p->ofs_next );
	fprintf( stderr, "dummy1:                 %s\n", hexbuf->b );
	fprintf( stderr, "len_page:               0x%lx\n", p->len_page );
}
// }}}
void parse_hbin_page_long( struct hive *hdesc, struct hbin_page *p ) // {{{
{
	struct hbin_block *hb = &p->start_block;

	parse_hbin_page( hdesc, p );

	do {
		parse_block_struct( hdesc, hb );
	}while( ( hb = hbinb_get_next( hdesc, p , hb ) ) );
} // }}}
/* The page header, I don't know if the 14 "dummy" bytes has a meaning,
 * they seem to be mostly NULLS
 */

int parse_block_struct(struct hive *hdesc, struct hbin_block *block) // {{{
{
	unsigned short id;

	fprintf(stderr, "== HBIN Block (F:0x%0lx) ", get_file_offset( hdesc, block ) );
	//fprintf(stderr, "block->size: %ld, %lu, 0x%0lx\n",block->size,block->size,block->size);
	fprintf(stderr, "size: (%ld/0x%lx) allocd: %s ", 
	hbinb_size( block ),
	hbinb_size( block ),
	hbinb_is_allocd( block )?"y":"n" );

	/*
	if (block->size == 0) {
		fprintf(stderr,
			"ACK, should not be: hbin_block->size == 0\n");
		//debugit(hdesc->buf,hdesc->size);
		return(0);
	}
  
	if (block->size < 0) {
		//block->size = -block->size;
		//hdesc->usetot += block->size;
		//hdesc->useblk++;
		fprintf(stderr, "USED BLOCK: size: %ld, 0x%0lx\n", -block->size, -block->size);
	      //hexdump(hdesc->buffer,vofs,vofs+block->size+4,1); 
	} else {
		//hdesc->unusetot += block->size;
		//hdesc->unuseblk++;
		//bzero(block->data, block->size-4);

		fprintf(stderr, "FREE BLOCK, size: %ld, 0x%lx\n", block->size, block->size ); 
	}
	*/

	id = *(unsigned short*)&block->data;
	fprintf( stderr, "contains: %s\n", id_to_str( id ) );

	return(block->size);
}
// }}}
// }}}

// {{{ **** HBIN access and manipulation functions
struct hbin_page *find_page_start(struct hive *hdesc, void *p) // {{{ Find page containing p
{
	/* hdesc = hive
	* 
	* returns: offset to start of page (and page header)
	*/
	struct hbin_page *h;
	//TODO: this is way slow

	//try the quick way first
	h = get_ptr_off_h( hdesc, get_heap_offset( hdesc, p ) & 0xFFFFF000 );
	if( hbinp_is_valid( h ) )  {
		ASSERT( (char*)p > (char*)h && (char*)p < (char*)h + h->ofs_next, "Impossible." );
		return h;
	}
	h = NULL;
	while( ( h = hbinp_get_next_hbinp( hdesc, h ) ) ) {
		if( p > (void*)h && p < (void*)h + h->ofs_next ) return h;
	}

	DIE("Couldn't find page containing pointer %p", p );

	return NULL;
}
// }}}
int hbinp_set_dirty( struct hive *hdesc, struct hbin_page *p ) // {{{ Mark a page as dirty
{
	//we don't do dirty pages yet, we just do dirty hives...
	hdesc->state |= HMODE_DIRTY;

	return 0;
} // }}}
int hbinb_merge( struct hive *hdesc, struct hbin_block *b1, struct hbin_block *b2 ) // {{{ Merge two free blocks into one
{
	ASSERT( 0 < b1->size, "hbinb_merge(): block 1 is not free." );
	ASSERT( 0 < b2->size, "hbinb_merge(): block 1 is not free." );

	b1->size+= b2->size;

	hive_avail_catalog( hdesc, find_page_start( hdesc, b1 ), b1 );

	return 0;
} // }}}
int hbinb_free( struct hive *hdesc, struct hbin_block *b ) { // {{{
	/* hdesc - hive
	* blk   - offset of block, MUST POINT TO THE LINKAGE!
	* return: 0 == success, -1 == fail
	* Will CHANGE HIVE IF SUCCESSFUL (changes linkage)
	*/

	struct hbin_block *bprev, *bnext; 
	struct hbin_page *p;

	//TODO: I don't get this, freeing an hbin block will not effect the
	//size of the file.  Isn't HMODE_NOALLOC just about not
	//growing/shrinking the file for ntfs write?
	if (hdesc->state & HMODE_NOALLOC) {
		ERR_UNIX( EACCES, 
		"Hive %s is in no-allocation mode, can't free blocks."
		, hdesc->filename);
		return -1;
	}

	//  size = get_int(hdesc->buf + b);
	//size = b->size;
	ASSERT(b->size < 0, "hbinb_free(): trying to free already free block!");

	if( hdesc->alloc_max && hbinb_size( b ) > hdesc->alloc_max ) {
		hdesc->alloc_max = hbinb_size( b );
	}
	b->size = -b->size;

	p = find_page_start(hdesc, b);

	bprev = hbinb_get_prev( hdesc, p, b );
	bnext = hbinb_get_next( hdesc, p, b );

	if( bnext && !hbinb_is_allocd( bnext ) ) { hbinb_merge( hdesc, b, bnext ); }
	if( bprev && !hbinb_is_allocd( bprev ) ) { hbinb_merge( hdesc, bprev, b ); }

	hbinp_set_dirty( hdesc, p );

	hive_avail_catalog( hdesc, p, b );

	return 0;
} // }}}
struct hbin_block* hbinp_first_hbinb( struct hbin_page *p ) // {{{
{
	return (struct hbin_block*)&p->start_block;
} // }}}
struct hbin_block* hbinb_get_prev( struct hive *hdesc, struct hbin_page *p, struct hbin_block *b ) // {{{ find the hbin block that comes before b or NULL if there isn't one
{
	struct hbin_block *result, *n;

	ASSERT( ( (void*)b > (void*)p ) && ( (void*)p + p->ofs_next > (void*)b ),
			"hbinb_get_prev() called with block outside of given page.");

	result = hbinp_first_hbinb( p );

	if( result == b ) return NULL;

	while( NULL != (n = hbinb_get_next( hdesc, p, result ) ) ) {
		if( n == b ) return result;
		ASSERT( b > n, "We seem to have missed the block we are looking for.");
		result = n;
	}

	//We should not be here ever
	DIE("hbinb_get_prev(): Ran out of blocks in page looking for block that must be in page.");

	return NULL;
} // }}}
struct hbin_block* hbinb_get_next( struct hive *hdesc, struct hbin_page *p, struct hbin_block *b ) // {{{ find the hbin block that comes after b || NULL 
{
	struct hbin_block *r = (void*)b + hbinb_size( b );
	if( (void*)r - (void*)p >= p->ofs_next ) return NULL;
	return r;
} // }}}
struct hbin_page* hive_get_last_hbinp( struct hive *hdesc ) // {{{
{
	return check_hbin_page( hdesc, get_ptr_off_f( hdesc, hdesc->head->lastblock ) );
} // }}}
struct hbin_page* hive_get_first_hbinp( struct hive *hdesc ) // {{{
{
	return check_hbin_page( hdesc, get_ptr_off_h( hdesc, 0 ) );
} // }}}
struct hbin_block* hbinp_find_free_hbinb( struct hive *hdesc, struct hbin_page *p, int size ) // {{{ Find free hbin block in a page (or NULL)
{
	struct hbin_block *b = hbinp_first_hbinb( p );
	int csize = 0;
	struct hbin_block *sb = NULL;

	do {
		if( !hbinb_is_allocd( b ) ) {
			if( hbinb_size( b ) >= size ) {
				if( 0 == csize || hbinb_size( b ) < csize ) {
					return sb=b;
				}
			}
			else
				hive_avail_catalog( hdesc, p, b );
		}
	} while( ( b = hbinb_get_next( hdesc, p, b ) ) ); 

	return sb;
} // }}}
struct hbin_page* hbinp_get_next_hbinp( struct hive *hdesc, struct hbin_page *p ) // {{{ returns page following p if it exists, NULL otherwise
{
	void *next;
	if( NULL == p ) { return hive_get_first_hbinp( hdesc ); }
	next = (void*)p + p->ofs_next;
	if( next > (void*)hdesc->buf + hdesc->head->lastblock ) return NULL;
	struct hbin_page *r = check_hbin_page( hdesc, next );
	if( p->ofs_next >= hdesc->head->lastblock ) DIE("hbinp_get_next_hbinp(): already on the last page"); // return NULL;
	return r;
} // }}}
struct hbin_block* hive_avail_seek( struct hive *hdesc, int size ) // {{{
{
	struct hbin_block *r = NULL;
	int i=0;
	if( size >= HBB_AVAIL_MAX )
		return NULL;
	for( i=size; i<HBB_AVAIL_MAX; i++ ) {
		struct hbin_page **hbp = &(hdesc->hbb_avail[i]);

		if( *hbp ) {
			r = hbinp_find_free_hbinb( hdesc, *hbp, size );

			if(r) 
				return r;
			else
				*hbp = NULL;
		}
	}

	return NULL;
} // }}}

void hive_avail_catalog( struct hive *hdesc, struct hbin_page *p, struct hbin_block *b ) // {{{
{
	int s = hbinb_size( b );

	if( s < HBB_AVAIL_MAX && !hbinb_is_allocd( b ) ) {
		if(!hdesc->hbb_avail[s]) {
			hdesc->hbb_avail[s] = p;
		}
	}
} // }}}
int hbinp_biggest_cont( struct hive *hdesc, struct hbin_page *p ) // {{{
{
	struct hbin_block *b = hbinp_first_hbinb( p );
	int r = 0;

	do {
		if( !hbinb_is_allocd( b ) && hbinb_size( b ) > r )
			r = hbinb_size( b );
	} while( ( b = hbinb_get_next( hdesc, p, b ) ) );

	return r;
} // }}}
int hbinp_total_free( struct hive *hdesc, struct hbin_page *p ) // {{{
{
	struct hbin_block *b = hbinp_first_hbinb( p );
	int r = 0;

	do {
		if( !hbinb_is_allocd( b ) )
			r+= hbinb_size( b );
	} while( ( b = hbinb_get_next( hdesc, p, b ) ) );

	return r;
} // }}}
struct hbin_block* hive_find_free_hbinb( struct hive *hdesc, int size ) // {{{ Search hive for free block or NULL
{
	static struct hbin_page *p = NULL;
	struct hbin_page *po = NULL;
	struct hbin_block *r = NULL;

	//DB( DB_0, "hive_find_free_hbinb(): size %x, alloc_max %x\n", size, hdesc->alloc_max );
	if( hdesc->alloc_max > 0 && size >= hdesc->alloc_max ) {
		//DB(DB_WARN, "Failed to allocate %d bytes before, not looking for free %d bytes.",
		//		hdesc->alloc_max, size );
		ERR_WIN(rlERROR_BROKEN_ERROR,
				"Failed to allocate %d bytes before, not looking for free %d bytes.",
				hdesc->alloc_max, size);
		return NULL;
	}

	if( NULL == p || 4 == size )
		p = hive_get_first_hbinp( hdesc );

	po = p;

	if( r = hive_avail_seek( hdesc, size ) )
		return r;


	//DB( DB_0, "hive_find_free_hbinb() running scan...\n");
	do {
		//DB( DB_0, "Chk page %#8x: %#x\n", get_file_offset( hdesc, p ), hbinp_biggest_cont( hdesc, p ) );
		if( NULL != ( r = hbinp_find_free_hbinb( hdesc, p, size ) ) ) return r;
		p = hbinp_get_next_hbinp( hdesc, p );
		if( NULL == p ) p = hive_get_first_hbinp( hdesc );
	} while( ( p != po ) );


	//DB( DB_0, "alloc_max %x -> %x\n", hdesc->alloc_max, size );
	hdesc->alloc_max = size;

	ERR_WIN(rlERROR_BROKEN_ERROR, "Couldn't find free hbin block > %d in hive.", size);

	return NULL;
} // }}}
void hbinb_split( struct hive *hdesc, struct hbin_block *hbb, int len ) // {{{ Makes one hbin block into two, the first with data space >= len NOTE: does nothing if there is not enough free space for another block
{
	ASSERT( !hbinb_is_allocd( hbb ), "Given hbb is already allocated." );
	ASSERT( hbinb_dlen( hbb ) >= len, "This hbb block is too small.");

	//hive_scan( hdesc );
	//ASSERT( 0 <= hbinp_scan( hdesc, find_page_start(hdesc, hbb ) ), "HBIN PAGE Broken." );

	int new_size = len + sizeof( struct hbb_head );
	if( new_size % 8 )
		new_size += 8 - (new_size % 8);

	if( hbinb_size( hbb ) - new_size >= HBB_MIN_LEN ) {
		//WARN("Attempting split of %x...", hbinb_size( hbb ) );
		struct hbin_block *hbn = (struct hbin_block*)((char*)hbb + new_size);
		int nblk_size = hbinb_size( hbb ) - new_size;
		ASSERT( 0 == nblk_size % 8, "Bad value");
		hbn->size = nblk_size; 
		hbb->size = new_size;

		hive_avail_catalog( hdesc, find_page_start(hdesc, hbb ), hbn );
		//WARN("Split: -> %x, %x", hbinb_size( hbb ), hbinb_size( hbn ) );

		//ASSERT( 0 <= hbinp_scan( hdesc, find_page_start(hdesc, hbb ) ), "HBIN PAGE Broken." );
	}
} // }}}
struct hbin_block *hbin_alloc(struct hive *hdesc, void *ofs, int size) // {{{ Allocate a block of requested size if possible
{
	/* hdesc - hive
	* pofs - If >0 will try this page first (ptr may be inside page)
	* size - number of bytes to allocate
	* returns: 0 - failed, else pointer to allocated block.
	* This function WILL CHANGE THE HIVE (change block linkage) if it
	* succeeds. 
	*/
	//int req_size; //required block data size
	struct hbin_block *blk = NULL;
	int trailsize, oldsz;

	if (hdesc->state & HMODE_NOALLOC) {
		ERR_UNIX( EACCES, 
			"Hive %s is in no-allocation mode, can't allocate blocks."
			, hdesc->filename);
		return NULL;
	}

	/* Check current page first */
	if (ofs) {
		struct hbin_page *p;
		p = find_page_start(hdesc, ofs );
		blk = hbinp_find_free_hbinb( hdesc, p, size + sizeof( struct hbb_head ) );
	}

	/* Then check whole hive */
	if (!blk)
		blk = hive_find_free_hbinb( hdesc, size + sizeof( struct hbb_head ));

	if (blk) {  
		hbinb_split( hdesc, blk, size );
		blk->size = -blk->size;

		hdesc->state |= HMODE_DIRTY;

	} else {
		//DIE("Allocation failure.");
		ERR_UNIX( EACCES,
			"hbin_alloc: failed to alloc %d bytes, and hive expansion not implemented yet!",size);
	}
	return blk;
}
// }}}

// }}}
// {{{ **** NK access and modification functions


// Unsorted

void nk_vb_get_path( vbuf *dest, struct hive *hdesc, struct nk_key *nk ) // {{{ Quick-n-dirty path generator for debugging
{
	VBUF_STATIC( namebuf, 10 );

	vbuf_strset( dest, "" );

	while(1) {
		switch( nk->type ) {
			case KEY_ROOT:
				vbuf_strnprepend( dest, "/", 1 );
				return;
			default:
				nk_vb_get_name( nk, namebuf );

				vbuf_strnprepend( dest, "/", 1 );
				vbuf_strnprepend( dest, namebuf->b, namebuf->dlen );
		}
		nk = nk_get_parent( hdesc, nk );
	}
} // }}}

int nk_get_path(struct hive *hdesc, struct nk_key *nk, char *path, int maxlen) // {{{ traceback - trace nk's back to root, APPENDS TO *path!!!
{
	//There is one of these implemented in nstdreg.c, which would be the
	//one to use.
	//
	//I left this here because it might be useful for something

	/* building path string as we go.
	* nkofs  = offset to nk-node
	* path   = pointer to pathstring-buffer
	* maxlen = max length of path-buffer
	* return: length of path string
	*/
	/* int newnkofs; */
	char tmp[ABSPATHLEN+1];

	maxlen = (maxlen < ABSPATHLEN ? maxlen : ABSPATHLEN);

	if (nk->type == KEY_ROOT) {   /* We're at the root */
		return(strlen(path));
	}

	strncpy(tmp,path,ABSPATHLEN-1);

	if ( (strlen(path) + nk->len_name) >= maxlen-6) {
		snprintf(path,maxlen,"(...)%s",tmp);
		return(strlen(path));   /* Stop trace when string exhausted */
	}
	*path = '\\';
	memcpy(path+1, nk->keyname, nk->len_name);
	strncpy( path+nk->len_name+1, tmp, maxlen);
	return(nk_get_path(hdesc, nk_get_parent(hdesc, nk), path, maxlen)); /* go back one more */
} // }}}

#if 0
int vb_fixname( vbuf *out, char *namein, int len ) // {{{ ascii/utf16 name -> utf8 name (for NK names only (I think))
{
	int r = -1;
	if( len > 1 && namein[len-1] == '\0' ) //It's PROBABLY UTF16
		r = vb_utf16to8( out, namein, len );
	else
		r = vb_iso8859_1toutf8( out, namein, len );

	return r;
}
// }}}
#endif
int nk_scan_subkeys( struct hive *hdesc, struct nk_key *nk ) // {{{
{
	VBUF_STATIC( hr1, 10 );
	VBUF_STATIC( hr2, 10 );
	VBUF_STATIC( pbuf, 10 );
	VBUF_STATIC( lastname, 10 );
	VBUF_STATIC( namebuf, 10 );
	VBUF_STATIC( rawbuf, 10 );
	lx_key *lx;

	vbuf_strset( lastname, "" );

	char lf_hash_buf[4];

	if( nk_get_no_subkeys( nk ) > 0 ) {
		lx = nk_get_lx( hdesc, nk );

		int j;
		for(j = 0; j < lx->head.no; j++) {
			struct nk_key *nks = nk_get_subkey( hdesc, nk, j );
			struct nk_key *nkp = nk_get_parent( hdesc, nks );
			if( nkp != nk ) {
				WARN("NK parent is not real parent.");
			}
			//parse_nk_struct( hdesc, nks );
			nk_vb_get_rawname( nks, rawbuf );
			ASSERT( 0 == nk_vb_get_name( nks, namebuf ), "Error decoding name: %s", strerror( errno ) );
			//WARN("NK: %s", namebuf->b );

			if( 0 < winstrcmp( lastname->b, namebuf->b ) ) {
				WARN("Subkey order mismatch between \"%s\" and \"%s\"", lastname->b, namebuf->b );
			}

			vbuf_set( lastname, namebuf->b, namebuf->dlen );

			switch( lx->head.id ) {
				case ID_LF_KEY:
				{
					//WARN("LF Key...");
					struct lf_key *lf = (struct lf_key*)lx;
					vb_utf16_ensure_zero_termination( rawbuf );
					lf_hash( namebuf->b, lf_hash_buf );
					if( 0 != memcmp( lf->hash[j].name, lf_hash_buf, 4 ) ) {

						nk_vb_get_path( pbuf, hdesc, nk );
						WARN("=========================================");
						WARN("%s", pbuf->b );
						WARN("LF hash check failed:");
						WARN("LF hash check: given hash, calculated hash, name");
						hexdump( lf->hash[j].name, 0, 4, 1 );
						hexdump( lf_hash_buf, 0, 4, 1 );
						hexdump( rawbuf->b, 0, rawbuf->dlen, 1 );
						parse_lf_names( hdesc, lf );

						parse_nk_nice( hdesc, nks );

						WARN("=========================================");
					}
				}
				break;
			case ID_LH_KEY:
				{
					struct lh_key *lh = (struct lh_key*)lx;
					//hexdump( rawbuf->b, 0, rawbuf->dlen, 1 );
					long hash = lh_hash( rawbuf->b );
					//long hash2 = lh_hash2( rawbuf->b );

					lh_hash_dec( hr1, lh->hash[j].hash );
					lh_hash_dec( hr2, hash );

					//WARN("Foo: %s, %08lx == %08lx, %s, %s\n", namebuf->b, lh->hash[j].hash, hash, hr1->b, hr2->b );
					if( lh->hash[j].hash != hash ) {
						parse_nk_nice( hdesc, nks );
						//WARN("diff: %ld", ( lh->hash[j].hash - hash ) );
						//DIE("Hash mismatch on key: >%s< %s != %s", namebuf->b, hr1->b, hr2->b );

						long h = labs(hash - lh->hash[j].hash);
						while( h ) {
							//WARN("H: %ld, %ld", h, h % 37 );
							h = h/37;
						}
						WARN("Hash mismatch on key: >%s< %ld != %ld", namebuf->b, hash, lh->hash[j].hash );
						//WARN("Althash: %ld", hash2 );

						hexdump( rawbuf->b, 0, rawbuf->blen, 1 );

					}


					/*
					if( lh->hash[j].hash != hash2 ) {
						parse_nk_nice( hdesc, nks );
						//WARN("diff: %ld", ( lh->hash[j].hash - hash ) );
						//DIE("Hash mismatch on key: >%s< %s != %s", namebuf->b, hr1->b, hr2->b );

						long h = labs(hash2 - lh->hash[j].hash);
						while( h ) {
							//WARN("H: %ld, %ld", h, h % 37 );
							h = h/37;
						}
						WARN("Hash2 mismatch on key: >%s< %ld != %ld", namebuf->b, hash2, lh->hash[j].hash );
						WARN("Normhash: %ld", hash );

						hexdump( rawbuf->b, 0, rawbuf->blen, 1 );

					}
					*/
				}
				break;
			case ID_LI_KEY:
				break;

			case ID_RI_KEY:
				break;
			default:
				DIE("Unknown lx type: %x", lx->head.id );
			}
		}
	}
} // }}}
int nk_scan( struct hive *hdesc, struct nk_key *nk ) // {{{
{
	VBUF_STATIC( namebuf, 10 );
	VBUF_STATIC( pbuf, 10 );

	int i;

	/*
	if( nk->type == 0x00 ) {
		nk_vb_get_path( pbuf, hdesc, nk );
		WARN("-----------------------------------------------");
		WARN("UNICODE: %s", pbuf->b );
		if( -1 == nk_vb_get_name( nk, namebuf ) ) {
			WARN("Unicode conversion failed.");
		}
		WARN("Name: %s", namebuf->b );
		parse_nk_struct( hdesc, nk );
		WARN("-----------------------------------------------");
	}
	*/

	if( nk->type != 0x00 && nk->type != 0x20 && nk->type != 0x2c ) {
		nk_vb_get_path( pbuf, hdesc, nk );
		WARN("-----------------------------------------------");
		WARN("%s", pbuf->b );
		WARN("Type: %x", nk->type );
		parse_nk_struct( hdesc, nk );
		WARN("-----------------------------------------------");
	}

	nk_scan_subkeys( hdesc, nk );
	

	//parse_nk_nice( hdesc, nk );

	for( i=0; i < nk_get_no_subkeys( nk ); i++ ) {
		nk_scan( hdesc, nk_get_subkey( hdesc, nk, i ) );
	}

	/*
	nk_vb_get_name( nk, namebuf );
	WARN("nk_scan() called on %s", namebuf->b);
	*/
} // }}}
void nk_vb_get_rawname( struct nk_key *nk, vbuf *out ) // {{{ Dumps the name of a NK in "out" verbatim
{
	static char z[] = "";
	vbuf_set( out, nk->keyname, nk->len_name );

	/*
	if( out->dlen > 1 && out->b[out->dlen-1] == '\0' ) {
		WARN("NK raw name Is terminated (and UTF16?).");
		hexdump( out->b, 0, out->dlen, 1 );
	}
	else {
		vbuf_append( out, z, 1 );
	}
	*/
} // }}}
int nk_vb_get_name( struct nk_key *nk, vbuf *out ) // {{{ 
{
	VBUF_STATIC( rawbuf, 10 );
	VBUF_STATIC( termbuf, 10 );
	VBUF_STATIC( rtest, 10 );
	
	nk_vb_get_rawname( nk, rawbuf );


	int r = -1;
	switch( nk->type ) {
		case NK_TYPE_UTF16:
			vbuf_set( termbuf, rawbuf->b, rawbuf->dlen );
			vb_utf16_ensure_zero_termination( termbuf );
			r = vb_utf16to8( out, termbuf->b, termbuf->dlen );
			vb_utf8to16( rtest, out->b, out->dlen );
			if( rtest->dlen - 2 > rawbuf->dlen || 0 != memcmp( rawbuf->b, rtest->b, rawbuf->dlen ) ) {
				WARN("----------------");
				hexdump( rawbuf->b, 0, rawbuf->dlen, 1 );
				hexdump( rtest->b, 0, rtest->dlen, 1 );
				DIE("Reverse check failed on \"%s\".", out->b);
			}
			break;
		case NK_TYPE_ISO8859:
			vb_utf8_ensure_zero_termination( rawbuf );
			r = vb_iso8859_1toutf8( out, rawbuf->b, rawbuf->dlen );
			break;
		default:
			DIE("Unknown NK type: %x", nk->type );
	}

	return r;
}
// }}}
#if 0
int nk_get_name( struct nk_key *nk, char *buf, int n) // {{{ 
{
	int sz = (nk->len_name < n - 1 )? nk->len_name:n - 1;
	//fprintf(stderr, "nk_get_name() starting: n = %d\n", n );
	memcpy( buf, nk->keyname, sz );
	//fprintf(stderr, "nk_get_name() raw:\n");
	//hexdump( buf, 0, sz, 1 );
	buf[sz] = '\0';
	sz = fixname( buf, sz, n );
	//fprintf(stderr, "nk_get_name(0 returning: sz = %d, n = %d\n", sz, n );

	if( sz > n ) DIE("buffer is too small." );

	buf[sz] = '\0';
	if( nk->len_name != sz && gverbose )
		fprintf(stderr, "nk_get_name() oddball: %d -> %d: %s\n", nk->len_name, sz, buf );

	//fprintf(stderr, "nk_get_name() fixed:\n");
	//hexdump( buf, 0, sz, 1 );

	//if( gverbose )
	//fprintf(stderr, "nk_get_name() returning from %p, \"%s\", sz=%d\n", nk->keyname, buf, sz );

	if( sz < nk->len_name) DIE( "nk_get_name(): I was given a buffer that is too small :(" );
	return sz;
} // }}}
#endif
int nk_get_classname( struct hive *hdesc, struct nk_key *nk, char *buf, int n) // {{{
{
	int sz = (nk->len_classnam < n - 1 )? nk->len_classnam:n - 1;
	char *p = get_ptr_off_h( hdesc, nk->ofs_classnam );

	memcpy( buf,  p, sz );
	buf[sz] = '\0';
	//fixname( buf, sz, n );

	if( gverbose ) fprintf(stderr, "nt_get_classname() returning \"%s\"\n", buf );

	return sz;
}
// }}}
long nk_get_class_len( struct nk_key *nk ) { // {{{
	return nk->len_classnam + 1;
}
// }}}
long nk_get_name_len( struct nk_key *nk ) // {{{
{
	//since some names are stored in ASCII and will become larger when
	//converted to UTF8 we need to convert the name to find the length :P

	VBUF_STATIC( buf, 10 );
	return nk_vb_get_name( nk, buf );
} // }}}
long nk_get_no_subkeys( struct nk_key *nk ) { // {{{
	return nk->no_subkeys;
}
// }}}
struct hbin_block *nk_get_hbin_block( struct nk_key *nk ) // {{{
{
	//yeah, this sucks (the '4' is the size integer, look at hbin_block struct)
	return (struct hbin_block*)((char*)nk - 4);
}
// }}}
struct nk_key* nk_get_parent( struct hive *hdesc, struct nk_key *nk ) // {{{
{
	struct hbin_block *hb = get_hbin_block( hdesc, get_ptr_off_h( hdesc, nk->ofs_parent ) );
	return check_nk_ptr( hdesc, hb->data );
}
// }}}

// Access functions

void lf_hash( const char *name, char* buf ) // {{{ in: name == UTF-8, out: buf == Generate an lf-style hash
{
	VBUF_STATIC( isobuf, 10 );

	memset( buf, 0, 4 );

	//WARN("lf_hash: %s", name );
	//hexdump( name, 0, strlen( name ) + 1, 1 );
	int r = vb_utf8toiso8859_1( isobuf, name, strlen( name ) + 1 );
	if( -1 == r ) {
		WARN("non-ascii characters, hash is zero (should be ok, this is what Windows seems to do): %s", strerror( errno ) );
		memset( buf, 0, 4 );
		//return;
	}
	int sz = isobuf->dlen - 1;
	if( sz > 4 ) sz = 4;
	//WARN("lf_hash returns:");
	strncpy( buf, isobuf->b, sz );
	//hexdump( buf, 0, 4, 1 );
}
// }}}
void lh_hash_dec( vbuf *out, unsigned long hash ) // {{{ Display lh_hash in base 37
{
	vbuf_strset( out, "" );
	VBUF_STATIC( nbuf, 10 );

	while( hash ) {
		vbuf_printf( nbuf, "%02ld ", (hash % 37) );
		vbuf_strnprepend( out, nbuf->b, nbuf->dlen );
		hash = hash/37;
	}
} // }}}
unsigned long lh_hash_old( const char *name ) //{{{ Generate an lh-style hash (broken on some values)
{
	int i=0;
	long hash=0;
	for (i = 0; i < strlen(name); i++) {

		hash *= 37;
		if( toupper( name[i] ) < 0xe0 )
			hash += toupper(name[i]);

		if( toupper(name[i]) > 127 )
			WARN("xx: %x %x", (unsigned int)name[i], toupper( name[i] ) );
	}

	//fprintf(stderr, "lh_hash: %s -> %ld\n", name, hash);
	return hash;
}
// }}}
unsigned long lh_hash( const char *name ) //{{{ Generate an lh-style hash (testing)
{
	//TODO: I'm not sure about 0xe0
	int i=0;
	long hash=0;
	for (i = 0; i < strlen(name); i++) {

		hash *= 37;
		if( toupper( name[i] ) < 0xe0 )
			hash += toupper(name[i]);

		if( toupper(name[i]) >= 0xe0 )
			hash += toupper(name[i]) & 0xdf;
	}

	//fprintf(stderr, "lh_hash: %s -> %ld\n", name, hash);
	return hash;
}
// }}}
#if 0
long lf_findkey( struct hive *hdesc, long no_subkeys, lx_key *lxkey, const char *name ) //{{{ returns: offset | -1
{
	//Is this used anymore?  Should be using lx_
	char hash[4];
	char buf[SZ_MAX];
	long min=0, max=no_subkeys, i, chkval;

	lf_hash( name, hash );

	//fprintf(stderr, "\n** ENTER: lf_findkey(%s): THIS IS SOMEWHAT UNTESTED CODE (hash=%c%c%c%c)\n", name, hash[0], hash[1], hash[2], hash[3]);

	if( no_subkeys <= 0 ) {
		return -1;
	}

	//parse_lf_struct( hdesc, &(lxkey->lfkey) );

	while( 1 ) { //binary search for *any* matching hash
		chkval = min + (max - min)/2;
		int c = strncasecmp( hash, lxkey->lfkey.hash[chkval].name, 4 );
		//fprintf(stderr, "Weee: min:%ld max:%ld c:%d hash:%s lf-hash:%s name:%s\n", min, max, c, hash, lxkey->lfkey.hash[chkval].name, name);
		if( c == 0 ) break;

		if( min >= max ) {
			//fprintf(stderr, "lf_findkey(): couldn't find value.\n");
			return -1;
		}

		if( c > 0 ) min = chkval + 1;
		if( c < 0 ) max = chkval - 1;
	}

	//chkval should now point to a matching hash
	min = chkval;
	max = chkval;

	//figure out which hashes we need to check
	while( ( max + 1 < no_subkeys ) && ( 0 == strncasecmp( hash, lxkey->lfkey.hash[max+1].name, 4 ) ) ) {
		max++;
	}

	while( ( min - 1 >= 0          ) && ( 0 == strncasecmp( hash, lxkey->lfkey.hash[ min-1 ].name, 4 ) ) ) {
		min--;
	}

	//fprintf(stderr, "Found range: %ld - %ld\n", min, max );

	{
		struct nk_key *subkey = lx_get_subkey( hdesc, lxkey, min );
		nk_get_name( subkey, buf, SZ_MAX );
		//fprintf(stderr, "%ld, %ld %s -> %s\n", min, max, name, buf );
		//fflush(stderr);
	}

	for( i=min; i<=max; i++) {
		struct nk_key *subkey = lx_get_subkey( hdesc, lxkey, i );
		nk_get_name( subkey, buf, SZ_MAX );
		if( gverbose ) {
			fprintf(stderr, "lf_findkey() \"%s\" == \"%s\"?\n", name, buf );
			fflush(stderr);
		}
		if( 0 == strcmp( name, buf ) ) {
			//fprintf(stderr, "lf_findkey(): success\n");
			return i;
		}
	}

	return -1;
} // }}}
#endif
int winstrncmp( const char *s1, const char *s2, int n ) // {{{ Compare two strings using Microsoft Alphabet Order (TM)
{
	int i;
	//WARN("winstrncmp(): comparing %s to %s", s1, s2 );
	for( i=0; n<0 || i<n; i++ ) {
		if( 0 == s1[i] && 0 == s2[i] ) return 0;
		int v1 = toupper(s1[i]);
		int v2 = toupper(s2[i]);
		if( 0 != v1 - v2 ) {
			return v1 - v2;
		}
	}
	return 0;
} // }}}
int winstrcmp( const char *s1, const char *s2 ) // {{{ Compare two strings using Microsoft Alphabet Order (TM)
{
	return winstrncmp( s1, s2, -1 ); //this is nice
} // }}}
char *comp_buf = NULL;
int cblen = 0;
int lx_comp( struct hive *hdesc, lx_key *lx, long leafno, const char *name ) // {{{ like strcmp, but not TODO: comp -> cmp
{
	int c;
	struct nk_key *ck;
	int result;
	VBUF_STATIC( namebuf, 10 );

	/* This is broken for UTF-16 names
	if( lx->head.id == ID_LF_KEY ) { //we can take a shortcut here
		//fprintf(stderr, "LF Shortcut: %s <> %s\n", name, lx->lfkey.hash[leafno].name );
		char hash[4];
		lf_hash( name, hash );
		c = winstrncmp( name, lx->lfkey.hash[leafno].name, 4 );
		if( c != 0 ) {
			ck = lx_get_subkey( hdesc, lx, leafno );
			nk_vb_get_name( ck, namebuf );
			WARN("Shortcut:");
			hexdump( name, 0, strlen( name ), 1 );
			hexdump( namebuf->b, 0, namebuf->dlen, 1 );
			hexdump( lx->lfkey.hash[leafno].name, 0, 4, 1 );
			hexdump( hash, 0, 4, 1 );
			return c;
		}
	}
	*/

	ck = lx_get_subkey( hdesc, lx, leafno );

	if( cblen < nk_get_name_len( ck ) + 1 ) {
		cblen = nk_get_name_len( ck ) + 1;
		comp_buf = F_REALLOC( comp_buf, cblen );
	}
	nk_vb_get_name( ck, namebuf );
	//WARN("Calling winstrcmp() on:");
	//hexdump( name, 0, strlen( name ), 1 );
	//WARN("LEN: %d", namebuf->dlen );
	//hexdump( namebuf->b, 0, namebuf->dlen, 1 );
	result = winstrcmp( name, namebuf->b ); 

	//nk_get_name( ck, comp_buf, cblen );
	//fprintf(stderr, "lx_comp: %s <> %s\n", name, comp_buf );
	//result = winstrcmp( name, comp_buf  ); 
	//fprintf(stderr, "[%ld], Compare name [%s] buf [%s] == %d\n", leafno, name, comp_buf, result );
	return( result );

} // }}}
int nk_comp( struct hive *hdesc, struct nk_key *nk, long no, const char *name ) // {{{
{
	ASSERT( no >= 0 && no < nk->no_subkeys, "nk_comp() no outside of range.");
	return lx_comp( hdesc, nk_get_lx( hdesc, nk ), no, name );
} // }}}
int nk_comp_val( struct hive *hdesc, struct nk_key *nk, long no, const char *name ) // {{{
{
	VBUF_STATIC( namebuf, 10 );
	//static struct varbuf *vb = NULL;
	//if( NULL == vb ) vb = vbuf_alloc(10);
	ASSERT( no >= 0 && no < nk->no_values, "nk_comp_val() no outside of range.");
	struct vk_key *vk = nk_get_value( hdesc, nk, no );

	//vbuf_resize( vb, vk->len_name + 1 );
	//vb->dlen = vk->len_name;
	//vk_get_name( vk, vb->b, vb->blen );
	vb_vk_get_name( namebuf, vk );

	return winstrcmp( name, namebuf->b );
} // }}}
long nk_seek_nk( struct hive *hdesc, struct nk_key *nk, const char *name ) //{{{ seeks first key >= name or -1 if not found
{
	//TODO: gross
	long min=0;
	long max=nk_get_no_subkeys( nk ) - 1;
	int found=0;
	long chkval;

	if( nk->no_subkeys == 0 ) return -1;
	//printf("Seeking %s in\n", name );
	//parse_nk_nice( hdesc, nk );

	do { //binary search 
		chkval = min + (max - min)/2;
		int c = nk_comp( hdesc, nk, chkval, name );

		//fprintf(stderr, "IN  Weee: min:%ld max:%ld c:%d name:%s\n", min, max, c, name);

		if( c > 0 ) { //chkval < name
			if( chkval < max )
				min = chkval + 1;
			else
				return -1;
		}

		if( c < 0 ) { //chkval > name
			max = chkval;
			found=1;
		}
		if( c == 0 ) {
			ASSERT( chkval < nk->no_subkeys, "Out of bounds" );
			return chkval;
		}

		if( max == min ) {
			ASSERT( max < nk->no_subkeys, "Out of bounds" );
			if( 0 >= nk_comp( hdesc, nk, max, name ) )
				return max;
			else
				return -1;
		}

		//fprintf(stderr, "OUT Weee: min:%ld max:%ld c:%d name:%s\n", min, max, c, name);
	} while( min < max ); 

	return ( chkval < nk_get_no_subkeys( nk ) )? chkval: -1;
}
// }}}
long nk_seek_vk( struct hive *hdesc, struct nk_key *nk, const char *name ) //{{{ seeks first key >= name or -1 if not found
{
	long min=0;
	long max=nk->no_values - 1;
	long chkval;

	if( nk->no_values == 0 ) return -1;
	//printf("Seeking %s in\n", name );
	//parse_nk_nice( hdesc, nk );

	do { //binary search 
		chkval = min + (max - min)/2;
		int c = nk_comp_val( hdesc, nk, chkval, name );

		//fprintf(stderr, "IN  Weee: min:%ld max:%ld c:%d name:%s\n", min, max, c, name);

		if( c > 0 ) min = chkval + 1;
		if( c < 0 ) max = chkval;
		if( c == 0 ) {
			ASSERT( chkval < nk->no_values, "Out of bounds" );
			return chkval;
		}

		if( max == min ) {
			ASSERT( max < nk->no_values, "Out of bounds" );
			return max;
		}

		//fprintf(stderr, "OUT Weee: min:%ld max:%ld c:%d name:%s\n", min, max, c, name);
	} while( min < max ); 

	return ( chkval < nk->no_values )? chkval: -1;
} // }}}
int nk_delete_subkey( struct hive *hdesc, struct nk_key *nkparent, const char *name ) // {{{ Removes a subkey from an nk_key given a name 
{
	int subkeyno;
	struct nk_key *subkey;

	if( -1 == ( subkeyno = nk_seek_nk( hdesc, nkparent, name ) ) ) {
		//TODO: is ENOENT right here?
		ERR_UNIX( ENOENT, "nk_delete_subkey(): subkey %s not found.", name );
		return -1;
	}

	if( 0 != nk_comp( hdesc, nkparent, subkeyno, name ) ) {
		//TODO: is ENOENT right here?
		ERR_UNIX( ENOENT, "nk_delete_subkey(): subkey %s not found.", name );
		return -1;
	}

	subkey = nk_get_subkey( hdesc, nkparent, subkeyno );

	if( subkey->no_subkeys > 0 ) {
		//TODO: is ENOEMPTY right here?
		ERR_UNIX( ENOTEMPTY, "nk_delete_subkey(): subkey has subkeys, I cannot delete it.");
		return -1;
	}

	while( subkey->no_values > 0 ) {
		nk_delete_value( hdesc, subkey, 0 );
	}

	lx_key *lx = nk_get_lx( hdesc, nkparent );

	if( -1 == subkeyno ) {
		//TODO: is ENOENT right here?
		ERR_UNIX( ENOENT, "nk_delete_subkey(): subkey %s not found.", name );
		return -1;
	}

	if( 0 != lx_remove_entry( hdesc, lx, subkeyno ) ) {
		return -1;
	}

	nkparent->no_subkeys--;

	return 0;
}
// }}}
lx_key *nk_get_lx( struct hive *hdesc, struct nk_key *nk ) // {{{
{

	//if( nk->no_subkeys <= 0 ) WARN("Attempting to get lx for nk with no subkeys.");
	ASSERT( nk->ofs_lf > -1, "Attempt to get invalid lx record. :P");
	struct hbin_block *hb = get_hbin_block( hdesc, get_ptr_off_h( hdesc, nk->ofs_lf ) );


	ASSERT( hbinb_is_allocd( hb ), "Hbin block is not allocated.");
	
	//struct hbin_block *hb = nk_get_hbin_block( nk );
	lx_key *result = pull_lxkey( hdesc, hb->data );

	return result;
}
// }}}
struct v_list *nk_get_vlist( struct hive *hdesc, struct nk_key *nk ) // {{{
{
	struct v_list *vlist;
	struct hbin_block *hb;

	hb = get_hbin_block( hdesc, get_ptr_off_h( hdesc, nk->ofs_vallist ) );
	vlist = (struct v_list *)(hb->data);

	return vlist;
}
// }}}
struct vk_key *nk_get_value( struct hive *hdesc, struct nk_key *nk, int off ) // {{{
{
	//void *p;
	struct vk_key *result;

	struct hbin_block *hb;
	struct v_list *vlist = nk_get_vlist( hdesc, nk );

	ASSERT( off >= 0 && off < nk->no_values, "nk_get_value(): index outside of range.");

	hb = get_hbin_block( hdesc, get_ptr_off_h( hdesc, vlist->vofs[off] ) );

	result = check_vk_ptr( hdesc, hb->data );

	return result;
}
// }}}
int nk_find_value( struct hive *hdesc, struct nk_key *nk, const char *name ) // {{{ return offset of value of name 'name' (or -1)
{
	VBUF_STATIC( namebuf, 10 );
	int i;
	for( i=0; i < nk->no_values; i++ )
	{
		vb_vk_get_name( namebuf, nk_get_value( hdesc, nk, i ) );
		//vk_get_name( nk_get_value( hdesc, nk, i ), buf, SZ_MAX );
		//WARN(" %s == %s ??", name, namebuf->b );

		if( gverbose )fprintf(stdout, "nk_find_value() \"%s\" == \"%s\"?\n", name, namebuf->b );
		if( 0 == strcmp( name, namebuf->b ) )
			return i;
	}

	return -1;
}
// }}}
int nk_find_subkey( struct hive *hdesc, struct nk_key *nk, const char *name ) // {{{
{
	int off = nk_seek_nk( hdesc, nk, name );

	if( -1 == off ) return -1;
	if( 0 != nk_comp( hdesc, nk, off, name ) ) return -1;
	return off;
}
// }}}
struct nk_key *nk_get_subkey( struct hive *hdesc, struct nk_key *nk, int off ) // {{{
{
	ASSERT( nk_get_no_subkeys(nk) > off, "nk_get_no_subkeys(nk) > off" );
	return lx_get_subkey( hdesc, nk_get_lx( hdesc, nk ), off );
}
// }}}

struct hbin_block *lx_get_hbinb( lx_key *lx ) // {{{
{
	//TODO: this -4 stuff sucks
	return (struct hbin_block*)((char*)lx - 4);
} // }}}
lx_key *lx_get_lx( struct hive *hdesc, lx_key *lx, int off, int *off_out ) // {{{ retrieve "leaf" lx if not already
{
	//struct hbin_block *hb;

	switch( lx->head.id ) {
		case ID_LF_KEY:
		case ID_LH_KEY:
		case ID_LI_KEY:
			*off_out = off;
			return lx;

		case ID_RI_KEY:
			return ri_find_lx( hdesc, (struct ri_key*)lx, off, off_out, NULL ); 

		default:
			DIE("lx_get_lx(): unknown ID 0x%x (%s)",
				lx->head.id, id_to_str( lx->head.id ) );
	}

	DIE("lx_get_lx(): how did I get here?");
	return NULL;
} // }}}
lx_key *lx_get_insert_lx( struct hive *hdesc, lx_key *lx, int off, int *off_out ) // {{{ retrieve "leaf" lx if not already
{
	lx_key *r;
	//struct hbin_block *hb;

	switch( lx->head.id ) {
		case ID_LF_KEY:
		case ID_LH_KEY:
		case ID_LI_KEY:
			*off_out = off;
			return lx;

		case ID_RI_KEY:
			r = ri_find_lx( hdesc, (struct ri_key*)lx, off, off_out, NULL ); 
			//TODO: This is "retry out of ignorance", which is ungood
			if(!r && off > 0) {
				WARN("Retrying out ignorance @ %d", off );
				r= ri_find_lx( hdesc, (struct ri_key*)lx, off - 1, off_out, NULL ); 
			}
			return r;

		default:
			DIE("lx_get_lx(): unknown ID 0x%x (%s)",
				lx->head.id, id_to_str( lx->head.id ) );
	}

	DIE("lx_get_lx(): how did I get here?");
	return NULL;
} // }}}
long lx_get_hash_off( lx_key *lx, int off ) // {{{
{
	ASSERT( off >= 0 && off < lx->head.no, "offset (%d) is outside of bounds (%d).", off, lx->head.no );
	switch( lx->head.id ) {
		case ID_LI_KEY:
			return lx->likey.hash[off].ofs_nk;
		case ID_LF_KEY:
			return lx->lfkey.hash[off].ofs_nk;
		case ID_LH_KEY:
			return lx->lhkey.hash[off].ofs_nk;
		case ID_RI_KEY:
			return lx->likey.hash[off].ofs_nk;
		default:
			DIE("lx_get_hash_off(): Unknown lx type %x", lx->head.id );
	}

	DIE("lx_get_hash_off(): how did I get here?");
	return -1;
} // }}}
struct nk_key *lx_get_subkey( struct hive *hdesc, lx_key *lx, int off ) // {{{
{
//	lx_key *lx = nk_get_lx( hdesc, nk );
	struct hbin_block *hb;
	int toff;
	//ASSERT( off < lx->head.no, "Record outside of bounds" );
	lx_key *tlx = lx_get_lx( hdesc, lx, off, &toff );

	get_ptr_off_h( hdesc, lx_get_hash_off( tlx, toff ) );

	hb = get_hbin_block( hdesc, get_ptr_off_h( hdesc, lx_get_hash_off( tlx, toff ) ) );

	return check_nk_ptr( hdesc, hb->data );
}
// }}}
void *gen_trav_path(struct hive *hdesc, struct nk_key *nk, const char *ppath, int type) // {{{ follow 'nk'-nodes based on a path-string,
{

/* 
 * in: hdesc, nk, ppath, type;
 * path - null-terminated pathname (relative to nk, \ is separator) CAN BE NULL
 * type - type to return 0=nk 1=vk
 * return: (type) || NULL if not found
 *
 * TODO: this is HUGE and gross
 */
	char *buf;
	char *path = (char*)ppath;
	char tok[SZ_MAX];
	struct nk_key *tnk; //the current nk
	VBUF_STATIC( tmp, 10 );

	char debugbuf[SZ_MAX];

	buf = hdesc->buf;

	tnk = nk;

	while( 1 ) {
		int token_type = gettoken( tok, SZ_MAX, &path, '\\' );
		if( gverbose ) {
			debugbuf[0] = '\0';
			nk_get_path( hdesc, tnk, debugbuf, SZ_MAX );
			fprintf(stderr, "trav_path(): (path:%s;tok:%s) main loop @ %s\n", path, tok, debugbuf);
		}

		switch( token_type  )
		{
			case TOK_DELIM:
				// Start from root if path starts with DELIM
				tnk = hdesc->root_key;
				//hb = hdesc->root_block;

				if( gverbose ) { fprintf( stderr, "trav_path(): path is absolute, using nk=%p\n", nk ); fflush( stderr ); };
				break;

			case TOK_BUF_SMALL:
				DIE("Token buffer was too small.");
				return NULL;

			case TOK_ERROR:
				DIE("Token error.");


			case TOK_PARENT:
				/* Time to go up a level */
				tnk = nk_get_parent( hdesc, tnk );
				break;

			case TOK_CURRENT:
				break;

			case TOK_ELEMENT:
			{
				/* got a path element */

				//fprintf(stderr, "nk_trav_path() seeks tok %s within an NK:\n", tok );
				//parse_nk_nice( hdesc, tnk );

				int kno=0;

				if( 0 <= ( kno = nk_find_subkey( hdesc, tnk, tok ) ) ) {
					// first try to retrieve an NK child of this NK

					tnk = nk_get_subkey( hdesc, tnk, kno );
				}
				else { // otherwise look for a value and hope we are at the end
					//Searching for a VK
					if( -1 == ( kno = ( nk_find_value( hdesc, tnk, tok ) ) ) )
						goto return_null;

					if( TOK_EMPTY != gettoken( tok, SZ_MAX, &path, '\\' ) ) goto return_null;

					if( type != 1 ) goto return_null;

					return nk_get_value( hdesc, tnk, kno );
				}
			}
			break;

			case TOK_EMPTY:
				/* All done ! */
				if( gverbose ) {
					fprintf( stderr, "trav_path( %s ): returning with: \n", path);
					parse_nk_nice( hdesc, tnk );
				}
				switch( type ) {
					case 0:
						return tnk;
					case 1:
					{
						int kno;
						if( -1 == ( kno = ( nk_find_value( hdesc, tnk, "" ) ) ) )
							goto return_null;

						return nk_get_value( hdesc, tnk, kno );
					}
					default:
						DIE("I have no idea how I got here.");
				}

			default:
				DIE("gettoken() came back with something weird." );
		}
	}

	/* Not found */

return_null:
	fprintf(stderr, "gen_trav_path() is returning NULL.\n");
	parse_nk_nice( hdesc, tnk );
	nk_vb_get_path( tmp, hdesc, tnk );
	fprintf(stderr, "    %s\n", tmp->b );
	return NULL;
}
// }}}
struct nk_key *nk_trav_path(struct hive *hdesc, struct nk_key *nk, const char *ppath) // {{{ follow 'nk'-nodes based on a path-string,
{
	return (struct nk_key*)gen_trav_path( hdesc, nk, ppath, 0 );
} // }}}
struct vk_key *vk_trav_path(struct hive *hdesc, struct nk_key *nk, const char *ppath) // {{{ follow 'nk'-nodes based on a path-string,
{
	return (struct vk_key*)gen_trav_path( hdesc, nk, ppath, 1 );
} // }}}
struct hbin_block *trav_path(struct hive *hdesc, int i_vofs, char *path, int type) // {{{ Recursevely follow 'nk'-nodes based on a path-string,
{ //TODO: this is for backward-compat and should be destroyed
	struct nk_key *nk=NULL;

	if( i_vofs != 0 ) {
		nk = check_nk_ptr( hdesc, get_ptr_off_f( hdesc, i_vofs ) );
		if( gverbose ) { fprintf( stderr, "trav_path(): path is relative=0x%p\n", nk ); fflush( stderr ); };
		if( NULL == ( nk =
			check_nk_ptr( hdesc, get_ptr_off_f( hdesc, i_vofs ) ) ) )
		fprintf( stderr, "trav_path() passed something dumpy: 0x%x\n", i_vofs );
	}
	else {
		nk = hdesc->root_key;

		if( gverbose ) { fprintf( stderr, "trav_path(): i_vofs == 0, trying root nk=%p\n", nk ); fflush( stderr ); };
	}

	nk = gen_trav_path( hdesc, nk, path, type );

	if( NULL == nk ) return NULL;

	return nk_get_hbin_block( nk );
}
// }}}
struct nk_key* get_root_key( struct hive *hdesc ) // {{{
{
	return hdesc->root_key;
}
// }}}

// Modification functions

int ri_free_child( struct hive *hdesc, struct ri_key *ri, int no ) // {{{ Remove a direct child of an RI
{
	struct hbin_block *child;
	if( no < 0 || no >= ri->head.no ) {
		ERR_UNIX( EINVAL, "ri_free_child(): attempt to delete entry that is out of range.");
		return -1;
	}

	child = get_ptr_off_h( hdesc, ri->hash[no] );

	if( ri->head.no - 1 > no ) {
		//TODO; hash sizes are assumed the same
		memmove( &ri->hash[no],
			&ri->hash[no+1],
			sizeof( li_offset ) * (ri->head.no - ( no + 1 ) )  );
	}

	ri->head.no--;

	hbinb_free( hdesc, child );
	return 0;
} // }}}
int lh_drop_entry( struct hive *hdesc, struct lh_key *lh, int no ) // {{{ Remove entry from lh
{
	struct hbin_block *child;

	if( no >= lh->head.no ) {
		ERR_UNIX( EINVAL, "lx_drop_entry(): attempt to delete entry that is out of range.");
		return -1;
	}

	child = get_ptr_off_h( hdesc, lh->hash[no].ofs_nk );
			/*
	free_block(hdesc,
		lh->hash[no].ofs_nk + REGF_HEADER_SIZE);
		*/
	if( lh->head.no - 1 > no )
	{
		 //TODO; hash sizes are assumed the same
		memmove( &lh->hash[no],
			&lh->hash[no+1],
			sizeof( struct lh_hash ) * (lh->head.no - ( no + 1 ) )  );
	}

	lh->head.no--;

	hbinb_free( hdesc, child );

	return 0;
} // }}}
int lx_remove_entry( struct hive *hdesc, lx_key *lx, int no ) // {{{ Removes an entry from an lx
{
	switch( lx->head.id ) {
		case ID_LF_KEY:
		case ID_LH_KEY:
		case ID_LI_KEY:
		{
			struct hbin_block *child;
			void *p;
			int slotsize = lx_get_slot_size( lx );

			if( no >= lx->head.no ) {
				ERR_UNIX( EINVAL, "lx_drop_entry(): attempt to delete entry that is out of range.");
				return -1;
			}

			p = lx_get_raw_slot( lx, no );
			child = get_ptr_off_h( hdesc, lx_get_hash_off( lx, no ) );

			if( lx->head.no - 1 > no ) {
				memmove( p, p + slotsize, 
				slotsize * ( lx->head.no - no - 1 ) );
			}

			lx->head.no--;

			hbinb_free( hdesc, child );

			return 0;
			/*
			struct lh_key *lh= &lx->lhkey;

			return lh_drop_entry( hdesc, lh, no );
			*/
		}
		case ID_RI_KEY:
		{
			int lxs_no;
			int lxp_off;
			int result;
			lx_key *lxs = ri_find_lx( hdesc, &lx->rikey, no, &lxs_no, &lxp_off );
			if( NULL == lxs ) return -1;
			result = lx_remove_entry( hdesc, lxs, lxs_no );
			if( lxs->head.no == 0 ) {
				ri_free_child( hdesc, &lx->rikey, lxp_off );
				//DIE("lx is empty, should be fixed.");
			}
			return result;
			DIE("lx_remove_entry(): It's an RI key, what do I do?!");
		}
		default:
			DIE("lx_remove_entry(): ID == 0x%x, confused.\n",
					lx->head.id );
	}

	return 0;
} // }}}
void nk_set_lx( struct hive *hdesc, struct nk_key *nk, lx_key *lx ) // {{{ 
{
	nk->ofs_lf = get_heap_offset( hdesc, lx_get_hbinb( lx ) );
} // }}}
struct nk_key *nk_alloc_utf16( struct hive *hdesc, void *p, const char *name ) // {{{ Allocate a new nk_key in same page as p, if possible
{
	struct nk_key *result;
	struct hbin_block *nknewblk;
	VBUF_STATIC( utf16buf, 10 );
	VBUF_STATIC( utf8buf, 10 );

	if( -1 == vb_utf8to16( utf16buf, name, strlen( name ) + 1 ) )
		return NULL;

	if( -1 == vb_utf16to8( utf8buf, utf16buf->b, utf16buf->dlen ) ) {
		DIE("Reverse convert failed.");
	}
	ASSERT( strlen( name ) + 1 == utf8buf->dlen && 0 == memcmp( name, utf8buf->b, utf8buf->dlen ), "Reverse convert comp failed.");

	lx_key *lx;
	//allocate the new nk
	if( NULL == ( nknewblk = hbin_alloc(hdesc, p, sizeof(struct nk_key) + utf16buf->dlen - 2) ) ) {
		return NULL;
	}

	//allocate the new nk's lx
	if( NULL == ( lx = lx_alloc( hdesc, p, ID_LF_KEY, 0 ) ) ) { //TODO: Should I use LF or LH?  What if this is for an RI?
		hbinb_free( hdesc, nknewblk );
		return NULL;
	}

	result = (struct nk_key*)nknewblk->data;

	//populate it
	result->id            = ID_NK_KEY;
	result->type          = KEY_UTF16;
	result->ofs_parent    = 0; //get_heap_offset( hdesc, nkparent ); //TODO: update this on insertion
	result->no_subkeys    = 0;
	result->no_values     = 0;
	result->ofs_vallist   = -1;
	result->ofs_sk        = 0; //TODO: update this on insertion //nkparent->ofs_sk; /* Get parents for now. 0 or -1 here crashes XP */
	result->ofs_classnam  = -1;
	result->len_name      = utf16buf->dlen - 2;
	result->len_classnam  = 0;
	memcpy(result->keyname, utf16buf->b, result->len_name ); //Yup, we don't zero-terminate in the registry (in general)

	nk_set_lx( hdesc, result, lx );

	return result;
} // }}}
struct nk_key *nk_alloc( struct hive *hdesc, void *p, const char *name ) // {{{ Allocate a new nk_key in same page as p, if possible
{
	struct nk_key *result;
	struct hbin_block *nknewblk;
	lx_key *lx;
	if( NULL == ( nknewblk = hbin_alloc(hdesc, p, sizeof(struct nk_key) + strlen(name)) ) ) {
		return NULL;
	}

	if( NULL == ( lx = lx_alloc( hdesc, p, ID_LF_KEY, 0 ) ) ) { //TODO: Should I use LF or LH?  What if this is for an RI?
		hbinb_free( hdesc, nknewblk );
		return NULL;
	}

	result = (struct nk_key*)nknewblk->data;

	//populate it
	result->id		= ID_NK_KEY;
	result->type          = KEY_NORMAL;
	result->ofs_parent    = 0; //get_heap_offset( hdesc, nkparent ); //TODO: update this on insertion
	result->no_subkeys    = 0;
	result->ofs_lf        = get_heap_offset( hdesc, lx_get_hbinb( lx ) );
	result->no_values     = 0;
	result->ofs_vallist   = -1;
	result->ofs_sk        = 0; //TODO: update this on insertion //nkparent->ofs_sk; /* Get parents for now. 0 or -1 here crashes XP */
	result->ofs_classnam  = -1;
	result->len_name      = strlen(name);
	result->len_classnam  = 0;
	strncpy(result->keyname, name, strlen(name));

	return result;
} // }}}
int lx_get_slot_size( lx_key *lx ) // {{{
{
	switch( lx->head.id ) {
		case ID_LI_KEY:
			return sizeof( struct li_hash );
		case ID_LF_KEY:
			return sizeof( struct lf_hash );
		case ID_LH_KEY:
			return sizeof( struct lh_hash );
		default:
			DIE("Unknown type %x.", lx->head.id );
	}
	DIE("How did I get here?");
	return -1;
} // }}}
void *lx_get_raw_slot( lx_key *lx, int off ) // {{{ return the offset to a slot in the array
{
	ASSERT( off >= 0 && off < lx->head.no, "Out of bounds" );
	switch( lx->head.id ) {
		case ID_LI_KEY:
			return &((struct li_key*)lx)->hash[off];

		case ID_LF_KEY:
			return &((struct lf_key*)lx)->hash[off];

		case ID_LH_KEY:
			return &((struct lh_key*)lx)->hash[off];

		case ID_RI_KEY:
			return &((struct ri_key*)lx)->hash[off];

		default:
			DIE("lx_get_rawslot(): unknown type.");
			return NULL;
	}
} // }}}
void lx_set_slot( struct hive *hdesc, lx_key *lx, int off, char *name, struct hbin_block *ptr )
{
	void *slot = lx_get_raw_slot( lx, off );
	switch( lx->head.id ) {
		case ID_LI_KEY:
			((struct li_hash*)slot)->ofs_nk = get_heap_offset( hdesc, ptr );
			return;
		case ID_LF_KEY:
			lf_hash( name, ((struct lf_hash*)slot)->name );
			((struct lf_hash*)slot)->ofs_nk = get_heap_offset( hdesc, ptr );
			return;
		case ID_LH_KEY:
			((struct lh_hash*)slot)->hash = lh_hash( name );
			((struct lh_hash*)slot)->ofs_nk = get_heap_offset( hdesc, ptr );
			//parse_nk_nice( hdesc, (struct nk_key*)ptr->data );
			return;
		case ID_RI_KEY:
			DIE("lx_set_slot(): RI: I'm not sure this makes sense...");
		default:
			DIE("lx_set_slot(): unknown type.");
			return;
	}
}
//TODO: static
char *lxinsertbuf = NULL;
int lxinsertbuflen = 0;
lx_key *lx_insert_nk( struct hive *hdesc, lx_key *lx, int no_subkeys, int offset, struct nk_key *nkchild ) // {{{ allocate bigger lx_key, insert child and return
{
	//TODO: big and ugly
	int slot_size;
	lx_key *nlx;
	VBUF_STATIC( nknamebuf, 10 );

	if( -1 == nk_vb_get_name( nkchild, nknamebuf ) ) 
		return NULL;

	//struct hbin_block *hnkold = get_ptr_off_h( hdesc, nkparent->ofs_lf );;

	//parse_hbin_page_long( hdesc, find_page_start( hdesc, lx ) );

	if( lx->head.id == ID_RI_KEY ) {
		int toff;
		struct ri_key *ri = (struct ri_key*)lx;
		int lx_off, ri_bucket_off;

		WARN("I'm going in @ %d -> %d!", offset, toff);

		lx_key *slx;

		if( offset == no_subkeys )
			slx = ri_find_lx( hdesc, ri, offset - 1, &lx_off, &ri_bucket_off);
		else
			slx = ri_find_lx( hdesc, ri, offset, &lx_off, &ri_bucket_off);

		ASSERT( NULL != slx, "slx is null." );

		if( ( nlx = lx_insert_nk( hdesc, slx, no_subkeys, lx_off, nkchild ) ) ) {
			ri->hash[ri_bucket_off];
			return lx;
		}
		else
			return NULL;

	}

	slot_size = lx_get_slot_size( lx );

	if( NULL == (nlx = lx_alloc( hdesc, lx, lx->head.id, lx->head.no + 1 ) ) )
		return NULL;


	//parse_hbin_page_long( hdesc, find_page_start( hdesc, nlx ) );

	if( lx->head.no > 0 )
		memcpy( lx_get_raw_slot( nlx, 0 ), lx_get_raw_slot( lx, 0 ), slot_size * offset );

	if( lx->head.no - offset ) {
		memcpy( lx_get_raw_slot( nlx, offset + 1 ), lx_get_raw_slot( lx, offset ),
				slot_size * (lx->head.no - offset ) );
	}

	if( lxinsertbuflen < nk_get_name_len( nkchild ) + 1 ) {
		lxinsertbuflen = nk_get_name_len( nkchild ) + 1;
		lxinsertbuf = F_REALLOC( lxinsertbuf, lxinsertbuflen );
	}

	//nk_get_name( nkchild, lxinsertbuf, lxinsertbuflen );
	//lx_set_slot( hdesc, nlx, offset, lxinsertbuf, nk_get_hbin_block( nkchild ) );

	lx_set_slot( hdesc, nlx, offset, nknamebuf->b, nk_get_hbin_block( nkchild ) );

	return nlx;
} // }}}
int nk_is_sane( struct hive *hdesc, struct nk_key *nk ) // {{{ returns: 1 == good, 0 == bad
{
	//TODO: this could check a lot more stuff

	int i;
	for( i=0; i<nk->no_subkeys; i++ ) {
		struct nk_key *snk = nk_get_subkey( hdesc, nk, i );
		if(!snk) DIE("Bad subkey!"); // return 0;

		if( !nk_is_sane( hdesc, snk) ) return 0;
	}

	return 1;
} // }}}
int nk_insert_nk( struct hive *hdesc, struct nk_key *nkparent, struct nk_key *nkchild ) // {{{ Insert an allocated nk_key into parent
{
	long seekval;

	struct hbin_block *hlxold = NULL;
	VBUF_STATIC( namebuf, 10 );

	nkchild->ofs_parent = get_heap_offset( hdesc, nkparent );
	nkchild->ofs_sk = nkparent->ofs_sk;

	lx_key *nlx;
	//lx_key *nlx = lx_alloc( hdesc, nkparent, nk_get_lx( hdesc, nkparent )->head.id, nkparent->no_subkeys + 1 );

	//if( NULL == nlx ) return -1;

	/*
	if( insertbuflen < nk_get_name_len( nkchild ) + 1 ) {
		insertbuflen = nk_get_name_len( nkchild ) + 1;
		insertbuf = F_REALLOC( insertbuf, insertbuflen );
	}
	*/

	if( 0 != nk_vb_get_name( nkchild, namebuf ) ) {
		return -1;
	}

	seekval = nk_seek_nk( hdesc, nkparent, namebuf->b );

	//nk_get_name( nkchild, insertbuf, insertbuflen );

	//seekval = nk_seek_nk( hdesc, nkparent, insertbuf );

	if( seekval > -1 ) { 
		if( 0 == nk_comp( hdesc, nkparent, seekval, namebuf->b ) ) {
			ERR_WIN( rlERROR_BROKEN_ERROR, "nk_insert_nk(): key already exists.");
			return -1;
		}
	}

	if( -1 == seekval ) seekval = nkparent->no_subkeys;

	//fprintf(stderr, "seekval: %ld\n", seekval );

	if( -1 == nkparent->ofs_lf ) { //TODO: we are allocating an empty lx so that lx_insert_nk can reallocate it, seems yucky
		lx_key *lx;
		WARN("UNTESTED, Allocating lx for pre-existing nk");
		if( NULL == ( lx = lx_alloc( hdesc, nkparent, ID_LF_KEY, 0 ) ) ) { //TODO: Should I use LF or LH?  What if this is for an RI?
			return -1;
		}

		nk_set_lx( hdesc, nkparent, lx );
	}
	else {
		hlxold = get_ptr_off_h( hdesc, nkparent->ofs_lf );;
	}
	lx_key *olx = nk_get_lx( hdesc, nkparent );
	if( NULL == ( nlx = lx_insert_nk(
			hdesc, olx, nk_get_no_subkeys( nkparent ), seekval, nkchild )
			)) return -1;

	//parse_hbin_page_long( hdesc, find_page_start( hdesc, hnkold ) );
	if( olx != nlx ) {
		nkparent->ofs_lf = get_heap_offset( hdesc, lx_get_hbinb( nlx ) );

		if( hlxold ) hbinb_free( hdesc, hlxold );

		nkparent->no_subkeys++;
	}

	ASSERT( nk_is_sane( hdesc, nkparent ), "Uh...I've got some bad news.");
	
	return 0;
} // }}}
struct nk_key *nk_create( struct hive *hdesc, const char *name, struct nk_key *nkparent ) // {{{ Allocate a new nk_key, insert it into the parent and return it
{
	/* hdesc - usual..
	* name  - name of key to add (UTF8)
	* parent - parent key
	* return: ptr to new nk_key, or NULL
	*/
	struct nk_key *result;
	VBUF_STATIC( namebuf, 10 );

	if( NULL == ( result = nk_alloc_utf16( hdesc, nkparent, name ) ) )
		return NULL;

	nk_vb_get_name( result, namebuf );

	//WARN("Trying to write: %s, aka:", namebuf->b );
	//hexdump( namebuf->b, 0, namebuf->dlen, 1 );

	if( 0 != nk_insert_nk( hdesc, nkparent, result ) ) {
		hbinb_free( hdesc, nk_get_hbin_block( result ) );
		return NULL;
	}

	return result;

}
// }}}
int nk_delete_value(struct hive *hdesc, struct nk_key *nk, int val_no) // {{{ Delete single value from key
{
	/* hdesc - yer usual hive
	* nkofs - current keyoffset
	* name  - name of value to delete
	* returns: 0 - ok, 1 - failed
	*/
	long *vlistkey; //, *newlistkey; //*tmplist, 

	struct vk_key *vk = nk_get_value( hdesc, nk, val_no );
	struct hbin_block *oldblk = get_ptr_off_h( hdesc, nk->ofs_vallist );

	vlistkey = (long*)oldblk->data;
	if( val_no >= nk->no_values ) {
		ERR_UNIX( EINVAL, "nk_delete_value(): value number outside of range.");
		return -1;
	}

	/* Delete vk and data */
	nk->no_values--;
	if( val_no < nk->no_values ) {
		memmove( &vlistkey[val_no], &vlistkey[val_no+1], sizeof( long ) * (nk->no_values - val_no) );
	}

	vk_delete( hdesc, vk );

	return 0;
}
// }}}
// }}}
// {{{ **** VK access and modification functions
void *vk_get_val( struct hive *hdesc, struct vk_key *vk ) /* {{{ Pull data from a key */
{
	void *r=NULL;
	HBB *hbb;
	static char empty[0];
	if (vk->len_data == 0) {
		DIE("IS THIS VALID?");
	}
#if 0
	if (vk->len_data == 0x80000000) {  /* Special inline case (len = 0x80000000) */
		//In this case the data are stored in the val_type field...but what's the type in that case?
		/*
		debug(
				fprintf(stderr, "Inline case, returning vk->val_type\n");
				hexdump((void*)&vk->val_type, 0, 4, 1 );
				);
				*/
		WARN("Got \"INLINE\"");
		return(&vk->val_type); /* Data (4 bytes?) in type field */
	}    
#endif

	if (vk->len_data == 0x80000000) {  /* Special inline case (len = 0x80000000) */
		// Only the type field is used here, regular data are zero-length
		/*
		debug(
				fprintf(stderr, "Inline case, returning vk->val_type\n");
				hexdump((void*)&vk->val_type, 0, 4, 1 );
				);
				*/
	//	WARN("Got \"INLINE\"");
		return empty;
		//return(&vk->val_type); /* Data (4 bytes?) in type field */
	}    


	/* Negative len is inline, return ptr to offset-field which in
	* this case contains the data itself
	* TODO: The original == 0x80000000 case was incorrect, is this correct?
	*/
	if (vk->len_data & 0x80000000) {
		//DIE("IS THIS VALID?");
		if( vk_get_length(vk) != 4 )
			WARN("len: %d\n", vk_get_length(vk) );
		return(&vk->ofs_data);
	}
	/* Normal return, return data pointer */
	hbb = get_hbin_block( hdesc, get_ptr_off_h( hdesc, vk->ofs_data ));
	r = hbb->data;

	if( vk_get_length( vk ) > hbinb_dlen( hbb ) ) {
		char badp[] = { 0x64, 0x62, 0x02 };
		if( 0 == memcmp( r, badp, 3 ) ) {
			/*
			parse_vk_struct( hdesc, vk );
			check_wtf( hdesc, hbb );
			*/
			//WARN("Value data is stored in DB_KEY.  Sorry, I don't do those yet.");
			ERR_WIN( rlERROR_OPEN_FAILED, "Value data stored in DB_KEY, I can't do those yet.  Sorry." );
		}
		else {
			ERR_WIN( rlERROR_OPEN_FAILED, "Value length > hbin data available, something is wrong." );
		}

		return NULL;
	}

	return r;
}
// }}}
int vk_get_val2( struct hive *hdesc, struct vk_key *vk, vbuf *vb ) /* {{{ Pull data from a key */
{
	void *r=NULL;
	HBB *hbb;
	VBUF_STATIC( utf8buf, 10 );

	if (vk->len_data == 0) DIE("VK Data length is zero.  I don't think this is valid (zero length == 0x80000000)."); 

	if (vk->len_data == 0x80000000) {  /* Special inline case (len = 0x80000000) */
		// Only the type field is used here, regular data are zero-length
		vbuf_set( vb, NULL, 0 );
		return 0;
	}    


	/* Negative len is inline, return ptr to offset-field which in
	* this case contains the data itself
	* TODO: The original == 0x80000000 case was incorrect, is this correct?
	*/
	if (vk->len_data & 0x80000000) {
		//DIE("IS THIS VALID?");
		//TODO: bounds check.
		if( vk_get_length(vk) != 4 )
			WARN("len: %d\n", vk_get_length(vk) );
		vbuf_set( vb, &vk->ofs_data, vk_get_length( vk ) );
		return 0;
		//return(&vk->ofs_data);
	}
	/* Normal return, return data pointer */
	hbb = get_hbin_block( hdesc, get_ptr_off_h( hdesc, vk->ofs_data ));
	r = hbb->data;

	if( vk_get_length( vk ) > hbinb_dlen( hbb ) ) {
		char badp[] = { 0x64, 0x62, 0x02 };
		if( 0 == memcmp( r, badp, 3 ) ) {
			/*
			parse_vk_struct( hdesc, vk );
			check_wtf( hdesc, hbb );
			*/
			//WARN("Value data is stored in DB_KEY.  Sorry, I don't do those yet.");
			ERR_WIN( rlERROR_OPEN_FAILED, "Value data stored in DB_KEY, I can't do those yet.  Sorry." );
		}
		else {
			ERR_WIN( rlERROR_OPEN_FAILED, "Value length > hbin data available, something is wrong." );
		}

		return -1;
	}

	switch( vk_get_type( hdesc, vk ) ) {
		case REG_SZ:
		case REG_EXPAND_SZ:
			if( 0 != vb_utf16to8( vb, r, vk_get_length( vk ) ) ) {
				ERR_WIN(rlERROR_BROKEN_ERROR, "vk_get_val2(): couldn't convert data from UTF-16 to UTF-8: %s", strerror( errno ) );
				return -1;
			}
			break;
		default:
			vbuf_set( vb, r, vk_get_length( vk ) );
			break;
	}

	return 0;
}
// }}}
ulong vk_get_type( struct hive *hdesc, struct vk_key *vk ) /* {{{ Pull data from a key */
{
	/*
	if (vk->len_data == 0x80000000) return vk->val_type; 
	if( vk->len_data & 0x80000000 ) {
		//DIE("I'm not sure this is correct.");
		return REG_DWORD;
	}
	*/
	return vk->val_type;
}
// }}}
int vk_get_length( struct vk_key *vk ) // {{{
{ 
	if (vk->len_data == 0x80000000) return 0; 
	if( vk->len_data & 0x80000000 ) { return 4; }
	return (vk->len_data & 0x7fffffff);
} // }}}
int vk_get_length2( struct hive *hdesc, struct vk_key *vk ) // {{{ returns the length of vk data, converts UTF16 to UTF8 first
{ 
	VBUF_STATIC( vb, 10 );
	switch( vk_get_type( hdesc, vk ) ) {
		case REG_SZ:
		case REG_EXPAND_SZ:
			vk_get_val2( hdesc, vk, vb );
			return vb->dlen;
		default:
			return vk_get_length( vk );
	}
} // }}}
int vk_get_name_len( struct vk_key *vk ) // {{{ return the length of the vk name were it encoded in UTF8
{
	VBUF_STATIC( namebuf, 10 );
	vb_vk_get_name( namebuf, vk );
	return namebuf->dlen;
} // }}}
void vk_vb_get_raw_name( vbuf *out, struct vk_key *vk ) // {{{
{
	vbuf_set( out, vk->keyname, vk->len_name );
} // }}}
void vb_vk_get_name( vbuf *vb, struct vk_key *vk ) // {{{ Places name of vk in vb in utf8
{
	int r;
	VBUF_STATIC( rawbuf, 10 );

	vk_vb_get_raw_name( rawbuf, vk );

	switch( vk->flag ) {
		case 0:
			//UTF16
			vb_utf16_ensure_zero_termination( rawbuf );
			ASSERT( 0 == vb_utf16to8( vb, rawbuf->b, rawbuf->dlen ), "utf16 to utf8 problem.");
			break;
		case 1:
			//ISO8859
			vb_utf8_ensure_zero_termination( rawbuf );
			vb_iso8859_1toutf8( vb, rawbuf->b, rawbuf->dlen );
			break;
		default:
			DIE("vk_get_name(): I don't know how to deal with type %d", vk->flag );
	}

}
// }}}
#if 0
int vk_get_nameT( struct vk_key *vk, char *buf, int n ) // {{{
{
	int sz = vk->len_name;

	if( sz + 2 > n ) DIE("vk_get_name(): sorry, output buffer is too small %d <= %d.", sz + 2, n );

	if( 0 == sz ) { buf[0] = '\0'; return 0; }

	WARN("size: %d", sz );
	memcpy( buf, vk->keyname, sz );

	switch( vk->flag ) {
		case 0:
			//UTF16
			buf[sz]     = '\0';
			buf[sz + 1] = '\0';
			if( 0 != sz % 2 ) DIE("vk_get_name(): attempt to convert UTF16 string of non-multiple-of-two size.");
			utf16to8( buf, buf, n );
			break;
		case 1:
			//ISO8859
			buf[sz] = '\0';
			iso8859_1to8( buf, buf, n );	
			break;
		default:
			DIE("vk_get_name(): I don't know how to deal with type %d", vk->flag );
	}

	if( gverbose ) fprintf(stderr, "vk_get_name() returning \"%s\"\n", buf );

	return sz;
}
#endif
// }}}
int Aget_val_len(struct hive *hdesc, int vofs, char *path) // {{{ Get len of a value, given current key + path 
{
	struct vk_key *vkkey;
	int vkofs;
	int len;

	if( gverbose ) fprintf( stderr, "Moo3: 0x%x\n", vofs);
	vkofs = get_file_offset( hdesc, trav_path(hdesc, vofs,path,1) );
	if (!vkofs) return -1;
	vkofs +=4;
	vkkey = (struct vk_key *)(hdesc->buf + vkofs);

	len = vkkey->len_data & 0x7fffffff;

	if ( vkkey->len_data == 0x80000000 ) {  /* Special inline case, return size of 4 (dword) */
		len = 4;
	}

	return(len);
}
// }}}
struct hbin_block *vk_alloc( struct hive *hdesc, void *rel, const char *name, int type, const char *buf, ulong blen ) // {{{
{
	struct hbin_block *newhblock;
	struct hbin_block *hbdata;
	struct vk_key *vk;
	//ulong vlen;
	VBUF_STATIC( utf16buf, 10 );

	if( -1 == vb_utf8to16( utf16buf, name, strlen( name ) + 1 ) ) {
		ERR_WIN(rlERROR_BROKEN_ERROR, "vk_alloc(): couldn't convert name \"%s\" from UTF-8 to UTF-16: %s", name, strerror( errno ) );
		return NULL;
	}

	ASSERT( !(blen & 0x80000000), "vk_alloc(): blen & 0x80000000, I can't deal with this.");

	/* Allocate vk descriptor including its name */
	if( NULL == ( 
		newhblock = hbin_alloc(hdesc,
			rel,
			sizeof(struct vk_key) + utf16buf->dlen) ) ) {
		//fprintf( stderr, "vk_alloc(): couldn't allocate a block for the vk.\n");
		ERR_WIN( rlERROR_BROKEN_ERROR, "couldn't allocate block for vk");
		return NULL;
	}
	vk = (struct vk_key*)newhblock->data; 

	vk->ofs_data = 0;

	/* Allocate a data block unless we inline it */

	if (type == REG_DWORD || type == REG_DWORD_BIG_ENDIAN) {
		vk->len_data = 0x80000004;  /* Prime the DWORD inline stuff */
		memcpy( &vk->ofs_data, buf, 4 );
	} else {
		if( NULL == (hbdata = hbin_alloc( hdesc, rel, blen ) ) ) {
			hbinb_free( hdesc, newhblock );
			return NULL;
		}

		memcpy( hbdata->data, buf, blen );

		vk->ofs_data = get_heap_offset( hdesc, hbdata );
		vk->len_data = blen;
	}

	/* Fill in vk struct */
	vk->id = ID_VK_KEY;
	vk->len_name = utf16buf->dlen;

	vk->val_type = type;
	vk->flag     = 0;   /* 1 -> name is ascii, 0 -> name is utf16 */
	vk->dummy1   = 0;
	memcpy( vk->keyname, utf16buf->b, utf16buf->dlen );
	//strcpy((char *)&vk->keyname, name);  /* And copy name */

	return newhblock;

} // }}}
struct hvlist *hvlist_alloc( struct hive *hdesc, struct nk_key *nk, int no ) // {{{
{
	return (struct hvlist*)hbin_alloc(hdesc, nk, no * 4);
} // }}}
int is_ascii( unsigned char *buf, int blen ) // {{{ TODO: this should be elsewhere
{
	int i;
	for( i=0; i < blen; i++ ) {
		if( buf[i] >= 128 ) return 0;
	}

	return 1;
} // }}}
int nk_set_value(struct hive *hdesc, struct nk_key *nk, const char *name, int type, const char *buf, ulong blen ) // {{{ Add a value to an nk.
{
	/* Just add the metadata (empty value), to put data into it, use
	* put_buf2val afterwards
	* hdesc - hive
	* nkofs - current key
	* name  - name of value
	* type  - type of value
	* returns: 0 err, >0 offset to value metadata
	*/
	//  int oldvlist = 0; //newhblock;
	struct hbin_block *oldvlist = NULL;
	struct hbin_block *newvlist = NULL;
	struct hbin_block *vkhbin = NULL;
	struct vk_key *newvkkey = NULL;
	VBUF_STATIC( utf16buf, 10 );
	VBUF_STATIC( hexbuf, 10 );
	//char *blank="";

	//parse_nk_nice( hdesc, nk );
	//hive_scan( hdesc );
	ASSERT( NULL != name, "nk_set_value() got null name" );


	if( type == REG_SZ || type == REG_EXPAND_SZ ) {
		//if( !is_ascii( (unsigned char *)buf, blen ) ) {
			WARN("Converting non-ascii UTF8 to UTF16");
			if( -1 == vb_utf8to16( utf16buf, buf, blen ) ) {
				vbuf_strset( hexbuf, "" );
				vbuf_cathex( hexbuf, buf, blen );
				ERR_WIN(rlERROR_BROKEN_ERROR, "nk_set_value(): couldn't convert buf [len:%d: %s] from UTF-8 to UTF-16: %s", blen, hexbuf->b, strerror( errno ) );
				return -1;
			}

			buf = utf16buf->b;
			blen = utf16buf->dlen;
		//}
	}

	//TODO: check to see if the value exists already or not

	if (nk->no_values) oldvlist = get_ptr_off_h( hdesc, nk->ofs_vallist );

	//parse_vlist( hdesc, nk_get_vlist( hdesc, nk ), nk->no_values );

	if( NULL == ( newvlist = (struct hbin_block*)hvlist_alloc( hdesc, nk, nk->no_values + 1 ) ) )
		return -1;

	//fprintf(stderr, "newvlist: F:%lx\n", get_file_offset( hdesc, newvlist ) );

	if (oldvlist) {   /* Copy old data if any */
		memcpy(newvlist->data, oldvlist->data, nk->no_values * 4 + 4);
	}

	//parse_vlist( hdesc, (struct v_list*)newvlist->data, nk->no_values );

	if( 0x6e967c == get_file_offset( hdesc, newvlist->data ) ) {
		fprintf(stderr, "Is this the end?\n");
	}

	if( NULL == (vkhbin = vk_alloc( hdesc, newvlist, name, type, buf, blen ) ) ) {
		hbinb_free( hdesc, newvlist ) ;
		ERR_WIN( rlERROR_BROKEN_ERROR, "Couldn't allocate vkhbin." );
		return -1;
	}

	//parse_vlist( hdesc, (struct v_list*)newvlist->data, nk->no_values );

	//fprintf(stderr, "newvlist: F:%lx\n", get_file_offset( hdesc, vkhbin ) );

	newvkkey = (struct vk_key*)vkhbin->data;

	/* Add pointer in value list */
	*(int *)( newvlist->data + (nk->no_values * 4)) = get_heap_offset( hdesc, vkhbin );

	/* Finally update the key and free the old valuelist */
	nk->no_values++;
	nk->ofs_vallist = get_heap_offset( hdesc, newvlist );
	if (oldvlist) hbinb_free( hdesc, oldvlist );

	//parse_vlist( hdesc, nk_get_vlist( hdesc, nk ), nk->no_values );
	//parse_nk_nice( hdesc, nk );

	return 0;
}
// }}}
void vk_delete( struct hive *hdesc, struct vk_key *vk ) // {{{  Remove a vk-struct incl dataspace if any
{
	if ( !(vk->len_data & 0x80000000) && vk->ofs_data) {
		hbinb_free( hdesc, get_hbin_block( hdesc, get_ptr_off_h( hdesc, vk->ofs_data ) ) );
		//free_block(hdesc, vk->ofs_data + REGF_HEADER_SIZE);
	}

	//TODO:  The -4 crap has to go
	hbinb_free( hdesc, get_hbin_block( hdesc, (void*)vk - 4 ) );
}

// }}}
// {{{ **** LX functions
lx_key* pull_lxkey( struct hive *hdesc, void *p ) // {{{ if p is lx_key then return else DIE()
{
	lx_key *result = p;

	if( gverbose )fprintf(stderr, "pull_lxkey(): id: 0x%x (%s)\n", result->head.id, id_to_str( result->head.id ) );

	if( result->head.id == ID_LF_KEY ||
		result->head.id == ID_LH_KEY ||
		result->head.id == ID_LI_KEY ||
		result->head.id == ID_RI_KEY )
	{
		return result;
	}
	else {
		DIE("pull_lxkey(): Invalid LX_KEY (id=0x%x,%s) at 0x%lx\n",
			result->head.id,
			id_to_str( result->head.id ),
			get_file_offset( hdesc, p ) );
		return NULL;
	}
}
// }}}
lx_key *lx_alloc( struct hive *hdesc, void *p, int id, int count ) // {{{ Allocate lx_key of size count close (as possible) to p and return it
{
	struct hbin_block *hb;
	lx_key *result;
	int newsize;

	switch( id ) {
		case ID_LF_KEY:
			newsize = sizeof( struct lf_hash );
			break;
		case ID_LH_KEY:
			newsize = sizeof( struct lh_hash );
			break;
		default:
			ERR_UNIX( EINVAL,
			"lx_alloc(): type 0x%04x is not not 'lf' or 'lh'. \n",
			id);
			return NULL;
	}

	newsize *= count; 
	newsize += sizeof( struct lx_head );

	if( NULL == ( hb = hbin_alloc( hdesc, p, newsize ) ) ) return NULL;

	result = (lx_key*)&hb->data;

	result->head.no = count;
	result->head.id = id;

	return result;
}
struct li_key *ri_get_li( struct hive *hdesc, struct ri_key *ri, int off ) // {{{
{
	return check_li_ptr( hdesc, get_hbin_block(
		hdesc, get_ptr_off_h( hdesc, ri->hash[off] ) )->data );
}
// }}}
lx_key *ri_get_lx( struct hive *hdesc, struct ri_key *ri, int off ) // {{{
{
	return pull_lxkey( hdesc, get_hbin_block(
		hdesc, get_ptr_off_h( hdesc, ri->hash[off] ) )->data );
}
// }}}
lx_key *ri_find_lx( struct hive *hdesc, struct ri_key *ri, int off, int *lx_off, int *lxp_off) // {{{ find the immediate child of ri and the offset within that child
{
	int roff=0, cntr=0;
	for( roff=0; roff < ri->head.no; roff++ ) {

		lx_key *lx = ri_get_lx( hdesc, ri, roff );
		if( cntr + lx->head.no > off ) {
			if( lx_off )
				*lx_off = off - cntr;
			if( lxp_off )
				*lxp_off = roff;
			//fprintf(stderr, "Returning lx %d\n", lx->head.no);
			return lx;
		}
		cntr += lx->head.no;
	}


	ERR_UNIX( EILSEQ,
		"ri_find_lx(): seeking %d of only %d records.\n",
		off,
		cntr );

	return NULL;
	/*
	for( i=0,cntr=0; i<ri->head.no && cntr < off; i++ )
	{
		struct li_key *li = ri_get_li( hdesc, ri, i );

		if( cntr + li->head.no >= off )
			return lx_get_nk( hdesc, (lx_key*)li, off - cntr );
		cntr+=li->head.no;
	}
	*/
} // }}}
#if 0
struct nk_key *lx_get_nk( struct hive *hdesc, lx_key *lx, int off ) // {{{ This is old, use lx_get_subkey
{
	struct hbin_block *hb;
	switch( lx->head.id )
	{
	case ID_RI_KEY:
	{   /* Is it extended 'ri'-block? */

		lx_key *lxs;
		int lxs_off;
		int lxp_off;

		lxs = ri_find_lx(hdesc, &lx->rikey, off, &lxs_off, &lxp_off );

		return lx_get_nk( hdesc, (lx_key*)lxs, lxs_off );

		return NULL;

		break;
	}
	case ID_LF_KEY:
	{
		struct lf_key *lf= &lx->lfkey;
		WARN("lx_get_nk(): Plain handler starting.\n");
		if (lf->head.no <= 0 || off >= lf->head.no ) {
			return(0);
		}

		hb = get_hbin_block( hdesc, 
			get_ptr_off_h( hdesc, lf->hash[off].ofs_nk ));

		break;
	}

	case ID_LI_KEY:
	{
		struct li_key *li = (struct li_key*)lx;

		hb = get_hbin_block( hdesc, 
			get_ptr_off_h( hdesc, li->hash[off].ofs_nk ));
		break;
	}

	default:
		ERR_UNIX( EILSEQ,
		"lx_get_nk(): unknown ID: 0x%x (%s)",
			lx->head.id, id_to_str( lx->head.id ) );
		return NULL;
	}

	return check_nk_ptr( hdesc, hb->data );
} // }}}
#endif
// }}}


// }}}
// {{{ *** Hive control (load/save/close) etc 

void closeHive(struct hive *hdesc) // {{{ 
{
	DB( DB_VERB, "closing hive %s\n",hdesc->filename);
	hdesc_free( hdesc );
}
// }}}
void hdesc_free( struct hive *hdesc ) // {{{
{
	FREE(hdesc->filename);
	FREE(hdesc->buf);
	FREE(hdesc->hbb_avail);
	FREE(hdesc);
} // }}}
int writeHive(struct hive *hdesc) // {{{ Write the hive back to disk (only if dirty & not readonly )
{
	int len;

	if (hdesc->state & HMODE_RO) return(0);

	//TODO: I am disrespecting the DIRTY flag!  I am so disrespectful.
	//if ( !(hdesc->state & HMODE_DIRTY)) return(0);

	if ( !(hdesc->state & HMODE_OPEN)) { /* File has been closed */
		if (!(hdesc->filedesc = open(hdesc->filename,O_RDWR))) {
			ERR_UNIX( errno,"writeHive: open(%s) failed, FILE NOT WRITTEN!",
			hdesc->filename);
			return -1;
		}
		hdesc->state |= HMODE_OPEN;
	}  
	/* Seek back to begginning of file (in case it's already open) */
	lseek(hdesc->filedesc, 0, SEEK_SET);

	len = write(hdesc->filedesc, hdesc->buf, hdesc->size);
	if (len != hdesc->size) {
		ERR_UNIX( errno,"writeHive: write of %s failed.",hdesc->filename);
		return -1;
	}

	hdesc->state &= (~HMODE_DIRTY);
	return 0;
}
// }}}
struct hive *hdesc_alloc( const char *filename ) // {{{
{
	struct hive *hdesc;
	hdesc = F_MALLOC( sizeof( struct hive ) );//CREATE(hdesc,struct hive,1);
	size_t avs;

	hdesc->filename = str_dup(filename);
	hdesc->state = 0;
	hdesc->size = 0;
	hdesc->alloc_max = 0;
	hdesc->buf = NULL;
	avs = sizeof( struct hbin_block* ) * HBB_AVAIL_MAX;
	hdesc->hbb_avail = F_MALLOC( avs );
	bzero( hdesc->hbb_avail, avs );

	return hdesc;
} // }}}
struct hive *openHive(const char *filename, int mode) // {{{
{
	struct hive *hdesc=NULL;
	int fmode,r; //,vofs;
	int pages=0;
	struct stat sbuf;
	//unsigned long pofs;
	/* off_t l; */
	char *c;
	struct hbin_page *p=NULL;
	struct regf_header *hdr=NULL;
	//  int verbose = (mode & HMODE_VERBOSE);
	int verbose = gverbose;
	//  struct hbin_block *block;

	DB(DB_0, "openHive() starting on \"%s\" with mode %d.\n", filename, mode );

	if( NULL == ( hdesc = hdesc_alloc( filename ) ) )
		return NULL;

	if ( (mode & HMODE_RO) ) {
		fmode = O_RDONLY;
	} else {
		fmode = O_RDWR;
	}
	hdesc->filedesc = open(hdesc->filename,fmode);
	if (hdesc->filedesc < 0) {
		if( errno == EACCES ) {
			WARN("Error opening \"%s\" trying read-only.", hdesc->filename );
			fmode = O_RDONLY;
			mode &= HMODE_RO;
			hdesc->filedesc = open(hdesc->filename,fmode);

			if (hdesc->filedesc < 0) {
				ERR_UNIX( errno, "openHive(): Couldn't open hive file %s", hdesc->filename );
				goto fail;
			}
		}
	}

	if ( fstat(hdesc->filedesc,&sbuf) ) {
		//TODO: What is this about?
		ERR_UNIX( errno, "openHive(): stat failed on: %s", hdesc->filename );
		goto fail;
	}

	hdesc->size = sbuf.st_size;
	hdesc->state = mode | HMODE_OPEN;
	/*  fprintf(stderr,"hiveOpen(%s) successful\n",hdesc->filename); */

	/* Read the whole file */
	hdesc->buf = F_MALLOC(hdesc->size);
	r = read(hdesc->filedesc,hdesc->buf,hdesc->size);
	if (r < hdesc->size) {
		ERR_UNIX( (r<0)? errno:EFAULT,"Could not read file, got %d bytes while expecting %d",
		r, hdesc->size);
		goto fail;
	}

	//TODO: assert file_size > header size

	/* Pull some useful info */
	hdesc->head = (void *)hdesc->buf;
	hdesc->heap = hdesc->buf + REGF_HEADER_SIZE;
	hdesc->root_block = get_hbin_block( hdesc, ((void*)hdesc->heap + hdesc->head->ofs_rootkey) );
	hdesc->root_key = check_nk_ptr( hdesc, hdesc->root_block->data);

	if( DEBUG_LEVEL >= DB_VERB ) {
		parse_nk_nice( hdesc, hdesc->root_key );
		DB( DB_VERB, "openHive(): file size: 0x%x", hdesc->size );
		parse_regf_header( hdesc, hdesc->head );
	}

	hdr = (struct regf_header *)hdesc->buf;
	if (hdr->id != ID_REGF) {
		ERR_UNIX( EILSEQ,
			"openHive(%s): File does not seem to be a registry hive!\n",filename);
		return(hdesc);
	}

	p = NULL;
	while ( (p = hbinp_get_next_hbinp( hdesc, p ) ) ) {
		//int c=0;
		struct hbin_block *b;

		b = &p->start_block;

		pages++;

		if (p->ofs_next == 0) {
			DB( DB_WARN, "pagesize zero!");
			return(hdesc);
		}
	}

	DB(DB_VERB, "Loaded hive file %s, %d pages found", filename, pages );

	return(hdesc);

fail:
	if( NULL != hdesc )
		closeHive(hdesc);
	return NULL;

}
// }}}

// }}}
// }}}

char *display_ptr_offset( struct hive *hdesc, void *p ) // {{{
{
	VBUF_STATIC( r, 100 );
	vbuf_printf( r, "@FILE:%08lx", get_file_offset( hdesc, p ) );
	return r->b;
} // }}}
void parse_db_struct( struct hive *hdesc, struct hdb_key *db ) // {{{
{
	int i;

	fprintf( stderr, "%s == db\n", display_ptr_offset( hdesc, db ) );
	fprintf( stderr, "hbb_dlen: %08x\n", hbinb_dlen( (HBB*)db ));

	fprintf( stderr, "short1  : %08x\n", db->short1 );
	fprintf( stderr, "offset1 : %08x -> %08x\n",
			db->offset1, db->offset1 + REGF_HEADER_SIZE );

	fprintf( stderr, "unknown1: %08x -> %08x\n",
			db->unknown1, db->unknown1 + REGF_HEADER_SIZE );

	hexdump( (char*)db, 0, hbinb_size( (HBB*)db ), 1 );

	fprintf( stderr, "\n\n");
} // }}}

int good_hoff( struct hive *hdesc, long off ) { // {{{

	if( off > 0 && off < hdesc->size ) 
		return 1;
	else
		return 0;
} // }}}
int good_pointer( struct hive *hdesc, void *nfc ) { // {{{

	if( (char*)nfc < (char*)hdesc->heap || ( (char*)nfc - (char*)hdesc->heap ) > hdesc->size )
		return 0;
	else
		return 1;
} // }}}

// {{{ *** DB_KEY stuff (mostly exploratory, not very functional)

void parse_nfc_struct( struct hive *hdesc, int level, struct hbin_block *hb ) // {{{
{
	int i;

	struct hnfc1 *nfc = (struct hnfc1*)hb;

	for(i=0;i<level;i++) { fprintf( stderr, "\t" ); }
	fprintf( stderr, "== nfc1 (%s) (lvl %02d;hbb_dlen 0x%08d) nfc1\n", display_ptr_offset( hdesc, nfc ), level, hbinb_dlen( hb ));

	if( !good_pointer( hdesc, nfc ) ) {
		fprintf(stderr, "%s Bad pointer.\n", display_ptr_offset( hdesc, nfc ) );
		return;
	}

	for(i=0;i<3;i++) {
		int j;
		for(j=0;j<level;j++) { fprintf( stderr, "\t" ); }
		fprintf( stderr, "[r%08x -> %08x: %c]",
			nfc->u[i], nfc->u[i] + REGF_HEADER_SIZE, (nfc->u[i] & 0x80000000)?'n':'s' );
		void *p2 = get_ptr_off_h( hdesc, nfc->u[i] );
		if( good_pointer( hdesc, p2 ) )
			parse_hbin_head_short( hdesc, p2 );
		else
			fprintf(stderr, "\n");
	}
	fprintf( stderr, "\n");
	parse_hbin_block( hdesc, hb );
} // }}}
void check_nfc1( struct hive *hdesc, int level, void *p ) // {{{
{
	struct hbin_block *hb = get_hbin_block( hdesc, p );
	struct hnfc1 *n = (struct hnfc1*)hb;

	if( good_pointer( hdesc, hb ) ) {
		parse_nfc_struct( hdesc, level, hb );
		int i;
		for( i=0; i<3; i++ ) {
			struct hbin_block *h2 = get_ptr_off_h( hdesc, n->u[i] );
			DB( DB_INFO, "-------------------\n");
			if( good_pointer( hdesc, h2 ) ) {
				DB( DB_INFO, "%s HBB_dlen: 0x%lx\n",
						display_ptr_offset( hdesc, h2 ),
						hbinb_dlen( h2 ) );

				parse_hbin_block( hdesc, h2 );
				/*
				if( i >= 2 )
					parse_hbin_block( hdesc, h2 );
				else 
					fprintf(stderr, "Skipping data block.\n");
					*/
			}
			else
				DB( DB_INFO, "%s Bad pointer.\n", display_ptr_offset( hdesc, h2 ) );
			/*
			if( level < 1 ) {
				check_nfc1( hdesc, level + 1, get_ptr_off_h( hdesc, n->u[i] ) );
			}
			*/
		}
	}
	else {
		DB( DB_INFO, "%s Bad pointer.\n", display_ptr_offset( hdesc, hb ) );
	}
} // }}}
void parse_un1( struct hive *hdesc, struct un1 *un ) // {{{
{
	int i;
	DB( DB_INFO, "== un1 %s\n", display_ptr_offset( hdesc, un ) );
	for( i=0; i<4; i++) {
		DB( DB_INFO, " field %d: %04hx/%05hd\n", i, un->v[i], un->v[i] );
	}
} // }}}
void check_wtf( struct hive *hdesc, HBB *hbb ) // {{{
{
	struct hdb_key *dk = (struct hdb_key*)hbb;
	struct hbin_block *uk = get_ptr_off_h( hdesc, dk->unknown1 );

	struct un1 *un;

	//struct hbin_block *uk2 = get_ptr_off_h( hdesc, uk->data );

	parse_db_struct( hdesc, dk );

	if( good_pointer( hdesc, uk ) ) {
		parse_hbin_block( hdesc, uk );
		un = (struct un1*)uk;
		parse_un1( hdesc, un );
	}
	else {
		DB( DB_INFO, "Unknown1 == bad pointer\n");
	}

	check_nfc1( hdesc, 0, get_ptr_off_h( hdesc, dk->offset1 ) );

	/*
	fprintf(stderr, "====================\n" );
	fprintf(stderr, "Unknown1 : hbb_dlen: 0x%08x\n", hbinb_dlen( uk ) );

	fprintf(stderr, "%08x\n", *(short*)(uk->data + 4) - *(short*)(uk->data) );
	*/

	/*
	fprintf(stderr, "====================\n" );
	fprintf(stderr, "Unknown1': %08x\n", hbinb_size( uk2 ) );
	hexdump( uk2->data, 0, hbinb_size( uk2 ) - 8, 1 );
	*/
	

	/*
	WARN( "uh-oh. %hx %x->%x %x", dk->len_what, dk->unknown1, dk->unknown1 + REGF_HEADER_SIZE, dk->unknown2 );

	parse_hbin_block( hdesc, mb );
	hexdump( mb->data, 0, hbinb_size( mb ), 1 );
	fprintf(stderr, "Go here: %x\n", vk->ofs_data + REGF_HEADER_SIZE );
	hexdump( r, 0, 16, 1 );
	*/
} // }}}
// }}}
// vim600: set foldlevel=0 foldmethod=marker:
