/* {{{ preamble
 * regstructs.h
 *
 * Structs that represent the format of data stored in registry hives and
 * related functions
 *
 * Authors:	Sean Loaring
 *		Petter Nordahl-Hagen
 *
 * Copyright (c) 2005-2006 Sean Loaring
 * Copyright (c) 1997-2004 Petter Nordahl-Hagen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 }}} */

/* {{{ A Brief Primer on Registry Binary Format
 *
 * [Registry Hive File]	:= [regf_header][hbin_page][hbin_page]...[hbin_page]
 *
 * [hbin_page] 		:= [hbin_page_header][hbin_block][hbin_block]...[hbin_block]
 *
 * [hbin_block]		:= [hbin_block_header][record]
 *
 * [record]		:= [nk_key] | [vk_key] | [lx_key] | [v_list] | [vk_data]
 *
 * nk_key:	Corresponds to a "regsitry key"
 *
 * vk_key:	A value key - corresponds to a "registry value"
 *
 * lx_key:	A list of offsets to nk_keys (see below)
 *
 * v_list:	A list of offsets to vk_keys.  These are NOT (necessarily) stored alphabetically
 *
 * vk_data:	The actual data of a value.  No header, just raw binary
 * 
 *
 * An Explanation Of The Whole LX Mess
 *
 * Every NK (Key) record has a list of subkeys.  This list is refered to here as an LX Key.
 *
 * Sadly there are many different types of subkey lists.  An LX Key is a
 * generic reference to a list of some type or another.  Things that operate on
 * LX keys must do different things with the list, depending on its type.
 *
 * We find type 'id' in lx_head.  This id defines the list type:
 *
 * LI	(Used in Windows 3.x)
 *
 *	The LI key is just a list of offsets to NK records.
 *
 *	I haven't delt with LI keys.  They probably don't work with this
 *	software at this time.
 *
 * LF 	(Used in Windows NT4.0+?)
 *
 *	Like the LI the LF key holds a list of offsets to NK records.  On top
 *	of this it also stores the first four characters of the names of the NK
 *	records.  With LI one has to look at all of the NK children to find the
 *	one that is desired.  With LF one only need to look at all of the
 *	records that share the first four characters.
 *
 *	I know what you are thinking.  Why on earth would they use the first
 *	four characters rather than a hash?
 * 	
 * LH NT5.0 onward?
 *
 * 	This type is very similar to an LF record.  The difference is that they
 * 	store a hash of the NK name rather than the first four characters of
 * 	the name.  
 *
 * RI
 *
 *	The RI key is an ugly, ugly hack that is employed when an NK list grows
 *	so huge that it can't be stored in one block anymore.
 *
 *	Rather than storing a list of NK keys the RI stores a list of lists of RI keys.  
 *
 *
 * }}} */
#ifndef REGSTRUCTS_H
#define REGSTRUCTS_H

// {{{ struct regf_header
/* The first page of the registry file is some kind of header, lot of
 * it's contents is unknown, and seems to be mostly NULLs anyway.
 * Note also, that this is the only place in the registry I've been
 * able to find _any_ kind of checksumming
 */

struct regf_header {

  long id;            /* 0x00000000	D-Word	ID: ASCII-"regf" = 0x66676572 */
  long unknown1;      /* 0x00000004	D-Word	???? */
  long unknown2;      /* 0x00000008	D-Word	???? Always the same value as at 0x00000004  */
  char timestamp[8];  /* 0x0000000C	Q-Word	last modify date in WinNT date-format */
  long unknown3;      /* 0x00000014	D-Word	1 */
  long unknown4;      /* 0x00000018	D-Word	3 - probably version #. 2 in NT3.51 */
  long unknown5;      /* 0x0000001C	D-Word	0 */
  long unknown6;      /* 0x00000020	D-Word	1 */
  long ofs_rootkey;   /* 0x00000024	D-Word	Offset of 1st key record */
  long lastblock;     /* 0x00000028	D-Word	Offset to last hbin_block */
  long unknown7;      /* 0x0000002C	D-Word	1 */
  char name[0x1fc-0x2c];   /* Seems like the hive's name is buried here, max len unknown */
  long checksum;      /* 0x000001FC	D-Word	Sum of all D-Words from 0x00000000 to 0x000001FB */
};
// }}} 
struct key_head { // {{{ generic key header in which to store the key id
	unsigned short id;
};
// }}}
/// {{{ struct sk_key : security descriptor
/* Security descriptor. I know how it's linked, but don't know
   how the real security data is constructed, it may as well
   be like the higher level security structs defined by MS in its
   includes & NT docs. Currently, I have no use for it.
   Note that keys sharing the exact same security settings will
   most likely point to the same security descriptor, thus
   saving space and making it fast to make objects inherit settings
   (is inheritance supported? they speak of security inheritance as a "new"
    feature in the filesystem on NT5, even though I think it was
    also supported by the lower levels in the earlier versions)
*/
struct sk_key {

  short id;          /* 0x0000	Word	ID: ASCII-"sk" = 0x6B73        */
  short dummy1;      /* 0x0002	Word	Unused                         */
  long  ofs_prevsk;  /* 0x0004	D-Word	Offset of previous "sk"-Record */
  long  ofs_nextsk;  /* 0x0008	D-Word	Offset of next "sk"-Record     */
  long  no_usage;    /* 0x000C	D-Word	usage-counter                  */
  long  len_sk;      /* 0x0010	D-Word	Size of "sk"-record in bytes   */
  char  data[4];     /* Security data up to len_sk bytes               */

};

/// }}}

/* The page header, I don't know if the 14 "dummy" bytes has a meaning,
 * they seem to be mostly NULLS
 */

#define HBB_MIN_LEN ( sizeof( struct hbb_head ) + 4 )
struct hbb_head {
	long size;
};

struct hbin_block { // {{{
	long size; 	/* 0x0000 D-Word - length of block (including head), negative if used, positive if free*/
	char data[1];	/* 0x0004 ??? - The data */
};

typedef struct hbin_block HBB;
// }}}
struct  hbin_page { // {{{
	long id;          /* 0x0000	D-Word	ID: ASCII-"hbin" = 0x6E696268  */
	long ofs_from1;   /* 0x0004	D-Word	Offset from the 1st hbin-Block */
	long ofs_next;    /* 0x0008	D-Word	Offset to the next hbin-Block (from THIS ONE)  */
	char dummy1[14];
	long len_page;    /* 0x001C	D-Word	Block-size??? Don't look like it,
					I only use the next-offset in this program  */
		    /* 0x0020   First data block starts here           */
	struct hbin_block start_block;     
};
// }}}

struct lx_head { // {{{
	short id; //ID_LF_KEY or ID_LH_KEY
	short no;
};
// }}}
// {{{ struct lf_hash, lf_key
/* This is the subkeylist/hash structure. NT4.0+.
 * ID + count, then count number of offset/4byte "hash". (not true hash)
 * Probably changed from the 3.x version to make it faster to
 * traverse the registry if you're looking for a specific name
 * (saves lookups in 'nk's that have the first 4 name chars different)
 */

struct lf_hash {
	long ofs_nk;    /* 0x0000	D-Word	Offset of corresponding "nk"-Record  */
	char name[4];   /* 0x0004	D-Word	ASCII: the first 4 characters of the key-name,  */
} *hash;

struct lf_key {
	struct lx_head head;
	struct lf_hash hash[];
};

// }}}
// {{{ struct lh_hash, lh_key
struct lh_hash {
	long ofs_nk;    /* 0x0000 D-Word Offset of corresponding "nk"-Record  */
	long hash;      /* 0x0004 D-Word ASCII: the hash of the key name,  */
};


struct lh_key {

	struct lx_head head;
	  
	/* WinXP uses a more real hash instead (base 37 of uppercase name chars)  */
	/* 		padded with 0's. Case sensitiv!                         */

	struct lh_hash hash[];
};

// }}}
// {{{ struct li_hash, li_key
/* 3.x version of the above, contains only offset table, NOT
 * any start of names "hash". Thus needs 'nk' lookups for searches.
 */

struct li_hash {
	long ofs_nk;    /* 0x0000	D-Word	Offset of corresponding "nk"-Record  */
};
struct li_key {

	struct lx_head head;
	struct li_hash hash[];
};
// }}}
// {{{ struct ri_hash (when an li_key grows too big it gets hacked up into an ri_key
/* This is a list of pointers to struct li_key, ie
 * an extention record of many li's.
 * This happens in NT4&5 when the lf hashlist grows larger
 * than about 400-500 entries/subkeys??, then the nk_key->ofs_lf points to this
 * instead of directly to an lf. (and it switches to li!)
 * Likely to happen in HKLM\Software\classes (file extention list) and
 * in SAM when many users.
 */

typedef long li_offset;
struct ri_key {
	struct lx_head head;
	li_offset hash[];
};
// }}}
// {{{ union lx_key, lx_hash
typedef union { 
	struct lx_head head;
	struct lf_key lfkey;
	struct lh_key lhkey;
	struct li_key likey;
	struct ri_key rikey;
} lx_key; 

typedef union
{
	struct lf_hash lfhash;
	struct lh_hash lhhash;
} lx_hash;
// }}}

struct v_list { // {{{ the value list pointed to by nk_key->ofs_vallist
	int vofs[1];
};

struct hvlist { // {{{ the value list pointed to by nk_key->ofs_vallist
	struct hbb_head hbh;
	int ofs[1];
};

// }}}
// {{{ struct vk_key
/* This is the value descriptor.
 * If the sign bit (31st bit) in the length field is set, the value is
 * stored inline this struct, and not in a seperate data chunk -
 * the data then seems to be in the type field, and maybe also
 * in the flag and dummy1 field if -len > 4 bytes
 * If the name size == 0, then the struct is probably cut short right
 * after the val_type or flag.
 * The flag meaning is rather unknown.
 *
 * NOTE: len_data
 *
 * The values for len_data are as follows:
 *
 * len_data == 0x00000000	data length is 0
 * len_data == 0x80000000	data are stored in val_type field, TYPE == REG_DWORD (I think)
 * len_data & 0x80000000	data are stored in ofs_data field
 *
 */
struct vk_key {
				/* Offset	Size	Contents                 */
	short id;         	/* 0x0000	Word	ID: ASCII-"vk" = 0x6B76  */
	short len_name;   	/* 0x0002	Word	name length              */
	long  len_data;   	/* 0x0004	D-Word	length of the data (see NOTE: above) */
	long  ofs_data;   	/* 0x0008	D-Word	Offset of Data           */
	long  val_type;   	/* 0x000C	D-Word	Type of value            */
	short flag;       	/* 0x0010	Word	Flag - 0:name is utf16, 1:name is ascii (I think)*/
	short dummy1;     	/* 0x0012	Word	Unused (data-trash)      */
	char  keyname[1]; 	/* 0x0014	????	Name                     */
};
// }}}
// {{{ struct nk_key
/* This is the key node (ie directory) descriptor, can contain subkeys and/or values.
 * Note that for values, the count is stored here, but for subkeys
 * there's a count both here and in the offset-table (lf or li struct).
 * What happens if these mismatch is not known.
 * What's the classname thingy? Can't remember seeing that used in
 * anything I've looked at.
 */
struct nk_key {
	short id;             /*  0x0000	Word	ID: ASCII-"nk" = 0x6B6E                        */
	short type;           /*  0x0002	Word	root-key: 0x2C, iso-8859-1: 0x20, utf-16: 0x00 */
	char  timestamp[12];  /*  0x0004	Q-Word	write-date/time in windows nt notation         */
	long  ofs_parent;     /*  0x0010	D-Word	Offset of Owner/Parent key                     */
	long  no_subkeys;     /*  0x0014	D-Word	number of sub-Keys                             */
	char  dummy1[4];
	long  ofs_lf;         /*  0x001C	D-Word	Offset of the sub-key lf-Records               */
	char  dummy2[4];
	long  no_values;      /*  0x0024	D-Word	number of values                               */
	long  ofs_vallist;    /*  0x0028	D-Word	Offset of the Value-List                       */
	long  ofs_sk;         /*  0x002C	D-Word	Offset of the sk-Record                        */
	long  ofs_classnam;   /*  0x0030	D-Word	Offset of the Class-Name                       */
	char  dummy3[16];
	long  dummy4;         /*  0x0044	D-Word	Unused (data-trash)                            */
	short len_name;       /*  0x0048	Word	name-length                                    */ 
	short len_classnam;   /*  0x004A	Word	class-name length                              */
	char  keyname[1];     /*  0x004C	????	key-name                                       */
};
// }}}

// {{{ DB_KEY stuff
// 060713
// I have just found a new structure in the registry.  When value data grows
// beyond some size a db_key will take the place of the data.  
//
// {{{ struct db_key
struct hdb_key {
	struct hbb_head hb;
	short id;         
	short short1; //structure depth, node count?
	long  offset1; //a pointer to an hnfc1
	long  unknown1; //a pointer to another hnfc1 struct?  
};
// }}}
struct hnfc1 { // {{{
	struct hbb_head hb;
	long u[3]; // pointers to hbins that hold data or *maybe* to other hnfc1 structs
};
// }}}
struct un1 { // {{{  //unknown part of DB_KEY
	struct hbb_head hb;
	unsigned short v[4];
}; // }}}
// }}}
#endif

// vim600: set foldlevel=0 foldmethod=marker:
